{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This units contains some basic methods common to all frameworks.

 **********************************************************************}

{$longstrings on}
{$codepage utf-8}
{$include un_globaldefs.h}

unit framework_basic;
interface
  uses SysUtils, classes, math, cge;

  // TAbstractFramework is defined in cge

Const
  MaxKey = 255;
  MinimumPaddingToScreenEdges = 100;

type
  TPointerBitmap = record
    Bitmap: array[0..31, 0..31, 0..3] of byte;
    Andmask, XorMask: array [0..31] of dword;
  end;

  TBasicFramework = class (TAbstractFramework)
  protected
    f_wsco, f_state: TWindowStateCommandedByTheOs;
    MxFpsCull, MyFpsCull: Longint;
    PreviousFpsCullTsc, PreviousCcToMTsc: qword;
    f_focus: boolean;
    f_cursortexture: glUint;
    CursorCreated,
    f_HadMessages: boolean;
    f_LoseFocusFramesCount: integer;
    f_KeyState: array[0..MaxKey] of boolean;
    f_MouseX, f_MouseY: integer;
    f_PointerGrabbed,
    f_PointerVisible,
    f_prevUseHwCursor : boolean;
    PointerBitmap, PointerCache: TPointerBitmap;
    PointerXHotSpot, PointerYHotSpot: integer;
    formerScreenRect, formerClientRect: TWindowManagerRect;
    f_window_subsystem_initialized: boolean;
    CombinedMx, CombinedMy: float;
    rMxPrev, rMyPrev: longint;
    gcMxPrevi, gcMyPrevi: longint;
    MouseIsGamepad: boolean;
    BatteryStatusCheckMoment: TdateTime;
    PrevBatteryRemaining: integer;

    function IsHotkey(scancode: integer; HK: array of integer): boolean;
    function _readKeyState (i: integer): boolean; virtual;

    function GetScreenRect(fullscreen: boolean): TWindowManagerRect; virtual; abstract;//if not fullscreen then return the working area (excluding the taskbar)
    function ClipWindowToRect(win, lrect: TWindowmanagerRect): TWindowmanagerRect; //works with *outer* rect (including decorations)
    function GetWindowOuterRect(): TWindowManagerRect; virtual; abstract;   //Ask directly from the OS
    function GetWindowClientRect(): TWindowManagerRect; virtual; abstract;  //Ask directly from the OS
    function GetIsWindowFullscreen(): boolean; virtual; abstract; //Ask directly from the OS
    function GetIsWindowMinimized(): boolean; virtual; abstract; //Ask directly from the OS
    procedure CorrectFullscreenWindowSize; virtual; abstract; //if screen size changes between frames: for example, laptop cover open

    procedure LoadWindowRectFromConfig(); //screen rect must be initialized prior!
    procedure CheckWindowSize; virtual;
    procedure EmulateMouseViaGamepad; virtual; abstract;
    procedure _SwapBuffers; virtual; abstract; //calls xxxSwapBuffers(), nothing more
    procedure _CreateGLcursor;
    procedure _DrawGLcursor;
    procedure _UploadGLcursor;
    procedure Heartbeat; virtual; abstract; //processing window messages and such
    procedure MaybeMakeAScreenshot; virtual;//see Mother^.Display.Screenshot
    procedure Pulse; //engine logic abstracted from platform
  public
    AnyKey: boolean;

    class var FramesSinceLeavingFullscreen: int64;

    property WindowSubsystemInitialized: boolean read f_window_subsystem_initialized;
    property WindowStateCommandedByTheOs: TWindowStateCommandedByTheOs read f_wsco write f_wsco;

    //implementing abstract functions that make up the framework API
    function MouseX: float; override;
    function MouseY: float; override;
    procedure MainThreadMainLoop; override;
    procedure CheckBatteryStatus; override; // a stub, preferrably be overridden
                                            // by descendants
    procedure SwapBuffers; override; //performs a lot of timing calculations
                                     // wrapped around _SwapBuffers


    //dragged from the old paradigm, public to satisfy gamepad and tabled manager classes using these
    // These methods add input events to the mother queue
    procedure OnMouseMove;
    procedure OnPress(scancode: integer);
    procedure OnRelease(scancode: integer);
    procedure OnType(input: UnicodeChar);
    procedure OnPen(x, y, pressure, reserved1, reserved2, reserved3: single);
  end;

implementation



  function TBasicFramework.MouseX: float;
  begin
    Result:= CombinedMx;
  end;


  function TBasicFramework.MouseY: float;
  begin
    Result:= CombinedMy;
  end;


  procedure TBasicFramework.OnMouseMove;
   var ne: integer;
  begin
     if f_pointerGrabbed then begin
       CombinedMx+= f_MouseX;
       CombinedMy+= f_MouseY;
       gcMxPrevi:= round(CombinedMx);
       gcMyPrevi:= round(CombinedMy);
       MouseIsGamepad:= false;
       ne:= NewInputEvent();
       with Mother^.Input.Events[ne] do begin
         EventType:= CGM_MOUSEDELTA;
         mx:= MouseX();
         my:= MouseY();
       end;
     end
     else begin
       if (rMxPrev <> f_MouseX) or (rMyPrev <> f_MouseY)
       then begin
         //did it *really* move? Only then shutdown the fake pointer position made by the game controller right thumb
         CombinedMx:= f_MouseX;
         CombinedMy:= f_MouseY;
         gcMxPrevi:= round(CombinedMx);
         gcMyPrevi:= round(CombinedMy);
         MouseIsGamepad:= false;
         rMxPrev:= f_MouseX;
         rMyPrev:= f_MouseY;
       end;
       ne:= NewInputEvent();
       with Mother^.Input.Events[ne] do begin
         EventType:= CGM_MOUSEMOVE;
         mx:= MouseX();
         my:= MouseY();
       end;

       //because the inpus system reacts asynchronously, lagging behind game logic commands
       //we have to reset it, just in case
       ne:= NewInputEvent();
       with Mother^.Input.Events[ne] do begin
         EventType:= CGM_MOUSEDELTA;
         mx:= 0;
         my:= 0;
       end;
     end;
   end;

   function TBasicFramework.IsHotkey(scancode: integer; HK: array of integer): boolean;
   var
     i: integer;
   begin
     Result:=No;
     if Length(HK) = 0 then Exit;
     For i:=0 to Length(HK) - 2 do
       if not _readKeyState(HK[i]) then Exit;
     if HK[Length(HK) - 1] = scancode then Result:=Yes;
   end;

   procedure TBasicFramework.OnPress(scancode: integer);
   var //i: integer;
     ne: integer;
   begin
     if (scancode < 0) or (scancode > 255) then Die('Invalid scancode %0',[scancode]);
     if (scancode = ord(Key_Space)) then Module.ReactToSpace;
     if (scancode = ord(Key_Escape)) then Module.ReactToEsc;

     if Mother^.Module.State <> ms_Loaded
       then with Mother^.Input do begin
         TextInput:= false;
         GrabPointer:= false;
       end;

     If IsHotKey(scancode, AbortHotKey) then begin
       _RequestExit;
       if Mother^.Debug.Verbose then AddLog('Hot keys pressed - exit requested.');
       Exit;
     end;
     If IsHotKey(scancode, RestartHotKey) then begin
       Mother^.State.RestartRequested:=Yes;
       if Mother^.Debug.Verbose then AddLog('Hot keys pressed - restart requested.');
       Exit;
     end;
     If IsHotKey(scancode, LanguageHotKey) then begin
       if Mother^.Debug.Verbose then AddLog('Hot keys pressed - language reset.');
       MessageContainer.SetLanguage(-1);
       Exit;
     end;
     If IsHotKey(scancode, FullscreenHotkey) then begin
       if Mother^.Debug.Verbose then AddLog('Hot keys pressed - full screen/windowed switch requested.');
       if Mother^.Display.InFullScreenMode
         then Mother^.Display.RequestedFullScreenMode:= -1
         else Mother^.Display.RequestedFullScreenMode:= 1;
     end;


  {   if Mother^.State.DeveloperMode then
       For i:=1 to 10 do
         If IsHotKey(scancode, ModuleHotkey[i]) then begin
           Module.Reload(i);
           Exit;
         end;
  }

     If Mother^.State.DeveloperMode and IsHotKey(scancode, SessionRollBackHotKey) then begin
         Module.RollBackSession;
         Exit;
       end;

     If IsHotKey(scancode, ModuleHotkey[0]) then begin
         Module.GoToModuleSelectionMenu;
         Exit;
       end;
  {   If IsHotKey(scancode, MainModuleHotKey) then begin
       ModuleManager._ActiveModuleNum:=0;
       State.VerboseLog('Hot keys pressed - switching to default module.');
       Exit;
     end;}
     if IsHotKey(scancode, ConsoleToggleHotkey) then begin
       Console.AlwaysVisible:=not(Console.AlwaysVisible);
       Exit;
     end;

     if IsHotKey(scancode, QualityFactorManualOverrideToggleHotkey) then begin
       Mother^.Timer.QF_ManualOverride:=not(Mother^.Timer.QF_ManualOverride);
       if Mother^.Timer.QF_ManualOverride
         then AddLog(RuEn('Коэффициент качества переведён на ручное управление.','Quality Factor manual override on.'))
         else AddLog(RuEn('Коэффициент качества переключён обратно на автомат.','Quality Factor control switched back to automatic.'));
       Exit;
     end;

     if IsHotKey(scancode, ConsoleScrollUpHotkey) then begin
       if Console.AlwaysVisible then begin
         Console.ScrolledBack:=Console.ScrolledBack + 1;
         Exit;
       end;
       if Mother^.Timer.QF_ManualOverride then begin
         Mother^.Timer.QF += max(1, round(Mother^.Timer.QF / 15));
         if Mother^.Timer.QF > QF_MAX then Mother^.Timer.QF:= QF_MAX;
         EXit;
       end;
     end;

     if IsHotKey(scancode, ConsoleScrollDownHotkey) then begin
       if Console.AlwaysVisible then begin
         Console.ScrolledBack:=Console.ScrolledBack - 1;
         Exit;
       end;
       if Mother^.Timer.QF_ManualOverride then begin
         Mother^.Timer.QF-= max(1, round(Mother^.Timer.QF / 15));
         if Mother^.Timer.QF < QF_MIN then Mother^.Timer.QF:= QF_MIN;
         EXit;
       end;
     end;

     AnyKey:= True;
     ne:= NewInputEvent();
     with Mother^.Input.Events[ne] do begin
       EventType:= CGM_PRESS;
       key:= TKey(scancode);
       mx:= MouseX;
       my:= MouseY;
     end;
   end;

   procedure TBasicFramework.OnRelease(scancode: integer);
   var ne: integer;
   begin
     ne:= NewInputEvent();
     with Mother^.Input.Events[ne] do begin
       EventType:= CGM_RELEASE;
       key:= TKey(scancode);
       mx:= MouseX;
       my:= MouseY;
     end;
   end;

   procedure TBasicFramework.OnType(input: UnicodeChar);
   var ne: integer;
   begin
     AnyKey:= True;
     ne:= NewInputEvent();
     with Mother^.Input.Events[ne] do begin
       EventType:= CGM_TYPE;
       character:= input;
       mx:= MouseX;
       my:= MouseY;
     end;
   end;

   procedure TBasicFramework.OnPen(x, y, pressure, reserved1, reserved2, reserved3: single);
   var ne: integer;
   begin
  //AddLog('Pen tablet: x=%0   y=%1   pressure=%2', [x,y,pressure]);
     Mother^.PenTablet.ActivityDetected:= true;
     ne:= NewInputEvent();
     with Mother^.Input.Events[ne] do begin
       EventType:= CGM_PEN;
       PenX:= x;
       PenY:= y;
       PenPressure:= pressure;
       PenReserved1:= reserved1;
       PenReserved2:= reserved2;
       PenReserved3:= reserved3;
     end;
   end;


   // The main heartbeat event
   procedure TBasicFramework.Pulse;
   var
     delta: double;
     mycull: integer;
  //qq: qword;
   begin
     if Assigned(Mother^.LockupGuard) then Mother^.LockupGuard.Checkup;

     if Mother^.Gamepad.Present then EmulateMouseViaGamepad
     else begin
       CombinedMx:= f_MouseX;
       CombinedMy:= f_MouseY;
     end;

     Mother^.Display.WindowHasFocus:= Self.f_state in [wsco_Normal, wsco_Maximized];

  //addlog('%1.Pulse, module is %0',[Module, AnsiString(Self.ClassName)]);
     //if {Assigned(Module) and} (Focus
     //or (f_LoseFocusFramesCount <= FRAMES_TO_CONTINUE_RENDER_AFTER_LOSING_FOCUS)) then begin

     mycull:= FPS_CULL_DEFAULT ;
     delta:= abs(MouseX - MxFpsCull) + abs(MouseY - MyFpsCull);
     if (delta >= FPS_CULL_DEFAULT_MOUSE_DELTA_THRESHOLD)
       and not Mother^.Display.MouseIsBeyondOurWindow
       and Self.f_focus
       then mycull:= FPS_CULL_DEFAULT_MOUSE_MOVED;

     Mother^.Timer.CullFrameRate:= mycull;
     Mother^.Timer.SuspendSwapBuffersMeasurement:= false;

     Console.RecalculateFade;

     if Assigned(Module) then begin
       Module.CheckDate;
       Module.Pulse;
     end
     else
       Console.DrawBackground;

     Console.RecalculateFade;

     MeasureFps;

     NoteThreadOccupation(tlk_ConsoleRender);
     Console.Draw;
     RenderFpsCounter;

     if Assigned(Console) and Mother^.Display.LogoAnimationPlaying
       then Console.DrawLogo;

     NoteThreadOccupation(tlk_Working);

     {$ifndef cpuarm}
     RecheckTSCFreq;
     {$endif}


  //UsecDelta(@qq);
     SwapBuffers;

     CheckBatteryStatus;

  //addlog('> %0',[round(UsecDelta(@qq))]);
     if (Mother^.Timer.CullFrameRate < 300) then begin

       delta:= UsecDelta(@PreviousFpsCullTsc) - 1000000.0 / max(FPS_CULL_DEFAULT, Mother^.Timer.CullFrameRate);

       delta:= 1.0 * Sign(delta) * Math.min(abs(delta), 1000000.0 / FPS_CULL_DEFAULT); {
         куллинг дельты - абсолютно необходимая вещь. Бывает, кадр занимает секунд по пять,
         при старте или если антивирус дрочит при запуске модуля.
       }
       if delta < 0 then begin
          Mother^.Timer.HiResSleep(-Round(delta / 1000.0));
          delta:= 0;
       end;
       UsecDelta(@PreviousFpsCullTsc);
       PreviousFpsCullTsc+= round(delta);
       MxFpsCull:= round(MouseX);
       MyFpsCull:= round(MouseY);
     end;
     //end
     //else Sleep(NO_FOCUS_SLEEP_INTERVAL);
  //addlog('guz5');
  //addlog('* %0',[round(UsecDelta(@qq))]);
   end;

  function TBasicFramework.ClipWindowToRect(win, lrect: TWindowManagerRect): TWindowManagerRect;
  begin
    Result.left:= Math.max(win.left, lrect.left);
    Result.top:= Math.max(win.top, lrect.top);
    Result.width:= Math.min (win.width, lrect.width);
    Result.height:= Math.min(win.height, lrect.height);
  end;

   procedure TBasicFramework.CheckBatteryStatus;
   begin
     if PrevBatteryRemaining <> 0 then Exit;
       if Mother^.Debug.Verbose then AddLog('%0 does not support battery status, it always reports "AC power"', [AnsiString(Self.ClassName)]);
     Mother^.State.BatteryRemaining:= 255;
     PrevBatteryRemaining:= 1;
   end;

  procedure TBasicFramework.LoadWindowRectFromConfig;
  begin
    with Mother^.Display.WindowRect do begin
      left:= Mother^.Display.ScreenRect.left + Round(Mother^.Display.ScreenRect.width * Config.GetFloat('video', 'window_left', 0.2));
      top:= Mother^.Display.ScreenRect.top + Round(Mother^.Display.ScreenRect.height * Config.GetFloat('video', 'window_top', 0.2));
      width:= Round(Mother^.Display.ScreenRect.width * Math.Max(0.1, Config.GetFloat('video', 'window_width', 0.6)));
      height:= Round(Mother^.Display.ScreenRect.height * Math.Max(0.1, Config.GetFloat('video', 'window_height', 0.6)));
    end;
    Mother^.Display.WindowRect:= ClipWindowToRect(Mother^.Display.WindowRect, Mother^.Display.ScreenRect);
  end;


 Procedure TBasicFramework.SwapBuffers;
  var
    q_prev: qword;
  begin
    glGetError;

    if not Mother^.Display.UseHardware1bitCursor then _DrawGLcursor;

    MaybeMakeAScreenshot;

      NoteThreadOccupation(tlk_SwapBuffers);
    if Mother^.GAPI.SkipSwapBuffers then begin
      if Mother^.Debug.Verbose then AddLog('Skipped SwapBuffers().');
      Mother^.GAPI.SkipSwapBuffers:= false;
    end
    else begin
      UsecDelta(@q_prev);

      _SwapBuffers; //call the concrete implementation (wgl / glx / egl)

      if Mother^.GAPI.Mode in [gapi_GL21, gapi_DeprecatedImmediateGL ] then glFinish();
      {this is ignored by modern drivers
        (ogl 3.0+), but on legacy hardware with old drivers (ogl 2.x)
        this is *vital* for getting a proper measured value.

        Yes, I'm still testing on GeForce FX 5200
      }

      //this shit is IMPORTANT! Used in quality control & load balancing.
      Mother^.Timer.LastSwapBuffersDuration:= UsecDelta(@q_prev);
    end;

      NoteThreadOccupation(tlk_Working);

    if (Mother^.Timer.FPS < HW_POINTER_LOW_FPS_THRESHOLD) and not MouseIsGamepad
      then Mother^.Display.UseHardware1bitCursor:= Yes;
    if Mother^.Timer.FPS > HW_POINTER_HIGH_FPS_THRESHOLD
      then Mother^.Display.UseHardware1bitCursor:= No;

    //Otherwise in Windows 7 pointer is invisible when hovered over frame borders
    if Mother^.Display.MouseIsBeyondOurWindow and not MouseIsGamepad
                                    then Mother^.Display.UseHardware1bitCursor:= Yes;

    if (Mother^.Timer.QF > fixedfont_min_qf_to_use_hq_texture)
    and not Mother^.Text.FixedFontEmergencyDownsample
       then Mother^.Text.FixedFontQuality:= 0;
    if Mother^.Timer.QF < fixedfont_max_qf_to_use_lq_texture
       then Mother^.Text.FixedFontQuality:= 1;

    inc(CurrentFrame);
    CurrentFrameDate:= Now();

    SwapStubTexture; //so that if flashes horribly
  end;


  Procedure TBasicFramework.MainThreadMainLoop;
  begin
    try
      Mother^.State.ProgramInitialized:= Yes;

      repeat
        Heartbeat; //polling the window manager events and stuff
        Pulse; //includes calling the module DLL pulse event
        NowRestarting:= No;
      Until Mother^.State.ExitRequested or Mother^.State.RestartRequested;

      if Mother^.Debug.Enabled then AddLog('...exited the main loop.');
    except
      EverythingIsOk:=No;
      if not (ExceptObject is EDying) then Try
          Die(MI_ERROR_UNCAUGHT_EXCEPTION);
      Except
      End;
    end;
  end;

  procedure TBasicFramework.CheckWindowSize;
//  var    AtNSZ: boolean = false; //adjust to new screen size
  begin
    //each frame. A dumb but *robust* method.
    Mother^.Display.ScreenRect:= GetScreenRect(Mother^.Display.InFullScreenMode); //Width:= GetSystemMetrics(SM_CXSCREEN);
    Mother^.Display.FullscreenRect:= GetScreenRect(true);
    Mother^.Display.WindowRect:= GetWindowOuterRect;
    Mother^.Display.WindowClientRect:= GetWindowClientRect;
//    Mother^.Display.ScreenHeight:= GetSystemMetrics(SM_CYSCREEN);
//    GetClientRect(Mother^.Display.WindowHandle, rect);
    case f_wsco of
      wsco_Normal, wsco_NoFocus: begin

      //Was nice idea but it FUCKING DOESN"T WORK!
      //Windows 7 simply ignores my calls to SetWindowPos() and MoveWindow() and resizes the window how it sees fit.
(*
        if (formerw <> Mother^.Display.ScreenWidth) or (formerh <> Mother^.Display.ScreenHeight)
        then begin
          l:= round(Mother^.Display.WindowedOriginX * Mother^.Display.ScreenWidth);
          t:= round(Mother^.Display.WindowedOriginY * Mother^.Display.ScreenHeight);
          w:= round(Mother^.Display.WindowedWidth * Mother^.Display.ScreenWidth);
          h:= round(Mother^.Display.WindowedHeight * Mother^.Display.ScreenHeight);

          AddLog(RuEn(
            'Внезапно! Размер экрана изменился (было %0х%1, стало %2х%3). Корректирую размер и положение окна на %4,%5 (%6х%7).',
            'Suddenly! The screen size changed (was %0x%1, now is %2x%3). Correcting the size and position of the window to %4,%5 (%6x%7).')
            ,[formerw, formerh, Mother^.Display.ScreenWidth, Mother^.Display.ScreenHeight,l,t,w,h]);
          //simply reset some values to bogus so that the algorithms below activate:
          //ShowWindow(Mother^.Display.WindowHandle,SW_RESTORE);//FTF does it get minimized somehow during the transition?
          if not MoveWindow(Mother^.Display.WindowHandle,l,t,w,h, false) then Die('MoveWindow() retuened false!');
          AtNSZ:= true;
        end;
*)

        if (formerClientRect <> Mother^.Display.WindowClientRect)
          and not GetIsWindowFullscreen
            //Because f_wsco denotes *desired* state, not thereal state
            //of the window. Windows window management has its say too.
        then begin
          //that wine hack again (two frames after return from fullscreen,
          //  the window suddenly gets repositioned slightly to the left and top
          //  of the screen)
          if not ((Mother^.State.OS = ostWine) and (FramesSinceLeavingFullscreen < 3))

          then with Mother^.Display.WindowRect do begin
            Config.Float['video', 'window_width']:= width / Mother^.Display.ScreenRect.Width;
            Config.Float['video', 'window_height']:= height / Mother^.Display.ScreenRect.height;
            Config.Float['video', 'window_left']:= (left - Mother^.Display.ScreenRect.left) / Mother^.Display.ScreenRect.width;
            Config.Float['video', 'window_top']:= (top - Mother^.Display.ScreenRect.top) / Mother^.Display.ScreenRect.height;
          end;
          if Mother^.Debug.Verbose then AddLog('Application window resized or moved, the client area is %0, the entire window including decorations is %1',
            [RectToStr(Mother^.Display.WindowClientRect), RectToStr(Mother^.Display.WindowRect)]);
        end;
        formerClientRect:= Mother^.Display.WindowClientRect;
      end;
    wsco_Maximized: begin
      if
        Mother^.Display.ScreenRect <> formerScreenRect
        then begin
          AddLog(RuEn(
            'Внезапно! Размеры окна не равны размерам экрана: %0, было %1. Корректирую.',
            'Suddenly! Window size doesn''t match screen size: %0, was %1. Correcting.')
            ,[RectToStr(Mother^.Display.ScreenRect), RectToStr(formerScreenRect)]);

          CorrectFullscreenWindowSize;

          Mother^.Display.WindowRect:= GetWindowOuterRect;
          Mother^.Display.WindowClientRect:= GetWindowClientRect;
        end;
      formerScreenRect:= Mother^.Display.ScreenRect;
      end;
    end;

    case Mother^.Display.RequestedFullScreenMode of
      -1: begin
          if f_wsco = wsco_Maximized then begin
            f_wsco:= wsco_Normal;
             if Mother^.Debug.Verbose then AddLog('The window manager is commanded to switch to windoweed mode.');
            Config.Bool['video', 'fullscreen']:= false;
          end;
          Mother^.Display.RequestedFullScreenMode:= 0;
        end;
       1: begin
           if f_wsco <> wsco_Maximized then begin
             f_wsco:= wsco_Maximized;
              if Mother^.Debug.Verbose then AddLog('The window manager is commanded to switch to fullscreen mode.');
             Config.Bool['video', 'fullscreen']:= true;
           end;
           Mother^.Display.RequestedFullScreenMode:= 0;
        end;
    end;
    Mother^.Display.CanRender:= not GetIsWindowMinimized
      and (Mother^.Display.WindowClientRect.width >= 256)
      and (Mother^.Display.WindowClientRect.height >= 256)

  end;

  function TBasicFramework._readKeyState (i: integer): boolean;
  begin
    Result:=f_KeyState[i];
  end;

  procedure TBasicFramework._CreateGLcursor;
  begin
    case  Mother^.GAPI.Mode of
      gapi_GL21, gapi_DeprecatedImmediateGL , gapi_GLES2: begin
        glGenTextures(1, @f_cursortexture);
        glBindTexture(GL_TEXTURE_2D, f_cursortexture);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      end;
    else
      DieUnsupportedGLMode;
    end;
  end;

  procedure TBasicFramework._UploadGLcursor;
  begin
    if (f_cursortexture = 0) then Self._CreateGLcursor;
    case  Mother^.GAPI.Mode of
      gapi_GL21, gapi_DeprecatedImmediateGL , gapi_GLES2: begin
        glBindTexture(GL_TEXTURE_2D, f_cursortexture);
        glTexImage2d(GL_TEXTURE_2D, 0, GL_RGBA, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, @PointerBitmap.Bitmap);
      end;
    else
      DieUnsupportedGLMode;
    end;
  end;

  procedure TBasicFramework._DrawGLcursor;
  var
    x, y: integer;
    Mesh: TDumbUniMesh;
  begin
   if f_PointerVisible and not Mother^.Display.MouseIsBeyondOurWindow
   then begin
     case  Mother^.GAPI.Mode of
       gapi_DeprecatedImmediateGL: begin
         glEnable(GL_TEXTURE_2D);
         glEnable(GL_ALPHA_TEST);
         glEnable(GL_BLEND);
         glAlphaFunc(GL_GREATER, 0);
         glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
         glDisable(GL_CULL_FACE);
         glDisable(GL_DEPTH_TEST);
         glViewport(0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
         glMatrixMode(GL_PROJECTION);
         glLoadIdentity();
         glMatrixMode(GL_MODELVIEW);
         glLoadIdentity();
         glOrtho(0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height, 0, 0 ,1);
         glBindTexture(GL_TEXTURE_2D, f_cursortexture);
         x:= round(MouseX - PointerXHotSpot);
         y:= round(MouseY - PointerYHotSpot);
{         glColor4f(1, 1, 1, Mother^.Display.FadeIn);
         glBegin(GL_QUADS);
           glTexCoord2f(0, 1);
           glVertex2f(x, y + 32); //bottom left corner
           glTexCoord2f(0, 0);
           glVertex2f(x, y);
           glTexCoord2f(1, 0);
           glVertex2f(x + 32, y);
           glTexCoord2f(1, 1);
           glVertex2f(x + 32, y + 32);
         glEnd;
}
      end;
    else
      DieUnsupportedGLMode;
    end;
     Mesh:= TDumbUniMesh.Create;
     Mesh.AddQuad2d(
       1, 1, 1, Mother^.Display.FadeIn,
       x, y, x+32, y+32,
       0, 0, 1, 1);
     Mesh.Render;
     Mesh.Free;
   end;
  end;

  procedure TBasicFramework.MaybeMakeAScreenshot;
  var
    s: integer;
    q: Qword;
  begin
    with Mother^.Display.Screenshot do begin
      if request = -1 then begin
        if Mother^.Debug.Verbose then AddLog('Got a request to delete screenshot data...');
        request:= 0;
        if Assigned(pdata) then begin
          FreeMem(pdata);
          pdata:= nil;
          if Mother^.Debug.Verbose then AddLog('  deleted unnecessary screenshot data (%0x%1)',[Width, Height]);
        end;
        width:= 0;
        height:= 0;
      end
      else if request = 1 then begin
        if Mother^.Debug.Verbose then AddLog('Got a request to copy back buffer data for a screenshot...');
        if Assigned(pdata) then begin
          FreeMem(pdata);
          pdata:= nil;
          if Mother^.Debug.Verbose then AddLog('  deleted existing screenshot data (%0x%1)',[Width, Height]);
        end;
        case  Mother^.GAPI.Mode of
          gapi_GL21, gapi_DeprecatedImmediateGL , gapi_GLES2: begin
            Width:= Mother^.Display.WindowClientRect.Width;
            Height:= Mother^.Display.WindowClientRect.Height;
            Format:= GL_RGB8;
            s:= Width * Height * 3 + 100000;
            (*
              Why not directly into a vampyre image: it's a wrong thread,
              that's why. After the initial loading, mother *cannot* use Vampyre
              as any vampyre calls should come from a specific module thread.
              Thus, we cache it here and let the module deal with it in the background.
            *)
            GetMem(pdata, s);
            if Mother^.Debug.Verbose then AddLog('  allocated %0 Kilobytes)',[s div 1024]);
            if Mother^.Debug.Verbose then AddLog('  calling glReadPixels()...');
            Mother^.Timer.UsecDelta(@q);
            glPixelStorei(GL_PACK_ALIGNMENT, 1);
            glReadPixels(0,0,Width,Height,GL_RGB,GL_UNSIGNED_BYTE,pdata);
            if Mother^.Debug.Verbose then AddLog('  Ok, %0x%1 in %2 μs',[Width, Height, round(Mother^.Timer.UsecDelta(@q))]);
            request:= 2; //mark as done
          end;
        else
          DieUnsupportedGLMode;
        end;
      end;
    end;
  end;

end.
