{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{
    This unit defined the basic classes of the game world.

    The layers are used for multiplayer (basically, a large-scale lag
    compensation), for advanced AI routines that run series of mini-simulations,
    and for illusion/mind control spell effects.

}

{
Для выбранной модели мультиплеера, равно как для работы продвинутого ИИ, движку
нужна поддержка расслаиваемого игрового мира.

Проблемы на пути к достижению:

1. Создание полной копии игрового мира невозможно вследствие технических
ограничений. Чеперси тупо не хватит быстродействия. Наилучший результат -
миллион объектов в секунду, вряд ли это можно хотя бы удвоить оптимизацией.
То есть сохранение всего игрового мира в TMemoryStream займёт не менее 50
милисекунд - в несколько раз больше, чем допустимое время кадра (16мс).
А если быть реалистом - то 100..300, в зависимости от насколько толстыми у меня
получатся чанки. И это - только сохранение, а нужна ещё загрузка.
То есть, тупое решение в лоб отпадает.
2. Необходимость *быстро* получать и сравнивать контрольные суммы, нужно для
обнаружения рассинхронизации в сетевой игре.
3. Необходимость привязывать копии одного объекта в разных слоях к рендеру,
чтобы могли наследовать туда-обратно имеющийся меш, даже если не совсем точно
отражает состояние объекта.

Намечаемое решение:

Все игровые объекты надо делать на основе того или иного рода умных указателей.
Тогда порядок такой: объект А запрашивает у центрального менеджера "дай мне
указатель на объект Б для текущего слоя" и получает либо базовый объект
Изначального Слоя, если в текущем слое нет изменённой копии, либо эту самую
изменённую копию.
Причём, важно, чтобы было два разных вида запросов: для чтения и запрос копии
для модификации.

Недостатки:
1. Паскаль не предоставляет средств для создания синтаксического сахара для
этого. Любое потенциально изменяющее объект взаимодействие нужно будет оформлять
в коде как последовательность "получить на запись - вызывать методы -
подтвердить окончание изменений". Если "получить на запись" всегда делать
безусловным клонированием, то третий шаг возможно отложить на после пульса,
когда менеджер объектов проходит по всем, вызывавшимся на запись и обновляет
их контрольные суммы. Возможно, с полным сличением копий? Тогда не отличающиеся
от оригинала стираются (изменений не было). Требуется небольшая доработка
Чеперси, чтобы обеспечить методы обхода двух потенциально идентичных объектов с
оставлением контрольной суммы.
2. Нужен механизм поиска алгоритмических ошибок, связанных с изменением
объектов, полученных только для чтения: когда он уже получен, это ссылка и
уследить за всеми манипуляциями трудно.
3. Нужен отдельный сборщик мусора для массива менеджера объектов, в котором
хранятся ссылки на них, чтобы не допустить повторного использования занятого
индекса (возможно, сам менеджер делает обход через каждые N пульсов логики, и
выкидывает из массива все, которые scraped?).

Достоинства:
1. За счёт всего вышеописанного ггеморроя получаем больше контроля и меньше
потерь/утечек. Поскольку объектам не позволено хранить ссылки на другие объекты,
а только индексы в менеджере, облегчается и ускоряется работа Чеперси по оходу
графа, включая сборщик мусора.
2. Выигрыш в размере объектов в 64-битной версии: поскольку индекс - это
Longint, он вдвое меньше указателя.
}

  {$include mo_globaldefs.h}
  {$typeinfo on}

unit ga_basics;

interface
uses
  SysUtils, Classes, math, {$ifdef windows} Windows, {$endif}
   typinfo, chepersy, mo_classes, mo_hub,
    mo_assets
  , syncobjs
 ;

 const
   GAME_OBJECT_BIT = 0;
   GAME_OBJECT_MASK = $00000001;
   FLOATING_LAYER_FLAG_BIT = 1;
   FLOATING_LAYER_FLAG_MASK = $00000002;


   MOSLTPOT = 12; //4096
   MAX_OBJECT_SLEEPING_TICS = 1 shl MOSLTPOT; { This governs performance.
     No object will be able to sleep longer than this, thus limiting on
     performance saving. But there would be also as much dynamic arrays
     per each game sublayer, to hold object indices.
   }
   MOSLTMASK = $ffffffff shr (32 - MOSLTPOT);
   OLTMASK = $3fffffff;

type

  //forward declarations
  TAbstractGame = class;
  TGame = class;
  TGameLayer = class;
  TGameObject = class;

  TGameObjectSmartPointer = packed object
    index: Longint;
    procedure Free;
    procedure Scrape;
    procedure DecrementRefCount;
    procedure IncrementRefCount;
    function Read: TGameObject;
    function Modify: TGameObject;
  end;

  operator := (o: TGameObject) a: TGameObjectSmartIndex;

  operator := (o: TGameObject) a: TGameObjectSmartIndex;
  begin
    if (a.index <> 0) and Assigned(o) and (o.index = a.index) then exit;
    if a.index <> 0 then a.DecrementRefCount;
    if not Assigned(o)
      then a.index:= 0
      else begin
        a.index:= o.index;
        a.IncrementRefCount;
      end;
  end;

type
  TGameObject = class(TChepersyObject)
  public
    index,

    NextTicks: Longint; { Still looping, compared to the 64-bit universal time,
      but at 30 bits (and $3fffffff). This is 12 days of gameplay. }

    constructor Create;
    procedure RegisterFields; override;
    function CloneAnother(another: TChepersyObject): boolean; override; { Keeps
      the original refcount and index }
    procedure Scrape; override;

  end;

  TGameObjectRefCounts = array[0..1] of TSparceLongintArray;

  TAbstractGame = class (TChepersyObject)
  protected
    procedure Tick; virtual; abstract;  { In theory, advances the base layer
        one tick. Implementation-dependent. }

  public
    BaseLayer: TGamelayer;
    RefCounts: TGameObjectRefCounts;
                     { Why the reference count when the objects are linked
      by indexes that are fed to appropriate layers to get the instance?
      Because an object instance may be shared among the child layers when the
      parent layer changes the base object and the children have the
      split-on-write property (which they will most often do)

      More importantly, the reference counters play the role of an index space,
      allowing fast and efficient search for free indices in the space shared
      by the theoretically unlimited number of layers.

      The index here is the highest (sign) bit of the index, calculated via shr 31,
      so RefCounts[0] contains only positives and RefCounts[1] negatives.
      }

    CurrentLayer: TGamelayer; //for faster access, changes as layers are processed

    constructor Create;

    procedure RegisterFields; override;
    procedure AddInputEvent(e: PInputEvent); virtual; abstract; { receives
        input messages as retranslated by Logic (partly float)
        MUST convert it to an internal all-integer format suitable for bonding
        with game objects or transmitting over the network.
      }
    procedure Pulse(ticks: integer); virtual; abstract; { The number of ticks
        is calculated by the Logic, usually based on timing.
      }
  end;

  TGame = class(TAbstractGame)
    procedure Tick; override;
  public
    procedure Pulse(ticks: integer); override;
  end;

  TGameObjectLayerArray = array of TGameLayer;

  TGameTickSchedule = class; //forward declaration

  TickScheduleArray = array of TGameTickSchedule; { Not static to allow
    adjusting the size during development. Chepersy currently cannot convert
    static arrays to a different size! }

  TGameLayer = class(TChepersyObject)
  public
    LayerKind: Longint; //0 core layer 1 auxilary layer
    NoSplittingOnWrite: boolean; { Normal behavior: when the parent layer
      modifies an object, it makes itself a copy and distributes the old one
      to all its children who have this flag set to False, if they do not have
      their own copy yet.

      Warning: setting this to true makes the layer incoherent! Use it for
        special cases only (like permanent illusions on world chunks
        that persist until the chunk in question dematerializes.) }

    Container: TSparceObjectArray;
    Parent: TGamelayer;
    Children: TGameObjectLayerArray;
    TickSchedules: TickScheduleArray;

    FullGameplayTime: int64; { The universal time span
                                      is +/- 290 thousand years of realtime
                                        ( 9223372036854775807 miliseconds ) }
    LoopingGameplayTime: SmallInt; //MAX_OBJECT_SLEEPING_TICS long, looping

    procedure RegisterFields; override;

    procedure Add(index: Longint; o: TGameObject); virtual;
    function GetForModification(index: Longint): TGameObject; virtual;
    function Get(index: Longint): TGameObject; virtual;
    function Has(index: Longint): boolean; virtual;
    procedure Remove(index: Longint); virtual;

    procedure Tick; virtual;
    function SpawnMultiplayerFloatingLayer(): TGameLayer;
    constructor Create;
    destructor Destroy; override;
    procedure Scrape; override;
  end;


  TGameTickSchedule = class (TGameObject) { Why such bizarre architecture?
      To make spawning a new, empty layer almost a free action. }
  protected
    type
      Longint32Array = array[0..31] of Longint;
    var
      stA: Longint32Array;
      dyA: TArrayOfLongint;
      f_count: Longint;
  public
    property Count: longint read f_count;
    destructor Destroy; override;
    procedure RegisterFields; override;
    procedure Add(ind: Longint); { They are never removed, for
        simplicity's sake. If the object itself says it is scheduled for a
        latter tick, it just gets skipped.
      The per-tick data gets erased when that tick is passed, anyway.
      One index may be included several times or included in several STs. This
        is resolved by the game layer using object's own fields.
      }
    function Get(num: Longint): Longint;
  end;



(*

  { WARNING: never change the field order! This type is exempt from the usual
      Chepersy enum/set conversion rules as its set is compressed to 2 bytes!
      Never increase their number beyond 16! }
  {$push}
    {$packenum 2}
    TGameLayerEnum = (
      gla_Prime, { The basis for everything. In single player this is likely
        to be the only existing layer. In multiplayer it runs in the past,
        200..1000ms behind the real time. The delay determines if the
        network packets traveling longer than that get discarded.

        The prime index is unique in that it is flat and linear. *ALL* other
        layers and sublayers reserve space in this index, even if the object
        in question does not belong to Prime Basic. This also saves heavily
        because all single-layer objects do not have to have an entry in that
        layer / sublayer sparce index. The performance impact for the Prime
        Basic index itself (may need to read an object's field before rejecting
        it)is deemed negligible.

        All other layers and sublayers have a sparce index to hold their

     }

      gla_Multiplayer1,
      gla_Multiplayer2 { These two are periodically flipped, not unlike the
          front and back buffer of the screen.
        While the front layer ticks in real time, exactly one tick per
          primeval tick, the back layer begins as a copy of the primeval,
          then begins catching up with the front layer, at at least 2 ticks
          per one primeval tick. When it catches up with the front layer,
          it takes its place. The former front layer is dumped, then recreated
          as the back layer, to begin catching up.
       }
       {
        NOTE: This is the only way I can think of to achieve full lag
          compensation for something as complex as fully destructible terrain
          that supports floods.
       }

    );

    {$packset 2}
    TGameLayerSet = set of TGameLayerEnum;

 const

    MAX_SUBLAYERS = 4096; {
      Sublayers are used for basic AI thinking simulations and spell effects
        like clairvoyance or illusions. Basically, these are very small partial
        layers with noticeable performance impact per layered object, unlike
        the main layers.
      Unlike the main layers, an object could only belong to a maximum of 2
        sublayers: the base sublayer and one of the 4095. Meaning that no two
        illusion spells may be affecting a character at the same time.
      Potentially extensible to 32767.
    }

 type

    {$bitpacking on}
    TGameLayerIndex = bitpacked record
    { This is registered in Chepersy as having only one field, of type dword. }
      BelongsToBaseSublayer: 0..1;
      Reserved: 0..3; { Can be used to add new fields OR extend the sublayer
        range }
      SecondSublayerNum: 0..MAX_SUBLAYERS - 1;
      BelongsToLayers: TGameLayerSet;
    end;

  {$pop}


 type

  TGameObjectIndex = packed record
    index: Longint;
  end;

  TLayerableGameObject = class;
  TLAGOArray = array of TLayerableGameObject;

  //***TODO: static array support in Chepersy
  TLayerWakingList = array[0..MAX_OBJECT_SLEEPING_TICS - 1]
                              of array of TGameObjectIndex; {  I fully delegate
        any performance optimizations to the memory manager here. The
        individual arrays would be (in the worst case of one million objects,
        and with equal distribution) 244 elements long. The memory manager
        copes well in such cases (Increasing array size by 1 at adding each
        new element)  }
  TSublayerRec = record


  end;

  TLayerRec = record
    Sublayers: array[] of TSublayerRec;
    Objects: array of TLayerableGameObject;
    SublayerObjects

  end;

  TLayerRecArray = array[TGameLayerEnum] of TLayerRec;

  TGameManager = class (TCgeObject)
  protected
    f_current_layer: TGameLayerEnum;
    f_layers: TLayerRecArray;
  public

    procedure AllocateLayer(Which: TGameLayerEnum);
    procedure TickLayer(Which: TGameLayerEnum); { Ticks are one millisecond
        long, thus the game world is running on a virtual frequency of
        one kiloHertz. }
    property CurrentLayer: TGameLayerEnum read f_current_layer;
    procedure RegisterFields; override;
    function GetObjectForReading(Index: TGameObjectIndex): TLayerableGameObject;
    function GetObjectForModification(Index: TGameObjectIndex): TLayerableGameObject;
    function GetObject(Index: TGameObjectIndex; ForModification: boolean;
                           FromLayer: TGameLayerEnum; FromSublayer: smallint): TLayerableGameObject;
  end;

  TLayerableGameObject = class (TCgeObject)
  protected
  public
    Me: TGameObjectIndex;
    MyLayers: TGameLayerIndex;
    { And AGAIN we have +8 bytes added to each class instance, in addition to
      4(8) of Pascal and 8 of Chepersy. To a total of 20(24). Just for an empty
      object with no functionality. I am getting depressed :( }

    procedure RegisterFields; override;
    class function GetForReading(Index: TGameObjectIndex): TLayerableGameObject; virtual;
    class function GetForModification(Index: TGameObjectIndex): TLayerableGameObject; virtual;
  end;
*)

  TPhysicalGameObject = class (TGameObject)
  protected

  public
    procedure RegisterFields; override;

  end;

{ The planet is a dubdivided icosahedron.
  Thus, the cells are not strictly equilateral and there *are* 20 points
    per planet where only 5 cells meet instead of usual 6. }

  TWorldCoordinateInt = Longint; { In millimeters, origin is planet center,
      te y axis matches the planet rotational axis towards its north pole.
    Allows for planets up to 1000 kilometers in radius. Above that, overflow
      lurks. }

  TGalaxyCoordinateInt = int64; { In millimeters again, this is enough to
      hold the real life distance to Alpha Centauri. }

  TWorldVector = packed record
    x,y,z: TWorldCoordinateInt;
  end;

  TgalaxyVector = packed record
    x,y,z: TGalaxyCoordinateInt;
  end;


  TCellCoordinateInt = ShortInt; //16-bit signed

  { Each cell considers itself to be a perfect equilateral prism, with its
      neighbors being non-equilateral prisms, together forming a perfect flat
      world.
    The y coordinates of all cells are equal. The x and z axes are
      assigned haphazardly by the world icosahedron subdivision process, so
      coordinates of different cells are converted using a simple 2x2 matrix,
      with each cell having 3 transformation matrices, one for each neighbor
      its sides touch. Note that physics interacts only with these cells,
      delegating interaction with 3 (rarely 2) others across the edge vertices
      to them.
    Any physics performed in the cell space assumes this perfect flat world,
      which works with reasonable approximation for the planetary surface but
      breaks too far below or above it. In such cases, world coordinates are
      used. The cells have a lower limit, with the inner volume of the planet
      considered to be consisting of magma and gas spherticles operating
      in the world  space, which are generated on contact with moving
      tetrameshes and dismissed upon losing the said contact. }
  TCellCoordinates = packed record
    y: TWorldCoordinateInt;
    x,z: TCellCoordinateInt;
  end;

  TCellToCellXZConversionMatrix = packed record
    a0, b0, a1, b1: Longint;
  end;

  TCell = class;

  TCellNeighbors = array[0..2] of Longint;
  TCellNeighborXZMatrices = array[0..2] of TCellToCellXZConversionMatrix;

  TTrizmoxelGridCoordinates = packed record
    x,z: TCellCoordinateInt;
  end;

  TTrizmoxelGridCoordinatesArray = array of TTrizmoxelGridCoordinates;

  TCell = class (TPhysicalGameObject)
  protected

  public
    Neighbors: TCellNeighbors; { Some of them *may* be legitimately zero
      in partial layers, AI simulations or non-planetary deathmatch maps }
    NeighborXZMatrices: TCellNeighborXZMatrices;

  end;

var
  Game: TAbstractGame = nil; //a global variable for easier access. Myst be set by self during Tick().

implementation

  {$include mo_ga_basics_go.inc}

  {$include mo_ga_basics_la.inc}

  {$include mo_ga_basics_tss.inc}

end.
