{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


unit mo_3dbasics;

{$include mo_globaldefs.h}

interface

uses
  SysUtils, Classes, math,
  chepersy, mo_hub, typinfo,
  mo_classes, mo_assets, mo_gmathbase;

type

  TRenderable = class// (TGenericAsset)
  public
    procedure Render; virtual; abstract;
  end;

  FixedFont = class
    { This one designed to convert existing code based on immediate mode with
        minimum hassle. NOT for efficiency.

      Differ from mother's header
        while implementation uses the same include (see un_fixed_font_inc ) }
  protected
    class var ffColor: TVector4f;
  public
    class procedure SizeToFace(size: integer; face: Pinteger; zoom: PGLfloat );
    class function GetStringWidth(Size, numChars: integer): GLfloat;
    class function GetStringHeight(Size, numLines: integer): GLfloat;
    class procedure SetRenderState (noshadow: boolean);
    class procedure FitStringIntoRectangle (
       Size, numChars, numLines, Width, Height: integer;  fit: PStringFitRec);
    class procedure RenderString (fit: PStringFitRec; Left, Top: glFloat; w: PUnicodeChar);

    class procedure Color3f(r, g, b: GLfloat);
    class procedure Color4f(r, g, b, a: GLfloat);
    class procedure Color3fv(p: PGLfloat);
    class procedure Color4fv(p: PGLfloat);
  end;

  TAbstractCamera = class;

  TAbstractScene = class (TCgeObject)
    //upload vertex data to VBOs, prepare rendering scenarios.
    procedure Prepare(Camera: TAbstractCamera); virtual; abstract;

    //Some cameras, like the fisheye camera, or the stereo camera,
    //  may call this method twice per frame
    procedure Render(Camera: TAbstractCamera); virtual; abstract;
  end;

  { TAbstractCamera }

const
  PiHalf = Pi / 2;
  PiFourth = Pi / 4;


type
  TAbstractCamera = class (TCgeObject)
  protected
    f_fov, f_hdr, f_farplane, f_nearplane, f_prevX, f_prevY, f_prevWidth, f_prevHeight: float;
    f_view_normal, f_up_normal, f_right_normal: TffVector;
    f_scene: TAbstractScene;

    function _GetHdr(): float; virtual;
    procedure _SetFov(newfov: float); virtual; //FOV of less than one degree switches to isometric projection
    procedure _SetScene(newscene: TAbstractScene); virtual;

    procedure RenderBufferToScreen; virtual; abstract;
    procedure RenderSceneToBuffer; virtual; abstract;
    procedure PrepareToRender; virtual; abstract;

    procedure RecalculateNormals(newUpVec: PffVector = nil); virtual;
  public
    constructor Create(
      _scene: TAbstractScene = nil;
      _fov: float = PiFourth;
      _viewportWidth: float = 256.0;
      _viewportHeight: float = 256.0;
      _viewportX: float = 0.0;
      _viewportY: float = 0.0 );

    // The up normal does not need to be perpendicular to the view normal,
    //  it's more like a hint. A simple upward pointing vector is enough
    //  for a classical first person shooter situation.
    // Both use RecalculateNormals to build a true up normal.
    procedure Point(vnormal, unormal: TffVector); overload; virtual;
    // The second version uses previous up normal on the assumption
    //   that the camera angle changes little between frames.
    procedure Point(vnormal: TffVector); overload; virtual;

    procedure RegisterFields; override;

    //called by viewport before it calls Render()
    //it is strongly encouraged to completely override this method in actual camera classes
    procedure ViewportResize(
      ViewportX: float = 0;
      ViewportY: float = 0;
      VieportWidth: float = 256.0;
      ViewportHeight: float = 256.0);  virtual;

    //normally called by the viewport
    procedure Cycle; virtual;

    //normally called by the viewport
    function OnKey(
        key: TKey;
        ViewportMX, ViewportMY: float //ranges 0.0..1.0 in viewport.
      ): boolean; virtual;
    //like TControl, this can return false to stop event propagation.
    //Normally always return true

    //considers projection curvature (for the fisheye camera).
    //may return false (for elliptic-screen fisheye camera when the point
    //  falls outside te renderbuffer corners. In such a case, the vector is undefined.
    function MouseToVector(mouseX, mouseY: float; var pointer_normal: TffVector
      ): boolean; virtual;

    procedure Render(); virtual;

    property HDR: float read _GetHdr write f_hdr; //update to the readed value
    //may be delayed due to lazy multisampler re-creation

    property Scene: TAbstractScene read f_scene write _SetScene;

    property Fov: float read f_fov write _SetFov; // HORIZONTAL. In radians.
    //fov of less than one degree switches most cameras to orthogonal projection

    property Width: float read f_prevWidth;
    property Height: float read f_prevHeight;

  end;

  {$include un_gl_common.h} { a slightly higher level library for working transparently with all OpenGL flavors }



//f_width, f_height prevW, prevH,

implementation

  uses mo_module;

  {$include un_gl_common.inc}

  {$include un_fixed_font.inc}

  const OneDegree = Pi / 180;

  procedure TAbstractCamera.RegisterFields;
  begin
    RegType(typeinfo(TffVector), sizeof (TffVector), ['x', 'y', 'z', typeinfo(fixed32)]);
    inherited RegisterFields;
    ListFields([
      'fov', @f_fov, 'hdr', @f_hdr, 'farplane', @f_farplane, 'nearplane', @f_nearplane,
      'prevx', @f_prevX, 'prevy', @f_prevY, 'prevwidth', @f_prevWidth,
      'prevheight', @f_prevHeight,
        typeinfo(float),
      'view_normal', @f_view_normal,
      'up_normal', @f_up_normal,
      'right_normal', @f_right_normal,
        typeinfo(TffVector),
      'scene' , @f_scene, CPS_MANAGED_OBJECT
    ]);
  end;

  procedure TAbstractCamera.ViewportResize(ViewportX: float; ViewportY: float;
    VieportWidth: float; ViewportHeight: float);
  begin
    //it is strongly encouraged to completely override this method in actual camera classes
    f_prevX:= viewportX;
    f_prevY:= viewportY;
    f_prevWidth:=  VieportWidth;
    f_prevHeight:= ViewportHeight;
  end;

  function TAbstractCamera._GetHdr: float;
  begin
    Result:= f_hdr;
  end;

  procedure TAbstractCamera._SetFov(newfov: float);
  begin
     //FOV of less than one degree means isometric projection
    f_fov:= Math.max(0.0, Math.min(Pi * 120 / 180, newfov));
  end;

  procedure TAbstractCamera._SetScene(newscene: TAbstractScene);
  begin
    f_scene:= newscene;
  end;

  procedure TAbstractCamera.RecalculateNormals(newUpVec: PffVector);
  var calcRightVec: TffVector;
  begin
    if not Assigned(newUpVec) then newUpVec:= @f_up_normal;
    CrossProduct(f_view_normal, newUpVec^, calcRightVec);
    if SquaredLength(calcRightVec) > FF_ONE_THOUSANDTH
      then f_right_normal:= calcRightVec;
      //the new right vector could be close to zero
      // is happens if camera suddenly looks straight up or straight down.
      // We just cheat by using the old value of the right vector

    //finally calculate the real up vector perpendicular to the view vector
    CrossProduct(f_right_normal, f_view_normal, f_up_normal);
  end;

  const
    ForwardVector: TffVector = (x: (i: 0); y: (i: 0); z: (i: FF_MINUS_ONE));
    UpwardVector: TffVector = (x: (i: 0); y: (i: FF_ONE); z: (i: 0));

  constructor TAbstractCamera.Create(_scene: TAbstractScene; _fov: float;
    _viewportWidth: float; _viewportHeight: float;
    _viewportX: float; _viewportY: float);
  begin
    inherited Create;
    f_scene:= _scene;
    f_prevX:= _viewportX;
    f_prevY:= _viewportY;
    f_prevWidth:=  _viewportWidth;
    f_prevHeight:= _viewportHeight;
    _SetFov(_fov);

    //You obviously can't just use declared constants to assign to fixed-point
    //  variables: any such assignment breaks determinism
    f_view_normal:= ForwardVector;
    f_up_normal:= UpwardVector;
  end;

  procedure TAbstractCamera.Point(vnormal, unormal: TffVector);
  begin
    f_view_normal:= vnormal;
    RecalculateNormals(@unormal);
  end;

  procedure TAbstractCamera.Point(vnormal: TffVector);
  begin
    f_view_normal:= vnormal;
    RecalculateNormals();
    //previous up normal will be used to calculate new up normal
  end;



  procedure TAbstractCamera.Cycle;
  begin
    if Module.Logic.IsLoadingAssets then Exit;
    if Assigned(f_scene) then f_scene.Prepare(Self);
  end;

  function TAbstractCamera.OnKey(key: TKey; ViewportMX, ViewportMY: float
    ): boolean;
  begin
    Result:= true;
  end;

  function TAbstractCamera.MouseToVector(mouseX, mouseY: float;
    var pointer_normal: TffVector): boolean;
  begin
    if f_fov < OneDegree then begin
      //orthogonal projection

    end
    else begin
      //perspective projection

    end;
  end;

  procedure TAbstractCamera.Render;
  begin
    if Module.Logic.IsLoadingAssets then Exit;
    try
      PrepareToRender;
      RenderSceneToBuffer;
      RenderBufferToScreen;
    except
      Die(RuEn(
        'Сбой прорисовки камеры %0 (%1)',
        'Camera %0 (%1) failed to render'
        ), [string(ClassName), pointer(Self)]);
    end;
  end;


end.
