{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{
  An "asset" here is some entity that is not stored in the database directly
      but is either loaded from a file or generated using some known formula.

  Examples:
    - a texture or any other OpenGL object
    - an image or sound
    - a texture atlas
    - a polygonal representation of a game world cell or a character model

  When the game module unloads, it passes all its assets to the mother module
     for safekeeping. OpenGL objects are represented by handles, mother memory
     chunks are represented by actual pointers as they are allocated
     by the mother module's memory manager (which, of course, is a different
     one from the game module DLL's memory manager).

  When the game module loads, it claims assets back from the mother module
     then deletes unclaimed assets. Thus, in the developer version, reloading
     a recompiled game module is instaneous an no assets are lost (except FBOs
     and shaders that have to be re-created for some reason, otherwise OpenGL
     crashes)

  When there is no stored asset to claim, it is considered lost and has to be
    "actualized" (loaded from a corresponding file or generated).
    After loading, the module keeps displaying the "loading" screen until all
    its necessary assets are actualized.
}
unit mo_assets;

{$include mo_globaldefs.h}

interface

uses
  sysutils, Classes, math, md5,  chepersy, mo_hub, typinfo, syncobjs,
  mo_classes, mo_threads, mo_image, mo_gmathbase, mo_mstream;

  procedure RegisterAssetClasses;

//  procedure StoreResourceClassesList;
//  procedure DeleteUnclaimedAssets;

  procedure RetrieveAssets;  //from mother memory stream

  procedure StoreAssets; //to a mother memory stream


  procedure WalkSessionToListAssetsToStore; //called when not really saving
  //the session (like in the case of a rollback)


type
  TGenericAsset = class;
  TAssetArray = array of TGenericAsset;

  TAssetContainer = class (TCgeObject)
  public
    List: TAssetArray;
    procedure Add(A: TGenericAsset);
    procedure ClearAndFree;
    procedure ClearAndForget;
    function GetByHash(ph: PAssetHash): TGenericAsset;
    procedure Remove(a: TGenericAsset);
    procedure RegisterFields; override;
    constructor Create;
    destructor Destroy; override;
    procedure MergeAndFree(var c2: TAssetContainer); //free and nil
    procedure MassSetUnclaimed;
    procedure MassResurrectScraped;
    procedure ZeroAllNotActualized;
  end;

  TAssetManager = class
  public
    class var
      ToStore, //is filled during saving, via OnBeforeSaving of assets
      Retrieved, //is loaded from the mother. It is a graph root.
      //Assets must nust not point back to game objects

      Temporary, //do NOT get saved to session (mostly assets to replace
     //existing primary assets, stored in skipped fields).
     //get added to unclaimed assets after loading, to be deleted in the background

     ToDeleteInMainThread: TAssetContainer; // Asset's _Release may move it here
       //at reaching zero refcount instead of calling destructor

    class var ToDeleteInMainThreadCS: TCriticalSection;
    class procedure Initialize;
    class procedure Finalize;
    class procedure Pulse;
  end;

const
  ASSET_MASK_BIT = 19;
  ASSET_MASK = dword(1) shl ASSET_MASK_BIT;


//TGenericAsset doesn't *have* a handle, nor GenResource/FreeResource methods
//  that had been moved to THandleAsset.

type
  TAssetPropertiesEnum = (
    ap_optional,
    ap_replacement,
    ap_actualized,
    ap_unclaimed,
    ap_temporary
  );
  TAssetPropertiesSet = set of TAssetPropertiesEnum;

  CGenericAsset = class of TGenericAsset;
  TGenericAsset = class (TCgeObject)
  protected
    f_props: TAssetPropertiesSet;
    f_hash: TAssetHash; //a perfect 16-byte hash -
       //see the Mother^.AssetKeeper.GenHash() function exported by the core module

    procedure _setHash(h: TAssetHash);

    function _GetActualizedNilSafe: boolean; //not virtual. Safe to call on nil.
    function _GetActualized: boolean; virtual;
    procedure MarkAsActualized;

    procedure DieBecauseStupidDeveloperForgotToOverrideMethod(s: UnicodeString);

  public
    //Compares to Module.FrameNum, to correct priorities for optional assets.
    //The SortOrder function should take this into account, adding non-linearity where necessary.
    InQueueSinceFrame: qword;

    //simply because the actualizing algorithm can lead to an almost infinite postponing loop
    //  naturally, there's a chance to waste some 30% CPU cycles in some circumstances,
    //  which I deem negligible, but then, an array-based queue would grow out of proportions!
    //Making the queue a chain of pointers is the only practical solution.
    NextInQueue: TGenericAsset;

    //The *priority* (updating an attacking orge is more important than updating
    //  a cactus behind the player) is determined by sorting threaded tasks, not assets

    constructor Create(p: TAssetPropertiesSet = []);

    function IsValid: boolean; virtual;
 //   function Valid: boolean; //NOT virtual as it could be safely invoked
      //as TGenericAsset(nil).Valid
      //calls protected virtual method IsValid


    //constructor Create(Optional: boolean = false);
    property Hash: TAssetHash read f_hash write _setHash;
    procedure UpdateHash; virtual;

    procedure RegisterFields; override;
    procedure AfterLoading; override;
    procedure BeforeSaving; override;
    procedure OnModuleDestroy; override;

    function Actualize: boolean; virtual;
    //This function is only called in the main thread. For details see TModule.ActualizeAssets

    procedure OnZeroReferenceCount; override; {depending on MustBeFreedFromMainThread
      it may add Self to gv_AssetsToDelete instead of calling Destroy
    }
    function MustBeFreedFromMainThread: boolean; virtual; //by default returns false


    property Actualized: boolean read _GetActualizedNilSafe; //calls _GetActualized if not nil
    property AssetProperties: TAssetPropertiesSet read f_props;

    function IsNecessary: boolean; virtual; //By default returns true
    //There are two kinds of assets:
    // 1. Low-res and single-instance versions without which further processing
    //  is impossible. These are _necessary_ assets
    // 2. Hi-res mip levels and other upsampled shit being loaded
    //  in the background. These are not necessary and do not cause a pause
    //  for the loading screen.

    { WARNING: If a necessary asset is depending on an asset that is not
        marked as necessary, it results in a DEADLOCK!
        Because optional assets are actualized only after all necessary assets
        had been actualized!
    }

    function IsValidForServer: boolean; virtual; { By default returns false.
        The server does not have OpenGL nor other services initialized,
        so after loading it scrapes all assets it cannot use.
        Exceptions are the file manager, mother streams and such. Any exception
        has to override this method to return true.
    }

    function SortOrder: ffloat; override;

    procedure MarkAsUnclaimed;
    procedure MarkAsTemporary;

    //Copies all relevant properties of the other instance
    // and leaves it zeroed out, to be safely disposed of
    //Used a) on counterparts retrieved from mother
    //  b) on clones lazily updated in the background
    //  see TManagedFbo.Devour as an example: it's complicated!
    //A complex asset may refuse to devour if something doesn't match
    //  -- that's why checking Actualized afterwards is necessary.
    function Devour(c: TGenericAsset): boolean; virtual;

    procedure EraseContainer; virtual;
//procedure Scrape; override;
  end;


  TAssetInnerMatroska = class (TCgeObject)
    //This one is designed for various innards, like MD3 model vertexes and stuff
    //  Is a separate object because Chepersy can selective saving
    //    on a per-object basis, but not on per-field basis.
    //So any asset with heavy data has to be a two-stage matroshka.
  public
    constructor Create; //make self volatile (gets saved in the mother ci stream
    // but not in the session.
  end;


  THandleAsset = class(TGenericAsset)
    //for *anything* that can be described by a simple ptruint or pointer handle
  protected
    _handle: ptruint;  //may be used to contain glUint or pointer type values

    procedure OnGenResourceFailure(ErrorMessage: UnicodeString); virtual;
    //no common meaning.
    //each class tree should either implement its own mechanism,
    //or leave this alone

    function IsValid: boolean; override;
  public
    property Handle: ptruint read _handle;


    procedure RegisterFields; override;

    procedure FreeResource; virtual; abstract;
    procedure GenResource; virtual; abstract;
    procedure OnModuleDestroy; override;
    function Actualize: boolean; override;
    //By default, it just calls GenResource and returns true
    //In complex cases (like FBOs or images) you must
    //  redefine this function, to first check if all child assets
    //  are actualized and all threaded tasks are done. If not, simply return false.
    //This function is only called in the main thread. For details see TModule.ActualizeAssets

    function Devour(c: TGenericAsset): boolean; override;
    procedure EraseContainer; override;
    destructor Destroy; override;
  end;


  TTextureLikeAsset = class(THandleAsset)
  protected
    f_width,
    f_height: integer;
  public
    constructor Create(_width, _height: integer; p: TAssetPropertiesSet = []);
    property Width: integer read f_width;
    property Height: integer read f_height;
    procedure RegisterFields; override;
    function Devour(c: TGenericAsset): boolean; override;
  end;


  function TellInvalidAsset(a: TGenericAsset): UnicodeString; //for error messages

implementation

  uses mo_module, mo_texture, mo_fbo, mo_glsl;

  {$include mo_assets_basic.inc}

  {$include mo_assets_stor.inc}

  function TTextureLikeAsset.Devour(c: TGenericAsset): boolean;
  begin
    f_width:= (c as TTextureLikeAsset).Width;
    f_height:= (c as TTextureLikeAsset).Height;
    Result:= inherited Devour(c);
  end;

  constructor TTextureLikeAsset.Create(_width, _height: integer; p: TAssetPropertiesSet = []);
  begin
    inherited Create(p);
    f_width:= _width;
    f_height:= _height;
    f_width:= ValidateTextureDimension(f_width);
    f_height:= ValidateTextureDimension(f_height);
  end;

  procedure TTextureLikeAsset.RegisterFields;
  begin
    inherited;
    ListFields(['f_width', @f_width, 'f_height', @f_height, typeinfo(integer)]);
  end;

 {
    constructor TImage.Create(width, height: integer; format: glUint);
    begin
      inherited Create;
      pointer(_handle):= Mother^.Image.Create(width, height, format);
      AfterEfCheck;
    end;
    
    constructor TImage.CreateEmpty;
    begin
      inherited Create;
    end;

    function TImage.Clone: TImage;
    begin
      Result:=CImage(Self.ClassType).CreateEmpty;
      pointer(Result._handle):=Mother^.Image.Clone(pointer(_handle));
      AfterEfCheck;
    end;
    
    function TImage.ClonePow2Down(p: integer; warp: boolean): TImage;
    begin
      Result:=Cimage(Self.ClassType).CreateEmpty;
      pointer(Result._handle):=  Mother^.Image.CloneResample(
          pointer(_handle)
        , max(2, Mother^.GAPI.GetHigherPowerOf2(Width div (1 shl p)))
        , max(2, Mother^.GAPI.GetHigherPowerOf2(Height div (1 shl p)))
        , warp);
      AfterEfCheck;
    end;

    function TImage.CloneNextMipmapLevel(warp: boolean): TImage;
    var nextW, nextH: integer;
    begin
      nextW:= Width div 2;
      nextH:= Height div 2;
      if (nextW < 1) or (nextH < 1) then Exit(NIL);

      Result:=Cimage(Self.ClassType).CreateEmpty;
      pointer(Result._handle):=  Mother^.Image.CloneResample(
          pointer(_handle)
        , nextW
        , nextH
        , warp);
      AfterEfCheck;
    end;
}

{
    procedure TImage.ResampleToPow2(wrap: boolean);
    var
      x,y: integer;
    begin
      x:= Mother^.GAPI.GetHigherPowerOf2(Width);
      y:= Mother^.GAPI.GetHigherPowerOf2(Height);
      if (x <> Width) or (y <> Height) then Resample(x, y, wrap);
    end;
}
{
    constructor TImage.Create(FileName: Utf8String);
    begin
      inherited Create;
      Mother^.FileManager.GetFileHash(PAnsiChar(FileName), @Hash);
      AfterEfCheck;
      Module.AddNonActualizedAsset(Self);
      f_filename:=FileName;
    end;


    function TImage.Width: integer;
    begin
      Result:= Mother^.Image.GetWidth(pointer(_handle));
      AfterEfCheck;
    end;


    function TImage.Height: integer;
    begin
      Result:= Mother^.Image.GetHeight(pointer(_handle));
      AfterEfCheck;
    end;


    function TImage.Format: glUint;
    begin
      Result:= Mother^.Image.GetFmt(pointer(_handle));
      AfterEfCheck;
    end;


    function TImage.Pixels: pointer;
    begin
      Result:= Mother^.Image.GetPixels(pointer(_handle));
      AfterEfCheck;
    end;


    procedure TImage.Resize(newwidth, newheight: integer);
    begin
      Mother^.Image.Resize(pointer(_handle), newwidth, newheight);
      AfterEfCheck;
    end;


    procedure TImage.Resample(newwidth, newheight: integer; wrap: boolean);
    begin
      Mother^.Image.Resample(pointer(_handle), newwidth, newheight, wrap);
      AfterEfCheck;
    end;
}
(*
    procedure TImage.OnReload;
    begin
      inherited;
        if f_filename <> '' then begin
          pointer(_handle) := Mother^.Image.Load(PChar(f_filename));
          AfterEfCheck;
          Mother^.FileManager.GetFileHash(PAnsiChar(f_FileName), @Hash.Md5Hash);
          AfterEfCheck;
        end;
    end;
*)
{
    function TImage.Actualize : boolean;
    begin
      (*
        * Loading such asset as image consists of following steps:
          - adding a file manager task to load the image file into memory
          - waiting for it to complete
          - adding a background task of decoding the image file
          - waiting for it to complete
      *)

    end;

    procedure TImage.FreeResource;
    begin
      if _handle > 0 then Mother^.Image.Delete(pointer(_handle));
    end;
}
(*
    procedure TImage.GenResource;
    begin
      { Does nothing. It's here only to console the compiler
         who would otherwise whine about creating instance of a class
         with abstract methods}
    end;

    procedure TImage.RegisterFields;
    begin
      inherited;
      ListFields([
        'f_storedwidth', @f_storedwidth,
        'f_storedheight', @f_storedheight, typeinfo(integer),
        'f_storedfmt', @f_storedfmt, typeinfo(gluint),
        'f_filename', @f_filename, typeinfo(string),
        'Store', @Store,
        'Delayed', @Delayed, typeinfo(boolean)
      ]);
    end;
 *)


//***************************  TTEXTURE **************************************

{ TTexture }

{
    function TTexture.IsResident: boolean;
    var
      isres: GLBoolean = GL_TRUE;
    begin
      if not Assigned(glAreTexturesResident) then Exit(Yes);
      Result:= glAreTexturesResident(1, @_handle, @isres) = GL_TRUE;
    end;

    procedure TTexture.Bind;
    var
      isres: GLBoolean = GL_TRUE;
    begin
      if _handle = 0 then begin
        if Mother^.Debug.StopOnOpenGlErrors then Die(RuEn(
          'Попытка забиндить нулевую текстуру',
          'Attempt to bind texture number zero'));

        glBindTexture(f_target, Mother^.GAPI.StubTexture); //Mightily flashing each frame, too
        exit;
      end;
      if Module.LoadBalancer.Enabled and Assigned(glAreTexturesResident)
        then if glAreTexturesResident(1, @_handle, @isres) <> GL_TRUE
               then Module.LoadBalancer.TextureIsNotInTextureMemory(Self.AssumedWeight());
//addlog('TTexture.Bind(%0, %1)',[f_target, _handle]);
      glBindTexture(f_target, _handle);
      if not Mother^.Debug.StopOnOpenGlErrors then begin
        if glGetError() <> 0 then glBindTexture(f_target, Mother^.GAPI.StubTexture);
      end
      else CheckGlError;
    end;
    
    procedure TTexture.FreeResource;
    begin
      glDeleteTextures(1, @_handle);
      CheckGlError;
    end;

    procedure TTexture.GenResource;
    begin
      glGenTextures(1, @_handle);
      CheckGlError;
    end;
    
    procedure TTexture.RegisterFields;
    begin
      inherited;
      ListFields([
        'f_warped', @f_warped, typeinfo(boolean),
//        'f_target', @f_target, typeinfo(glEnum),
//        'f_minfilter', @f_minfilter,
//        'f_magfilter', @f_magfilter, typeinfo(integer)
      ]);
    end;
 }
 (*   procedure TTexture.Discard;
    begin
      Die(MI_ERROR_PROGRAMMER_NO_BAKA,
         [Self.ClassName +'.Discard() should never be called!']);
    end;
 *)


{ TStatic2dTexture }

{
    constructor TStatic2dTexture.Create(Source: TImage;
                             warped: boolean; minfilter, magfilter: integer);
    begin
      inherited Create;
      f_warped:=warped;
      f_target:=GL_TEXTURE_2D;
      f_minfilter:=minfilter;
      f_magfilter:=magfilter;
      f_source:=Source;
    end;
    
    Constructor TStatic2dTexture.CreateFromFile(FileName: AnsiString;
                             warped: boolean; minfilter, magfilter: integer);
    begin
      Self.Create(TImage.Create(FileName), warped, minfilter, magfilter);
    end;


    function TStatic2dTexture.OptimalFormat: gluint;
    begin
      case f_source.Format of
        GL_BGRA: Result:=GL_RGBA;
        GL_BGR: Result:=GL_RGB;
      else
        Result:=f_source.Format;
      end;
    end;

    procedure TStatic2dTexture.RebuildMips;
    var
      m: integer = 0;
      i1, i2: TImage;
    begin
      i1:= f_source;
      while Assigned(i1) do begin
        glTexImage2D (
          GL_TEXTURE_2D, m, OptimalFormat(), i1.Width, i1.Height,
                       0, i1.Format, GL_UNSIGNED_BYTE, i1.Pixels);

        //only build mip maps if texture is specified with them
        if (f_minfilter <> GL_LINEAR_MIPMAP_LINEAR)
          and (f_minfilter <> GL_LINEAR_MIPMAP_NEAREST)
          and (f_minfilter <> GL_NEAREST_MIPMAP_LINEAR)
          and (f_minfilter <> GL_NEAREST_MIPMAP_NEAREST)
          then Exit;

        i2:= i1.CloneNextMipmapLevel(f_warped);
        if (i1 <> f_source) then i1.Free;
        i1:= i2;
        Inc(m);
      end;
    end;
}

{
    procedure TStatic2dTexture.OnReload;
    var
      rep: glUint;
      vW, vH: glInt;
    begin
      inherited;
      if not WasLost then FreeResource;
      GenResource;
      vW:= ValidateTextureDimension(f_source.Width);
      vH:= ValidateTextureDimension(f_source.Height);
      if (vW <> f_source.Width) or (vH <> f_source.Height)
        then f_source.Resample(vW, vH, f_warped);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, _handle);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, f_minfilter);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, f_magfilter);
      if f_warped
        then rep:= GL_REPEAT
        else rep:= GL_CLAMP;
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, rep);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, rep);

      RebuildMips;
      f_source.Discard;
      Hash.Md5Hash:=Md5String(Md5Print(f_source.hash.Md5Hash));
      WasLost:=false;
    end;
 }
(*
  procedure TStatic2dTexture.RegisterFields;
    begin
      inherited;
      ListFields(['f_source', @f_source, typeinfo(TImage)]);
    end;

    function TStatic2dTexture.AssumedWeight: float;
    begin
      Result:= 0;
      if (_handle = 0) then Exit;
      Result:= ValidateTextureDimension(f_source.Width) * ValidateTextureDimension(f_source.Height) * ( 1 / (1024 * 1024));
    end;

{ TDynamic2dTexture }

    procedure TDynamic2dTexture.RegisterFields;
    begin
      inherited;
      ListFields([
        'f_width', @f_width,
        'f_height', @f_height, typeinfo(integer),
        'f_format', @f_format, 'f_imgformat', @f_imgformat, typeinfo(glUint),
        'f_runhash', @f_runhash, typeinfo(TAssetHash),
        'f_master', @f_master, typeinfo(TGenericAsset)
      ]);
    end;

    Constructor TDynamic2dTexture.Create(Master: TGenericAsset; format: glUint;
                             warped: boolean; minfilter, magfilter: integer;
                             ImageFormat: glUint = GL_RGB;
                             Optional: boolean = false);
    begin
//addlog('TDynamic2dTexture.Create(%0, %1,',[_width, _height]) ;
      inherited Create(Optional);
      f_master:= Master;
      if not f_master.RequestMasterDimensions(f_width, f_height)
      then begin
        f_width:= 256;
        f_height:= 256;
      end;
      f_width:= ValidateTextureDimension(f_width);
      f_height:= ValidateTextureDimension(f_height);

      f_warped:=warped;
      f_format:=format;
      f_imgformat:= ImageFormat; //required for legacy-compatible FBO
      f_target:= GL_TEXTURE_2D;
      f_minfilter:=minfilter;
      f_magfilter:=magfilter;
    end;
                             
{
    procedure TDynamic2dTexture.OnReload;
    begin
      inherited;
      if not WasLost then FreeResource;
      GenResource;
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, _handle);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, f_minfilter);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, f_magfilter);
      if f_warped
        then rep:= GL_REPEAT
        else rep:= GL_CLAMP;
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, rep);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, rep);

      RebuildMips;
      WasLost:=false;
    end;
 }

    function TDynamic2dTexture.OptimalFormat: gluint;
    begin
      result:=f_format;
    end;


    procedure TDynamic2dTexture.BeforeSaving;
    begin
      Mother^.AssetKeeper.GenHash(@hash);
      AfterEfCheck;
      inherited;
    end;


    procedure TDynamic2dTexture.AfterLoading; //cloned from TGenericDynamicAsset
    begin
      inherited;
      if not FActualized then begin
        f_master.RequestMasterDimensions(f_width, f_height);
        f_width:= ValidateTextureDimension(f_width);
        f_height:= ValidateTextureDimension(f_height);
      end;
    end;



    function TDynamic2dTexture.AssumedWeight: float;
    begin
      if (_handle = 0) then Exit(0);
      Result:= f_width * f_height * ( 1 / (1024 * 1024));
    end;


    
{*******************************************}

    constructor THostMemoryChunk.Create(sz: integer);
    begin
      inherited Create;
      f_size:=sz;
      GenResource;
    end;

    procedure THostMemoryChunk.RegisterFields;
    begin
      inherited;
      ListFields(['f_size', @f_size, typeinfo(integer)]);
    end;


    procedure THostMemoryChunk.FreeResource;
    begin
      Mother^.Memory.Free(pointer(_handle));
      AfterEfCheck;
      f_size:=0;
    end;

    procedure THostMemoryChunk.GenResource;
    begin
      if f_size = 0 then Exit;
      pointer(_handle):= Mother^.Memory.Get(f_size);
      AfterEfCheck;
    end;
  *)

{*****************************************************************************}
    
  { Now these two functions are used to cleanup any resources
    left unclaimed after the module is loaded.
    
    To do that properly, we need to know the resource type,
      to tell a vertex buffer from a chainsaw.
    This is implemented using the class information
      that Chepersy does already have. The list of TGenericAsset
      descendants and their class indices is stored in the core module
      and used later to tell which class is which.
      
    We cannot store the metaclasses themselves (which would be a lot simpler)
      because their pointers move from compile to compile
      and may even move if the module Dll base address changes between loadings.
  }
  
 { procedure StoreResourceClassesList;
  var
    //s: pointer;
    p: pointer;
    ms: TMotherStream;
    i: integer;
    LL: TArrayOfExportClassDesc;
  begin
    p:= Mother^.Core.GetAssetStorageMemoryStream();
    AfterEfCheck;
    if Assigned(p) then begin
      Mother^.Core.SetAssetStorageMemoryStream(NIL);
      AfterEfCheck;
      Mother^.Stream.Free(p);
      AfterEfCheck;
    end;
    ms:= TMotherStream.CreateAsMemoryStream(false);
    //s:= Mother^.Memory.CreateStream();

    Mother^.Core.SetAssetStorageMemoryStream(ms.Handle);//s);
    AfterEfCheck;

    LL:= chepersy.ExportClassesList(TGenericAsset);
    //L1:=ExportChepersyClassList; //full chepersy classes list, <name> - <class>
//writeln('length(LL)=',length(LL));
   (* TExportClassDesc = record
      Name: AnsiString;
      Metaclass: CChepersyObject;
      Index: integer;
    end;
    TArrayOfExportClassDesc = array of TExportClassDesc;
    *)

    ms.WriteDword(Length(LL));
    //Mother^.Memory.WriteIntToStream(s, Length(LL));
    //AfterEfCheck;
    For i:=0 to High(LL) do begin
      ms.WriteAnsiString(LL[i].Name);
//      Mother^.Memory.WriteIntToStream(s, Length(LL[i].Name));
//      AfterEfCheck;
//      Mother^.Memory.WriteToStream(s, @LL[i].Name[1], Length(LL[i].Name));
//      AfterEfCheck;
      ms.WriteDword(LL[i].Index);
//      Mother^.Memory.WriteIntToStream(s, LL[i].Index);
//      AfterEfCheck;
    end;
  end;
  

  procedure DeleteUnclaimedAssets;
  var
    i, j, m, n: integer;
    r: PAssetRecord;
    h: pointer;
    c: CGenericAsset;
    o: TGenericAsset;
    cn: AnsIString;
    ms: TMotherStream;
    NumOldClasses : integer;
    ocNames: array of AnsiString;
    ocIndexes: array of Integer;
    ociToClass: array of CGenericAsset;
    ociToName: array of AnsiString;
    LL: TArrayOfExportClassDesc;
    q: qword;
    t: double;
    total: integer = 0;
  begin
//writeln;
//writeln('CEI=',Mother^.Module.CurrentlyExecutingInd);
    if Mother^.Debug.Verbose then AddLog('Deleting unclaimed assets...');
    Mother^.Timer.UsecDelta(@q);
    h:= Mother^.Core.GetAssetStorageMemoryStream();
//addlog('s = %0',[s]);
    AfterEfCheck;

    if not Assigned(h) then Exit; //nothing to do, the module was never saved

    ms:= TMotherStream.CreateFromHandle(h, false);

    r:= Mother^.AssetKeeper.GetFirst();
    AfterEfCheck;

    if not Assigned(r) then Exit; //there are no lost resources to cleanup.

    LL:= chepersy.ExportClassesList(TGenericAsset);

    //Load the data left by the previous module instance
    ms.Position:= 0;
//    Mother^.Memory.SeekStream(s, 0);
//    AfterEfCheck;
    NumOldClasses:= ms.ReadDword;
    //NumOldClasses:= Mother^.Memory.ReadIntFromStream(s);
//    AfterEfCheck;
    SetLength(ocNames, NumOldClasses);
    SetLength(ocIndexes, NumOldClasses);
//writeln('NumOldClasses=',NumOldClasses);
    For i:=0 to NumOldClasses - 1 do begin
      ocNames[i]:= ms.ReadAnsiString;
//      j:=Mother^.Memory.ReadIntFromStream(s);
//      AfterEfCheck;
//      SetLength(ocNames[i], j);
//      Mother^.Memory.ReadFromStream(s, @ocNames[i][1], j);
//      AfterEfCheck;
      ocIndexes[i]:= ms.ReadDword;
//      ocIndexes[i]:=Mother^.Memory.ReadIntFromStream(s);
//      AfterEfCheck;
    end;
    ms.Free;

    SetLength(ociToClass, 0);
    For i:=0 to NumOldClasses - 1 do begin
      For j:= 0 to Length(LL) - 1 do begin
        if (LL[j].Name = ocNames[i]) then begin
          if ocIndexes[i] > High(ociToClass) then begin
            m:= Length(ociToClass);
            SetLength(ociToClass, ocIndexes[i] + 1);
            SetLength(ociToName, ocIndexes[i] + 1);
            for n:= m to High(ociToClass) - 1 do ociToClass[n]:= nil;
            for n:= m to High(ociToName) do ociToName[n]:= '<n/a>';
          end;
//writeln('i=',i,' ocIndexes[i]=',ocIndexes[i],' high(ociToClass)=',high(ociToClass));
          ociToClass[ocIndexes[i]]:= CGenericAsset(LL[j].Metaclass);
          break;
        end;
      end;
    end;

    //Now delete the unclaimed resources
    while Assigned(r) do begin
      if r^.cpsInd <= High(ociToClass) then begin
        c:= ociToClass[r^.cpsInd];
        cn:= ociToName[r^.cpsInd];
      end else begin
        c:= nil;
        //whoops, no such class in the new module instance.
        //Extracting the name from the old chepersy data
        cn:= '<n/a:' + IntToStr(r^.cpsInd) + '>';
        for i:= 0 to High(ocNames) do
          if ocIndexes[i] = r^.cpsInd then begin
            cn:= ocNames[i];
            break;
          end;
      end;
      if Assigned(c) then begin
        o:= c.CreateToDestroy(r^.Handle);
        o.Free;//TechnicalDestroy;
        inc(Total);
      end
      else
        Die(RuEn(
          'Не удалось освободить ничейный ресурс:'#10#13'  не найден класс, создавший его!'#10#13'  (индекс %0, имя %1, хэндл %2, хэш %3)',
          'Unable to free the unclaimed resource:'#10#13'  cannot find the class that created it!'#10#13'  (index %0, name %1, handle %2, hash %3)'),
          [r^.cpsInd, cn, pointer(r^.Handle), HashToString(@r^.Hash)]);
      r:=Mother^.AssetKeeper.GetNext();
      AfterEfCheck;
    end;
    Mother^.AssetKeeper.ClearList;
    AfterEfCheck;
    t:= Mother^.Timer.UsecDelta(@q);
    if Mother^.Debug.Verbose then AddLog('Deleted %0 assets in %1 microseconds (%2 nanoseconds per asset)', [total, round(t), round(t * 1000 / max(1, total))]);

  end;
 }



  { TAbstractVertexBuffer }
 (*
  constructor TAbstractVertexBuffer.Create(sz: integer);
  begin
    f_sz:= sz;
    GenResource;
  end;

  procedure TAbstractVertexBuffer.RegisterFields;
  begin
    inherited;
    ListFields(['f_sz', @f_sz, typeinfo(integer)]);
  end;

  { THwVertexBuffer }

  procedure THwVertexBuffer.FreeResource;
  begin

  end;

  procedure THwVertexBuffer.GenResource;
  begin

  end;

  procedure THwVertexBuffer.RegisterFields;
  begin
    inherited;
    
    
  end;

  function THwVertexBuffer.Buffer(): PVertexArray;
  begin

  end;

{ TSoftwareVertexBuffer }

  procedure TSoftwareVertexBuffer.FreeResource;
  begin
    SetLength(va, 0);
  end;

  procedure TSoftwareVertexBuffer.GenResource;
  begin
    SetLength(va, f_sz * (sizeof(TVertex) div sizeof(GLfloat)));
  end;

  procedure TSoftwareVertexBuffer.RegisterFields;
  begin
    RegType(typeinfo(TGlFloatArray), typeinfo(glFloat));
    inherited;
    ListFields(['va', @va, typeinfo(TGlFloatArray)]);
  end;

  function TSoftwareVertexBuffer.Buffer(): PVertexArray;
  begin
    Result:=@va[0];
  end;

{ TMemoryMappedFile }

  constructor TMemoryMappedFile.Create(FileName: AnsiString);
  var
    ufn: ansistring;
  begin
    if not Module.FileManager.Exists(FileName)
      then Die(RuEn(
        'Файл не существует: "%0"',
        'File does not exist: "%0"'
        ),
        [FileName]
      );
    inherited Create(Module.FileManager.GetSize(FileName));
    f_filename:= FileName;
    AfterEfCheck;
  end;

  procedure TMemoryMappedFile.RegisterFields;
  begin
    RegType(typeinfo(TDateTime));
    inherited;
    ListFields([
      'f_filename', @f_filename, typeinfo(TFileNameString),
      'file_modify_time', @file_modify_time, typeinfo(TDateTime)
    ]);
  end;

{
  procedure TMemoryMappedFile.OnReload;
  var
    ufn: ansistring;
    fs: TFileStream;
  begin
    inherited;
      if f_filename <> '' then begin
        ufn:=PCharToString(Mother^.FileManager.UnmangleFileName(PChar(f_filename)));
        AfterEfCheck;
        try
          fs:=TFileStream.Create(ufn, fmOpenRead);
        except
          Die(MI_CGEFILE_ERROR_READ, [ufn]);
        end;
        f_mem:= THostMemoryChunk.Create(fs.Size);
        fs.Read(pointer(f_mem._handle)^, f_mem.Size);
        pointer(_handle) := Mother^.Image.Load(PChar(f_filename));
        AfterEfCheck;
        Mother^.FileManager.GetFileHash(PAnsiChar(f_FileName), @Hash.Md5Hash);
        AfterEfCheck;
      end;
  end;
}
  procedure TMemoryMappedFile.FreeResource;
  begin
//    inherited FreeResource;
  end;

  procedure TMemoryMappedFile.GenResource;
  begin
//    inherited GenResource;
  end;

{
  procedure TMemoryMappedFile.Discard;
  begin
    inherited Discard;
  end;
 }

*)
  procedure RegisterAssetClasses;
  begin
    RegClass(TGenericAsset);
    regClass(THandleAsset);
    RegClass(TTextureLikeAsset);
    RegClass(TAssetContainer);
//    RegClass(TImage);
//	  RegClass(TStatic2dTexture);
//	  RegClass(TDynamic2dTexture);
//	  RegClass(THostMemoryChunk);
//	  RegClass(TMemoryMappedFile);
//	  RegClass(THwVertexBuffer);
//	  RegClass(TSoftwareVertexBuffer);
    mo_fbo.RegClasses;
    mo_glsl.RegClasses;
    mo_texture.RegClasses;
  end;

  function TellInvalidAsset(a: TGenericAsset): UnicodeString; //for error messages
  var ap: TAssetPropertiesSet;
  begin
    if not Assigned(A) then Exit('NIL');
    try
      if not (a is TGenericAsset)
        then Exit(RuEn(
          'Не является ресурсом: ',
          'Is not an asset: '
          ) + a.ClassName);
    except
      Exit(RuEn(
        'Указатель, не являющийся объектом ',
        'Invalid object '
        ) + IntToHex(ptruint(pointer(a)), 2 * sizeof(pointer)) + 'h '
        + (ExceptObject as Exception).Message);
    end;
//    if a.Scraped
//      then Exit(a.ClassName + RuEn(', удалён',', scraped'));
    if (a is THandleAsset) and ((a as THandleAsset).Handle = 0)
      then Exit(a.ClassName + RuEn(', нулевой хэндл',', zero handle'));
    if IsHashZero(a.Hash)
      then Exit(a.ClassName + RuEn(', пустой хэш',', empty hash'));
    if not a.Actualized
      then Exit(a.ClassName + RuEn(', не актуализован',', not actualized'));
    if not a.IsValid
      then begin
        ap:= a.AssetProperties;
        Exit(a.ClassName + ', IsValid = false, ' + RuEn('свойства: ','properties: ')
          + SetToStr(@ap, typeinfo(TAssetPropertiesEnum)));

      end;

    Result:= a.ClassName;
  end;

  class procedure TAssetManager.Finalize;
  begin
    if Assigned(ToDeleteInMainThreadCS) then begin
      while Assigned(ToDeleteInMainThread) or Assigned(Retrieved)
        do Pulse;

      ToDeleteInMainThreadCS.Free;
    end;
  end;

  class procedure TAssetManager.Initialize;
  begin
    ToDeleteInMainThreadCS:= TCriticalSection.Create;
  end;

  class procedure TAssetManager.Pulse;
  var
    q: qword;
    i, t, c : integer;
  begin
    if Assigned(ToDeleteInMainThread) then begin
      EnterWv(TAssetManager.ToDeleteInMainThreadCS);
      t:= Length(ToDeleteInMainThread.List);
      if (Mother^.Debug.Verbose) then begin
        AddLog('  Delayed deleting of %0 assets in the main thread...', [t]);
        Mother^.Timer.UsecDelta(@q); // Not free anymore. On an ARM platform it can take up to a microsecond.
      end;

      for i:= 0 to High(ToDeleteInMainThread.List) do
        ToDeleteInMainThread.List[i]:= nil; //will deadlock the CS if this is not the main thread

      if (Mother^.Debug.Verbose) then AddLog('  Deleted %0 delayed assets in %1µs',[t, round(Mother^.Timer.UsecDelta(@q))]);
      ToDeleteInMainThread:= nil;
      ToDeleteInMainThreadCS.Leave;
    end;


    //For *now*, just delete them at once, assuming that the cost of
    //  freeing whatever is negligible.
    //***TODO: benchmark and optimize the shit out of this if needed.

    if Assigned(Retrieved) then begin

      t:= Length(Retrieved.List);

      if (Mother^.Debug.Verbose) then begin
        AddLog('  Deleting %0 unclaimed assets...', [t]);
        Mother^.Timer.UsecDelta(@q); // Not free anymore. On an ARM platform it can take up to a microsecond.
      end;

      c:=0;
      for i:= 0 to High(Retrieved.List) do
        if Assigned(Retrieved.List[i]) then with Retrieved.List[i] do begin
          //if not Exists then continue;
          if Actualized
            then begin
              if (Mother^.Debug.Verbose) then AddLog('    ..deleting %0', [string(ClassName)]);
              c+= 1;
            end;
          Retrieved.List[i]:= nil;
        end;

      if (Mother^.Debug.Verbose) then AddLog('  Deleted %0 unclaimed assets in %1µs',[c, round(Mother^.Timer.UsecDelta(@q))]);
      Retrieved:= nil;
    end;
  end;

end.

