{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


unit mo_glsl;

{$include mo_globaldefs.h}

interface

uses
  sysutils, Classes, math, md5,  chepersy, mo_hub, typinfo, mo_classes,
    mo_gmathbase, mo_assets;

procedure CheckGLSLError(Handle: GLhandleARB; Param: GLenum; Explanation: UnicodeString; source: AnsiString = '');

procedure RegClasses;

type
  TShaderPropertiesEnum = (
    sp_free_shaders_on_destroy
  );
  TShaderPropertiesSet = set of TShaderPropertiesEnum;

  {

***TODO: Investigate if that was because of unrelated coding horrors!
  There are things man was not meant to keep between module instances.
    For example, GLSL shaders are guaranteed to crash (does the driver
      link them to our DLL somehow?)

    So these classes are not inherited from TGenericDynamicAsset anymore

    The class instances get saved and loaded, but the shaders are deleted
      at each module unload and recreated at each session load
      (because you *cannot* load the session without reloading the module DLL)
  }

  TAbstractShader = class(THandleAsset)
  private
    f_source: AnsiString;
    f_sprops: TShaderPropertiesSet;
    function MyType: GLenum; virtual; abstract;
  public
    constructor Create(source: AnsiString; sprops: TShaderPropertiesSet = []; aprops: TAssetPropertiesSet = []);
    procedure RegisterFields; override;
    procedure GenResource; override;
    procedure FreeResource; override;
    function Which(): UnicodeString; virtual; abstract;
    function MustBeFreedFromMainThread: boolean; override;
    property Handle: ptruint read _handle;
 //   procedure OnModuleDestroy; override;
  end;

  TVertexShader = class(TAbstractShader)
  private
    function MyType: GLenum; override;
  public
    function Which(): UnicodeString; override;
  end;

  TFragmentShader = class(TAbstractShader)
  private
    function MyType: GLenum; override;
  public
    function Which(): UnicodeString; override;
  end;

  TGlslProgram = class(THandleAsset)
    //does NOT free shaders on destruction
  protected
    f_vsh: TVertexShader;
    f_fsh: TFragmentShader;
    f_sprops: TShaderPropertiesSet;
    procedure CheckOrRecreateShaders;
    function IsHandleValid: boolean;
  public
    constructor Create(_vsh: TVertexShader; _fsh: TFragmentShader;
      sprops: TShaderPropertiesSet = []; aprops: TAssetPropertiesSet = []); overload;

    constructor Create(
      sprops: TShaderPropertiesSet = []; aprops: TAssetPropertiesSet = []); overload; //see GenerateShaderSource

    procedure RegisterFields; override;

    procedure GenResource; override;
    procedure FreeResource; override;
    procedure DeleteShaders;
    property VertexShader: TVertexShader read f_vsh;
    property FragmentShader: TFragmentShader read f_fsh;
    procedure Use;
 //   procedure OnModuleDestroy; override;

    function Actualize: boolean; override;
    procedure AfterLoading; override;

    //gets called after loading if asset is not actualized
    //In case shader source depends of QF, videocard quirks or other global settings
    //An empty string means "keep current shader", otherwise shader gets re-created
    //by default returns two empty strings if shaders are present,
    //or standard FFP emulation if shaders are not assigned
    procedure GenerateShaderSource(var v, f: AnsiString); virtual;
    function MustBeFreedFromMainThread: boolean; override;

  end;

implementation //**************************************************************
  uses mo_module;

  constructor TAbstractShader.Create(source: AnsiString;
    sprops: TShaderPropertiesSet = []; aprops: TAssetPropertiesSet = []);
  begin
    inherited Create(aprops);
    f_source:= source + #0;
    f_sprops:= sprops;
  end;

  procedure TAbstractShader.RegisterFields;
  begin
    RegType(typeinfo(TShaderPropertiesEnum));
    RegType(typeinfo(TShaderPropertiesSet));//, typeinfo(TShaderPropertiesEnum));
    inherited;
    ListFields([
      'source code', @f_source, typeinfo(AnsiString),
      'shader properties', @f_sprops, typeinfo(TShaderPropertiesSet)
    ]);
  end;

{  procedure TAbstractShader.OnModuleDestroy;
  begin
    Free;
  end;
}
  function TAbstractShader.MustBeFreedFromMainThread: boolean; begin Result:= true end;

  procedure TAbstractShader.GenResource;
  var
    len: GLuint;
    q: qword;
  begin
    try
      Mother^.Timer.UsecDelta(@q);
      _handle:= glCreateShader(MyType());
      CheckGlError;
      len:= Length(f_source);
      glShaderSource(_handle,  1, @(pointer(f_source)), NIL);//@len);
      CheckGlError;
      glCompileShader(_handle);
      CheckGLSLError(_handle, GL_OBJECT_COMPILE_STATUS_ARB,
        PervertedFormat(RuEn(
          'компиляции %0 шейдера',
          'compiling %0 shader'), [Which()]
        ),
        f_source
      );
      if Mother^.Debug.Verbose then AddLog(
        '    Compiled a %0 shader in %1 microseconds',[Which(), round(Mother^.Timer.UsecDelta(@q))]);
    except
      if _handle <> 0 then glDeleteShader(_handle);
      CheckGlError;
      _handle:= 0;
      Die(RuEn('Не удалось создать шейдер','Failed to create shader'));
    end;
    UpdateHash;
  end;

  procedure TAbstractShader.FreeResource;
  begin
    if _handle = 0 then Exit;
    glDeleteShader(_handle);
    CheckGlError;
  end;

  function TVertexShader.Which(): UnicodeString;
  begin
    Result:= RuEn('вершинного', 'vertex');
  end;


  function TFragmentShader.Which(): UnicodeString;
  begin
    Result:= RuEn('фрагментного', 'fragment');
  end;


  function TVertexShader.MyType: GLenum;
  begin
    Result:= GL_VERTEX_SHADER_ARB;
  end;

  function TFragmentShader.MyType: GLenum;
  begin
    Result:= GL_FRAGMENT_SHADER_ARB;
  end;



//    f_vsh: TVertexShader;
//    f_fsh: TFragmentShader;

  constructor TGlslProgram.Create(_vsh: TVertexShader; _fsh: TFragmentShader;
      sprops: TShaderPropertiesSet = []; aprops: TAssetPropertiesSet = []);
  begin
    inherited Create(aprops);
    f_vsh:= _vsh;
    f_fsh:= _fsh;
    f_sprops:= sprops;
  end;

  constructor TGlslProgram.Create(
      sprops: TShaderPropertiesSet = []; aprops: TAssetPropertiesSet = []);
  begin
    inherited Create(aprops);
    f_sprops:= sprops;
  end;

  procedure TGlslProgram.RegisterFields;
  begin
    inherited;
    ListFields([
      'vertex shader', @f_vsh, typeinfo(TVertexShader),
      'fragment shader', @f_fsh, typeinfo(TFragmentShader),
      'program properties', @f_sprops, typeinfo(TShaderPropertiesSet)
    ])
  end;

  function TGlslProgram.MustBeFreedFromMainThread: boolean; begin Result:= true end;

  function TGlslProgram.Actualize: boolean;
  begin
    if not Assigned(f_fsh) or not Assigned(f_vsh)
      then  CheckOrRecreateShaders;

    if not f_fsh.Actualized or not f_vsh.Actualized
      then Exit(false); //satisfy dependencies

    Result:= inherited;
  end;

  procedure TGlslProgram.AfterLoading;
  begin
    inherited;
    if not Actualized then CheckOrRecreateShaders;
  end;

  procedure TGlslProgram.CheckOrRecreateShaders;
  var
    f, v: AnsiString;
  begin
    GenerateShaderSource(v, f);
    if v = '' then begin
      if not Assigned(f_vsh) then Die(RuEn(
        'Внутренняя ошибка в %0.GenerateShaderSource(),'#10#13'  не указан исходный код вершинного шейдера.',
        'Internal error in %0.GenerateShaderSource(),'#10#13'  no source code provided for vertex shader.'),
        [string(Self.ClassName)]
      );
    end
    else f_vsh:= TVertexShader.Create(v, f_sprops, f_props);

    if f = '' then begin
      if not Assigned(f_fsh) then Die(RuEn(
        'Внутренняя ошибка в %0.GenerateShaderSource(),'#10#13'  не указан исходный код фрагментного шейдера.',
        'Internal error in %0.GenerateShaderSource(),'#10#13'  no source code provided for fragment shader.'),
        [string(Self.ClassName)]
      );
    end
    else f_fsh:= TFragmentShader.Create(f, f_sprops, f_props);
  end;

  procedure TGlslProgram.GenerateShaderSource(var v, f: AnsiString);
  begin
    if Assigned(f_vsh)
      then v:= ''
      else v:=
       'void main() {'
         + 'gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;'
         + 'gl_FrontColor = gl_Color;'
         + 'gl_TexCoord[0] = gl_MultiTexCoord0;'
       +'}';
    if Assigned(f_fsh)
      then f:= ''
      else f:=
       'void main(void) {'
         + 'gl_FragColor = gl_Color;'
       + '}';
    if Mother^.GAPI.MaxVShTextureUnits < 1 then Die(RuEn(
      'Невозможно использовать данный шейдер:'#10#13'  видеокарта не поддерживает выборки из текстуры в вершинном шейдере',
      'Impossible to use this shader:'#10#13'  video card does not support texture lookups from vertex shaders'
    ));
  end;


  procedure TGlslProgram.GenResource;
  var
    sta: GLint;
    var q: qword;
  begin

    try
      _handle:= glCreateProgram();
       CheckGlError;

      if not Assigned(f_vsh) then Die(RuEn(
        'TGlslProgram.GenResource: не хватает экземпляра класса вершинного шейдера',
        'TGlslProgram.GenResource: missing vertex shader class instance.'));
      if not Assigned(f_fsh) then Die(RuEn(
        'TGlslProgram.GenResource: не хватает экземпляра класса фрагментного шейдера',
        'TGlslProgram.GenResource: missing fragment shader class instance.'));

      if f_vsh.Handle = 0 then f_vsh.GenResource;
      if f_fsh.Handle = 0 then f_fsh.GenResource;

      Mother^.Timer.UsecDelta(@q);

      glAttachShader(_handle, f_vsh.Handle); //alias for glAttachObjectARB
      CheckGlError;

      glAttachShader(_handle, f_fsh.Handle);
      CheckGlError;

      glLinkProgram(_handle);
      CheckGlError;

      CheckGLSLError(_handle, GL_OBJECT_LINK_STATUS_ARB, RuEn(
        'сборки шейдеров в программу GLSL',
        'linking GLSL program'
      ));

      if Mother^.Debug.Verbose then AddLog(
        '    Linked a GLSL program in %0 microseconds, handle=%1',[round(Mother^.Timer.UsecDelta(@q)), pointer(_handle)]);

    except
      if _handle <> 0 then try
        glDeleteProgram(_handle);
        CheckGlError;
      except
        Die(RuEn('Не удалось удалить частично созданную программу','Failed to delete the partly created program'));
      end;
      _handle:= 0;
      Die(RuEn('Не удалось создать GLSL программу','Failed to create GLSL program'));
    end;
    UpdateHash;
  end;

  procedure TGlslProgram.FreeResource;
  begin
    if _handle = 0 then Exit;
    if Mother^.Debug.Verbose then AddLog(
      '  %0.FreeResource: Handle was %1',[string(self.classname), pointer(_handle)]);
//addlog(ExplainCallStack(1));
    if IsHandleValid then begin
      glDeleteProgram(_handle);
      CheckGlError;
    end;
    _handle:= 0;
  end;

  function TGlSlProgram.IsHandleValid: boolean;
  var p: glInt;
  begin
    if _handle = 0 then Exit(false);
    glGetProgramiv(_handle, GL_DELETE_STATUS, @p);
    if glGetError() <> 0 then Exit(false);
    Result:= p <> GL_TRUE;
  end;

  procedure TGlslProgram.DeleteShaders;
  begin
    if Assigned(f_vsh) then begin
      f_vsh.Free;
      pointer(f_vsh):= NIL;
    end;
    if Assigned(f_fsh) then begin
      f_fsh.Free;
      pointer(f_fsh):= NIL;
    end;
  end;


  procedure TGlslProgram.Use;
  begin

//addlog('_handle=%0',[pointer(_handle)]);
//CheckGlError;


    glUseProgram(0);
    glUseProgram(_handle);
    CheckGlError;
  end;


  procedure CheckGLSLError(Handle: GLhandleARB; Param: GLenum; Explanation: UnicodeString; source: AnsiString);
  var
    slen, res: GLint;
    s: AnsiString;
  begin
    glGetObjectParameterivARB(Handle, Param, @res);

    if (res <> GL_TRUE) or true then begin //Mother^.Debug.Verbose then begin
      SetLength(s, 100000);
      glGetInfoLogARB(Handle, Length(s), @slen, PGLcharARB(@s[1]));
      SetLength(s, slen);
    end;

    if res <> GL_TRUE then begin
      Mother^.State.Fatality:= Yes; //cuz opengl just hangs if we try to continue!
      if source <> ''
        then Die(RuEn(
          'Ошибка %0: %1'#10#13'Исходный код шейдера: %2',
          'Failed %0: %1'#10#13'Shader source: %2'), [Explanation, s, source])
        else Die(RuEn(
          'Ошибка %0: %1',
          'Failed %0: %1'), [Explanation, s]);
    end;

    if Mother^.Debug.Verbose then
      if source <> ''
        then AddLog('    Log of %0 : %1 ; Shader source: %2',[Explanation, s, source])
        else AddLog('    Log of %0 : %1',[Explanation, s]);
  end;

  procedure RegClasses;
  begin
    RegClass(TVertexShader);
    RegClass(TFragmentShader);
    RegClass(TGlslProgram);
  end;

end.

