{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{
    Note: at first, images were assets. But then I realized that they were
    too volatile for that, too temporary. Unnneeded after they are loaded
    into that texture or whatever.

    So I made images into threaded tasks.
    They live while the  owner asset actualizes. Then they die.

    On a side note, I should work on moving from Vampyre Imaging library
    to SDL2 image or something, and incorporate imaging into the framework.
    Because, as awesome as Vampyre is, I don't see it actually update since 2009.

    Note 2: this operates with opengl texture format constants, but the
    imaging library uses its own, different format representation.
    Format designations are converted on the fly.
}

unit mo_image;

{$include mo_globaldefs.h}

interface

uses
  sysutils, Classes, math, md5,  chepersy, mo_hub, typinfo, mo_classes,
    mo_gmathbase, mo_threads;


type

  TImageDecodeTask = class(TAbstractThreadedTask)
  private
    f_pinbuff, f_poutbuff: pointer;
    f_insize, f_outsize: integer;
    f_imghandle: pointer; //in mother, this resolves to a class instance
    f_width, f_height: integer;
    f_format: GLuint;
  protected
    function Perform: boolean; override;
  public
    property Width: integer read f_width;
    property Height: integer read f_height;
    property Format: GLuint read f_format;
    property ImageHandle: pointer read f_imghandle;
    property PPixels: pointer read f_poutbuff;
    property PixelsSize: integer read f_outsize;
    function GetType: TThreadedTaskType; override;

    constructor Create(pbuf: pointer; bsize: longint);
    destructor Destroy; override;
  end;

  TImageResampleTask = class(TImageDecodeTask)
  private
    f_source: TImageDecodeTask;
    f_resample_width,
    f_resample_height : integer;
    f_resample_format: GLuint;
    f_cycled: boolean;
  protected
    function Perform: boolean; override;
  public
    property Source: TImageDecodeTask read f_source;

    constructor Create(src: TImageDecodeTask; newx, newy: integer;
      newformat: GLuint; cycled: boolean = false);
  end;

  TImageSaveTask = class(TAbstractThreadedTask)
    //This one breaks the "all file operations in one thread" rule
    //  for simplicity's sake.
    //  Intended for making screenshots, saving atlas caches & stuff

    //! Deletes image in destructor, so if you need to keep it
    //  make sure to call ZeroHandle() before erasing your reference to this task!
  protected
    f_imghandle: pointer;
    f_file_name: TFileNameString;
    function Perform: boolean; override;
  public
    constructor Create(Imagehandle: pointer; FileName: TFileNameString);
    function GetType: TThreadedTaskType; override;
    procedure ZeroHandle;
    destructor Destroy; override;
  end;

  TImageDeleteTask = class(TAbstractThreadedTask)
  protected
    f_imghandle: pointer;
    function Perform: boolean; override;
  public
    constructor Create(Imagehandle: pointer);
    function GetType: TThreadedTaskType; override;
  end;

  TScreenshotMakeTask = class(TAbstractThreadedTask)
  protected
    function Perform: boolean; override;
  public
    constructor Create;
    function GetType: TThreadedTaskType; override;
  end;

  //The image cache: there is a clever trick of storing all checksums in one INI file,
  //  there should be no collisions as there is already "one image task at a time"
  //  interlocking in place to protect the not thread-safe Vampyre.
  //These tasks, again, break the "all file operations in one thread" rule
{
  TImageCacheReadTask = class(TImageDecodeTask)
  protected
    f_imghandle: pointer;
    f_checksum: RawByteString;
    f_file_name: TFilenameString;
    function Perform: boolean; override;
    //Failure is indicated by PPixels being NIL
    //Immediately deletes file on disk if checksum does not match
  public
    constructor Create(fn: TFilenameString; Checksum: RawByteString);
    function GetType: TThreadedTaskType; override;
  end;

  TImageCacheUpdateTask = class(TAbstractThreadedTask)
  protected
    f_checksum: RawByteString;
    f_file_name: TFilenameString;
    function Perform: boolean; override;
  public
    constructor Create(Imagehandle: pointer; FileName: TFilenameString;
      Checksum: RawByteString; DeleteImage: boolean = false);
    function GetType: TThreadedTaskType; override;
  end;
}

{

***TODO
  TImageEncodeTask = class(TAbstractThreadedTask)
  private

  public
  end;
}



implementation

  function TImageSaveTask.Perform: boolean;
  var
    q: qword;
    n: TFileNameString;
  begin
    Result:= true; //in case of failure, consider task completed as well

    if Mother^.Debug.Verbose then AddLog('%0.Perform',[RawByteString(Self.ClassName)]);
    if Mother^.Debug.Verbose then AddLog('  file name is %0', [f_file_name]);
    Mother^.Timer.UsecDelta(@q);


    n:= UnmangleFileName(f_file_name);
    MakeSurePathExists(ExtractFilePath(n));

    if Mother^.Debug.Verbose then AddLog('  saving %0', [n]);
    Mother^.Image.Save(PFileNameChar(n), f_imghandle);
    AfterEfCheck;

    if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecDelta(@q))]);
  end;

  function TScreenshotMakeTask.Perform: boolean;
  var
    q: qword;
    Fname, ext: TFileNameString;
    img: pointer;
    src, dst: Pbyte;
    x, y: integer;
    size: integer;
  begin
    Result:= true; //in case of failure, consider task completed as well
    if Mother^.Debug.Verbose then AddLog('%0.Perform',[RawByteString(Self.ClassName)]);
    if not Assigned(Mother^.Display.Screenshot.pdata)
      or (Mother^.Display.Screenshot.Request <> -2)
    then begin
      if Mother^.Debug.Verbose then AddLog('  aborted!');
      Exit;
    end;

    Mother^.Timer.UsecDelta(@q);
    {
    jpeg no work. Vampyre have internal errors.

    ext:= LowerCase(Utf8String(Mother^.Config.GetString('screenshots', 'file_type')));
    if (ext <> '.jpg') and (ext <> '.jpeg') and (ext <> '.png') then begin
      AddLog(RuEn(
        'Недопустимый вариант "%0" параметра file_type в секции [screenshots] файла настроек. Использую ".jpg" вместо этого.',
        'Invalid value "%0" in the file_type parameter of the [screenshots] config file section. Using ".jpg" instead.'),
        [ext]
      );
      ext:= '.jpg';
    end;
    }
    ext:= '.png';

    if Mother^.Debug.Verbose
      then with Mother^.Display.Screenshot do
        AddLog('  creating image %0x%1 %2...',[Width, Height, GlTextureFormatToString(Format)]);
    with Mother^.Display.Screenshot do
      img:= Mother^.Image.Create(Width, Height, Format);
    AfterEfCheck;
    Dst:= Mother^.Image.GetPixels(img);
    AfterEfCheck;
    size:= Mother^.Image.GetBytes(img);
    AfterEfCheck;
    Src:= Mother^.Display.Screenshot.pdata;

    //and nown turn this upside down and inside out:
    with Mother^.Display.Screenshot do
      for y:= 0 to Height - 1 do
        for x:= 0 to Width - 1 do begin
          (Dst + (x + y * Width) * 3 + 0)^:= (Src + (x + (Height - 1 - y) * Width) * 3 + 2)^;
          (Dst + (x + y * Width) * 3 + 1)^:= (Src + (x + (Height - 1 - y) * Width) * 3 + 1)^;
          (Dst + (x + y * Width) * 3 + 2)^:= (Src + (x + (Height - 1 - y) * Width) * 3 + 0)^;
        end;

    Fname:= '*H' + DS + 'screen-capture' + DS
      + RawByteString(Mother^.Module.PureName) + '_'
      + FormatDateTime('_yyyy_mm_dd_hh_nn_ss_zzz', Now()) + ext;
    Fname:= UnmangleFileName(Fname);
    MakeSurePathExists(ExtractFilePath(Fname));

    if Mother^.Debug.Verbose then AddLog('  saving %0', [Fname]);
    Mother^.Image.Save(PFileNameChar(FName), img);
    AfterEfCheck;

    Mother^.Image.Delete(img);
    Mother^.Display.Screenshot.Request:= -1; //command mother to free its image buffer

    if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecDelta(@q))]);
    AddLog(RuEn(
      'Сохранён снимок экрана,'#10#13'  %0',
      'Screenshot saved,'#10#13'  %0'
      ),[Fname]);
  end;


  constructor TImageSaveTask.Create(Imagehandle: pointer; FileName: TFileNameString);
  begin
    inherited Create;
    f_imghandle:= ImageHandle;
    f_file_name:= FileName;
    Reschedule;
  end;

  procedure TImageSavetask.ZeroHandle;
  begin
    Assert(Self.Completed, 'Cannot zero the handle until the task is completed!');
    f_imghandle:= nil;
  end;

  destructor TImageSaveTask.Destroy;
  begin
    if Assigned(f_imghandle) then TImageDeleteTask.Create(f_imghandle);
    inherited;
  end;

  constructor TScreenshotMakeTask.Create;
  begin
    inherited Create;
    Reschedule;
  end;

  function TImageSaveTask.GetType: TThreadedTaskType;
  begin
    Result:= ttt_Imaging;
  end;

  function TScreenshotMakeTask.GetType: TThreadedTaskType;
  begin
    Result:= ttt_Imaging;
  end;




  function TImageDecodeTask.Perform: boolean;
  var
    q: qword;
  begin
    Result:= true; //in case of failure, consider task completed as well
    if Mother^.Debug.Verbose then AddLog('TImageDecodeTask.Perform');
    if Mother^.Debug.Verbose then AddLog('  encoded size=%0', [ f_insize]);
    if not Assigned(f_pinbuff) or (f_insize <= 0)
      then Die(RuEn(
        'Попытка декодировать изображение из пустого буфера!',
        'Attempt to decode image from empty buffer!'));
//bgSay('doing byte(nil^):= 0 ...');
//byte(nil^):= 0;
    Mother^.Timer.UsecDelta(@q);
    f_imghandle:= Mother^.Image.LoadFromMem(f_pinbuff, f_insize);
    AfterEfCheck;
    f_poutbuff:= Mother^.Image.GetPixels(f_imghandle);
    AfterEfCheck;
    f_outsize:= Mother^.Image.GetBytes(f_imghandle);
    AfterEfCheck;
    f_width:= Mother^.Image.GetWidth(f_imghandle);
    AfterEfCheck;
    f_height:= Mother^.Image.GetHeight(f_imghandle);
    AfterEfCheck;
    f_format:= Mother^.Image.GetFmt(f_imghandle);
    AfterEfCheck;
    if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecDelta(@q))]);
    if Mother^.Debug.Verbose then AddLog('  dimensions %0x%1, format %2h', [f_width, f_height, GlTextureFormatToString(f_format)]);
  end;

  function TImageResampleTask.Perform: boolean;
  var
    q: qword;
  begin
    if not f_source.Completed then Exit(false); //postpone. This is unlikely situation, but still.

    Result:= true; //in case of failure, consider task completed as well
    if Mother^.Debug.Verbose then AddLog('TImageResampleTask.Perform');
    if Mother^.Debug.Verbose then AddLog('  source dimensions %0x%1, format %2h', [f_source.Width, f_source.Height, GlTextureFormatToString(f_source.Format)]);
    Mother^.Timer.UsecDelta(@q);
    if (f_resample_width = f_source.Width) and (f_resample_height = f_source.Height)
    then begin
      //sizes equal, only format differs.
      //in fact, I don't want to know if resample routine is smart enough.
      //we just convert the format;
      f_imghandle:= Mother^.Image.Clone(f_source.ImageHandle);
      AfterEfCheck;
      Mother^.Image.SetFmt(f_imghandle, f_resample_format);
      AfterEfCheck;
      if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecDelta(@q))]);
      if Mother^.Debug.Verbose then AddLog('  format %0h', [GlTextureFormatToString(f_format)]);
    end
    else begin
      f_imghandle:= Mother^.Image.CloneResample (f_source.ImageHandle,
        f_resample_width, f_resample_height, f_cycled, f_resample_format);
      AfterEfCheck;
      f_width:= Mother^.Image.GetWidth(f_imghandle);
      AfterEfCheck;
      f_height:= Mother^.Image.GetHeight(f_imghandle);
      AfterEfCheck;
      f_format:= Mother^.Image.GetFmt(f_imghandle);
      AfterEfCheck;
      if Mother^.Debug.Verbose then AddLog('  completed in %0 μs', [round(Mother^.Timer.UsecDelta(@q))]);
      if Mother^.Debug.Verbose then AddLog('  dimensions %0x%1, format %2h', [f_width, f_height, GlTextureFormatToString(f_format)]);
    end;
  end;

  constructor TImageDecodeTask.Create(pbuf: pointer; bsize: longint);
  begin
    inherited Create;
    f_pinbuff:= pbuf;
    f_insize:= bsize;
    Reschedule;
  end;

  constructor TImageResampleTask.Create(src: TImageDecodeTask;
    newx, newy: integer;  newformat: GLuint; cycled: boolean = false);
  begin
    f_source:= src;
    f_resample_width:= newx;
    f_resample_height:= newy;
    f_resample_format:= newformat;
    f_cycled:= cycled;
    Reschedule;
  end;

  destructor TImageDecodeTask.Destroy;
  begin
    if Assigned(f_imghandle) then begin
      //Mother^.Image.Delete(f_imghandle);  //OOPSIE. :( Wrong tread to call it from!
      //AfterEfCheck;
      TImageDeleteTask.Create(f_imghandle);
    end;
    inherited;
  end;

  function TImageDecodeTask.GetType: TThreadedTaskType;
  begin
    Result:= ttt_Imaging;
  end;




  function TImageDeleteTask.Perform: boolean;
  begin
    Result:= true;
    if Mother^.Debug.Verbose then AddLog('TImageDeleteTask.Perform (handle=%0)', [f_imghandle]);
    if Assigned(f_imghandle) then begin
      Mother^.Image.Delete(f_imghandle);
      AfterEfCheck;
      f_imghandle:= nil;
    end;
    if Mother^.Debug.Verbose then AddLog('  image deleted.');
  end;

  function TImageDeleteTask.GetType: TThreadedTaskType;
  begin
    Result:= ttt_Imaging;
  end;

  constructor TImageDeleteTask.Create(Imagehandle: pointer);
  begin
    inherited Create;
    f_imghandle:= Imagehandle;
    Reschedule;
  end;


end.
