{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


unit mo_loba;

{$include mo_globaldefs.h}

interface

uses
  SysUtils, Classes, math, chepersy, mo_hub, typinfo, mo_classes;

type

  TLoadBalancer = class //NOT managed, it is useless to store it, gets recreated at each run

    {
     This class manages load balancing, thus allowing the engine
        to achieve its awesome 100x hardware requirements range,
        utilizing anything from ancient office desktops of early 2000s
        to modern gaming rigs.
    }

  private
    suspended,
    suspend_frames,
    avframes: integer;
    avoqf,
    oqf: float;
    procedure CheckBounds;
  public
    constructor Create;
    destructor Destroy;
    procedure TextureIsNotInTextureMemory(megabytes: float);
    function Enabled: boolean;
    procedure Pulse;
    procedure Suspend; //ignore this frame and the next one;
    procedure EmergencyDrop; // halves qf
    function RecommendedTimeLimitForBackgroundTasksInMainThread: double; //microseconds
  end;

implementation

  uses mo_logic, mo_module;

//  QF_MIN = 10;
//QF_MAX = 999;
//  FPS_CULL_DEFAULT = 30;

  procedure TLoadBalancer.CheckBounds;
  begin
    if Mother^.State.BatteryRemaining <= 100
      then Mother^.Timer.QF:= min(Mother^.Timer.QF, min(100, Mother^.State.BatteryRemaining * 2.0));
    Mother^.Timer.QF:= math.max(QF_MIN, math.min(QF_MAX, Mother^.Timer.QF));
  end;

  procedure TLoadBalancer.Suspend;
  begin
    suspend_frames:= math.max (2, round(10 + 2 - (Mother^.Timer.QF / 10)));
    Mother^.Timer.SuspendSwapBuffersMeasurement:= true;
  end;

  procedure TLoadBalancer.EmergencyDrop;
  begin
    Mother^.Timer.QF:= math.min(100, Mother^.Timer.QF/2);
    CheckBounds;
  end;

  procedure TLoadBalancer.Pulse;
  var
    OptimalQf, TargetQFByFPS,
    delta, f, aniso, FPSFloor
       : float;
    when: TLogicNextPulseWhenEnum = lnf_WaitForRender;
  begin
    if Assigned(Module.Logic) then when:= Module.CachedLogicNextPulseWhen;

    case when of
      lnf_WaitForRender : begin
        FPSFloor:= Mother^.Timer.CullFrameRate - 5;
      end;
      lnf_Steady100Hz: begin
//addlog('%0',[Mother^.Display.WindowHasFocus]);
        if Mother^.Display.WindowHasFocus then begin
          FPSFloor:= 60.0;
          Mother^.Timer.CullFrameRate:= 999; //limited by logic pulsing at 100Hz
        end
        else begin
          FPSFloor:= Mother^.Timer.CullFrameRate - 5;
        end;
      end;
      lnf_ASAPRenderAt30: begin
        Mother^.Timer.CullFrameRate:= 30;
        FPSFloor:= 0; //Have NO effect on QF.
      end;
    else
      Die(MI_ERROR_PROGRAMMER_NO_BAKA, ['Unsupported Logic.NextPulseWhen='
        + GetEnumname(typeinfo(TLogicNextPulseWhenEnum),
            Ord(Module.Logic.NextPulseWhen))]);
    end;

    if Enabled then begin
      if suspend_frames > 0 then
        if suspended > 30
          then suspend_frames:= 0
          else begin
            Dec(suspend_frames); //Exit;
          end;
      if suspend_frames = 0 then
        Mother^.Timer.SuspendSwapBuffersMeasurement:= false;
    end;


//addlog ('%0  %1  %2',[FPSFloor, FPSFactor, TargetQFByFPS] );

    // at 16000 microseconds and 60 FPS, the call to SwapBuffers would take almost entire frame time.
    // at 64000, the videocard couldn't even keep 30 FPS (*cough* mesa software implementation in Debian *cough*)
    // 6-8 thousands us is meh (your average videocard with Aero transparent windows overlapping our own)
    //OptimalQf:=  (640000.0 / Mother^.Timer.AverageSwapBuffersDuration);
    f:= Mother^.Timer.AverageSwapBuffersDuration / 10000.0; // 1.0 => qf = 100
    if f > 1 then OptimalQf:= 100 / sqrt(f);
    if f < 1 then OptimalQf:= 100 + ((1/Math.max(f, 0.01)) - 1) * 1498;
    // 64.000 => 10  //15 FPS pure
    // 10.000 => 100 //100 FPS pure
    //  6.000 => 999 //333 FPS pure


    //SwapBuffers alone doesn't give the whole picture.
    if Mother^.Timer.FPS < FPSFloor  then begin
      TargetQFByFPS:= Mother^.Timer.QF * sqr(sqr(Mother^.Timer.FPS / FPSFloor));
      OptimalQF:= min(OptimalQF, TargetQFByFPS); //blend two methods smoothly
    end;


//    DbgSay(PervertedFormat('FPSFloor = %0; OptimalQf = %1; when=%2; lag=%3ms',
//      [FPSFloor, Round(OptimalQf), GetEnumName(typeinfo(TLogicNextPulseWhenEnum), ord(when)), round(Module.Logic.AccumulatedLag/1000)]));

    avoqf+= OptimalQf;
    inc (avframes);
    if avframes > 15 then begin
      oqf:= avoqf / avframes;
      avoqf:= 0.0;
      avframes:= 0;
    end;

    if oqf < 1 then Exit; //not initizlized yet

    if oqf > Mother^.Timer.QF then begin
      //it's *per frame*, folks. So delta should be tiny.
      delta:= math.min(0.099, 0.002 * sqrt(oqf - Mother^.Timer.QF));
    end
    else begin
      //there is NO limit to how fast it could drop
      delta:= - 0.0001 * sqr(Mother^.Timer.QF - oqf);
    end;


    {
      decrease slowly if window does not have focus. Otherwise it could rise
      to insane levels if window is not visible, up to causing overall system slowdown

    }
    if not Mother^.Display.WindowHasFocus and (Mother^.Timer.QF > 101)
      then delta:= Math.min(delta, - 1 / Math.max(10, Mother^.Timer.FPS));

    if Enabled then begin
      Mother^.Timer.QF += delta;
      CheckBounds;
    end;

    //4 at Timer.QF = 100, 16 at Timer.QF = 1600
    if Mother^.GAPI.Initialized
      then glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
        //16
        Math.min(Mother^.GAPI.MaxAnisotropy, Math.max(1.0, 4.0 * sqrt(Mother^.Timer.QF / 100)))
        //Math.max(1.0, Math.min(Mother^.GAPI.MaxAnisotropy, 4.0 * sqrt(Mother^.Timer.QF / 100)))
      );
//DbgSay('cap %0', [Mother^.Timer.CullFrameRate]);
  end;


  constructor TLoadBalancer.Create;
  begin
    inherited;
  end;

  destructor TLoadBalancer.Destroy;
  begin
    inherited;
  end;

  procedure TLoadBalancer.TextureIsNotInTextureMemory(megabytes: float);
  begin
    Mother^.Timer.QF-= megabytes;
    CheckBounds;
  end;

  function TLoadBalancer.Enabled: boolean;
  begin
    Result:= not Mother^.Timer.QF_ManualOverride;
  end;

  function TLoadBalancer.RecommendedTimeLimitForBackgroundTasksInMainThread: double; //microseconds
  begin
    //***TODO: Implement!
    Result:= 5000.0;
  end;

end.

