{
    This file is part of Chentrah,
    Copyright (C) 2004-2012 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


  {$include mo_globaldefs.h}
  {$typeinfo on}

unit mo_module;

interface
uses
  SysUtils, Classes, math, {$ifdef windows} Windows, {$endif}
    mo_hub, //main API
    mo_threads, //thread manager, contains parts of control logic
    mo_loba, //load balancer
    mo_logic //application control logic
   , syncobjs
 ;
  

const
  NonKeys: set of TKey = [KEY_NULL, KEY_MWHEEL_DOWN, KEY_MWHEEL_UP];
//  MainMenuInvocationKey: TKey = KEY_ESCAPE;
  JoystickHistoryLimit = 10; { The last movements saved for possible
    stochastical filtering. Analog joysticks are noisy beasts. }

type

  { TModule }

  TModule = class
  protected
    f_ThisRunFrame: integer;
    f_IsAlreadyWaitingForThreadsToTerminate: boolean;
    f_SessionHash,
    f_LoadedHash: TAssetHash;
    f_GoingToEraseSession: boolean;
    f_QuickSaveHotKey,
    f_QuickLoadHotKey,
    f_ScreenShotHotKey,
    f_MainMenuHotKey: TKeySetArray;
    f_FirstCycle: boolean;
    f_ControlLogicThreadStackSize,
    f_BackgroundThreadStackSize: ptrUint;
    LogicHadResponded: boolean; { to differ reactions to it not responding
                                                at startup and during gameplay }
    LogicNotRespondingFrames: longint;
  public

    Logic: TLogic;
    LoadBalancer: TLoadBalancer;
    Threadmanager: TDummyThreadManager;

    CachedLogicNextPulseWhen: TLogicNextPulseWhenEnum; { is filled
                                                        by CallLogicForRender }

    property ThisRunFrame: integer read f_ThisRunFrame;
    property IsAlreadyWaitingForThreadsToTerminate: boolean read f_IsAlreadyWaitingForThreadsToTerminate write f_IsAlreadyWaitingForThreadsToTerminate;
    property SessionHash: TAssetHash read f_SessionHash;
    property LoadedHash: TAssetHash read f_LoadedHash write f_LoadedHash;
    property GoingToEraseSession: boolean read f_GoingToEraseSession write f_GoingToEraseSession;
    property QuickSaveHotKey: TKeySetArray read f_QuickSaveHotKey;
    property QuickLoadHotKey: TKeySetArray read f_QuickLoadHotKey;
    property ScreenShotHotKey: TKeySetArray read f_ScreenShotHotKey;
    property MainMenuHotKey: TKeySetArray read f_MainMenuHotKey;
    property FirstCycle: boolean read f_FirstCycle;
    property ControlLogicThreadStackSize: ptrUint read f_ControlLogicThreadStackSize;
    property BackgroundThreadStackSize: ptrUint read f_BackgroundThreadStackSize;



    constructor Create;
    destructor Destroy; override;

//    procedure OnGetFocus; virtual;
//    procedure OnLoseFocus; virtual;
    function Pulse: boolean;
    function Cleanup: boolean;

    procedure ProcessMessages; virtual;

    function GetConfigStr(name: AnsiString; defval: AnsiString = ''): AnsiString;
    procedure SetConfigStr(name, val: AnsiString);
    procedure SetConfigInt (id: AnsiString; Value: integer);
    function GetConfigInt (id: AnsiString; Defval: integer = 0): integer;
    procedure Restart;
    procedure NewSessionHash;

  end;

  CModule = class of TModule;

var
  Module: TModule;

  MyLogClass: CLogic; //gets reset to the true metaclass in M_Init

implementation

uses mo_sessionstack, mo_assets;


  Constructor TModule.Create;
  begin
    inherited;

    {
      The control logic is created elsewhere!
      See
        TLogicThread.Execute
      or
        TDummyThreadManager.Pulse - for single-threaded mode
    }

    mo_sessionstack.GetSessionIDFromConfig;
    SessionID:= mo_sessionstack.SessionId;
    try
      f_QuickSaveHotKey:= ParseHotKeyString(GetConfigStr('quick_save_hot_key','F5'));
      f_QuickLoadHotKey:= ParseHotKeyString(GetConfigStr('quick_load_hot_key','F9'));
      f_ScreenShotHotKey:= ParseHotKeyString(GetConfigStr('screenshot_hot_key','F3'));
      f_MainMenuHotKey:= ParseHotKeyString(GetConfigStr('main_menu_hot_key','ESCAPE,GAMEPAD_START'));
    except
      Die('Не удалось загрузить настройки горячих клавиш','Failed to load hot key config values')
    end;
//    if chepersy.IsRegistered(TAbstractThreadedTask) then Die('Developer''s error! TAbstractThreadedTask must NOT be registered!');

    f_ControlLogicThreadStackSize:= Math.max(
      Mother^.Module.ControlLogicThreadStackSize,
      1024 * 1024 * GetConfigInt('control_logic_thread_stack_size_megabytes', 16)
    );
    f_BackgroundThreadStackSize:= Math.max(
      Mother^.Module.BackgroundThreadStackSize,
      1024 * 1024 * GetConfigInt('background_thread_stack_size_megabytes', 4)
    );

    f_FirstCycle:= true;
    Module:= Self;

    TAssetManager.Initialize;

    if Mother^.Module.MultithreadingDisabled
      then Threadmanager:= TDummyThreadManager.Create
      else Threadmanager:= TThreadManager.Create;

    LoadBalancer:= TLoadBalancer.Create;
  end;

  destructor TModule.Destroy;
  begin
    Module:= nil;
    inherited;
  end;


{  procedure TModule.OnGetFocus;
  begin
    f_frametime:=0;
  end;
  
  procedure TModule.OnLoseFocus;
  begin

  end;
}


  function TModule.Pulse: boolean;
  begin
    Result:= true;

    Inc(f_ThisRunFrame);

    if not ThreadManager.Started then try
      ThreadManager.Start;
    except
      Die(
        'Крах при запуске диспетчера потоков.',
        'Crashed trying to start the thread manager.');
    end;

    if Assigned(Logic) and Logic.Crashed
      then Logic:= nil; { Ignore the faulty thing. The memory manager is going
        to be purged anyway when the DLL reloads. }

    ProcessMessages;

    ThreadManager.RenderDoneLogicMayProceed.SetEvent; //if it waits for the next frame

    try
      if ThreadManager.CallLogicForRender {
        Synchronizes the main thread and the logic thread, calling
        Logic.Render that includes actualizing assets.

        ***TODO: make texture uploads and such into background tasks,
          to avoid executing them here and thus stalling the logic thread
          unnecessarily
      }

      then begin
        if (f_ThisRunFrame >= 10) and Assigned(Logic) and not Logic.IsLoadingAssets
          then LoadBalancer.Pulse;

        if not LogicHadResponded then begin
          LogicHadResponded:= true;
          Mother^.Display.BeginBlackout; //initiate fade in from black
          Mother^.Display.MousePointer.LoadDefault;
          AfterEfCheck;
        end;
        LogicNotRespondingFrames:= 0;
      end
      else begin
        //The control logic is suddenly busy.
        //Render a blank screen
        Inc(LogicNotRespondingFrames);

        if Mother^.Display.CanRender then begin
          if LogicNotRespondingFrames = 1 then begin
            Mother^.Display.BeginBlackout(1.0);
            //***TODO: Investigate why this stably leads to an AV
            //     (caught in LoadPicFromMemAs in cl_mother_api_imaging.inc )
            Mother^.Display.MousePointer.LoadHourglass;
            AfterEfCheck;
          end;

          if LogicHadResponded then begin { suddenly! doing something lenghty }
            if LogicNotRespondingFrames > 2 { do nothing for the first 2 frames. }
              then Logic.RenderBusyScreenInMainThread
              else Mother^.GAPI.SkipSwapBuffers:= true; { to avoid flickering
                of an unrendered frame. The mouse pointer, console and FPS counter
                will overlap, which is not as noticeable. }
          end
          else begin
            //too busy at startup
            Mother^.Display.RenderBackground;
            AfterEfCheck;
          end;
         end;
      end;
    except
      //screen it! Mother reraises when appropriate.
      NoteThreadOccupation(tlk_Working);
    end;

    ThreadManager.Pulse; //includes running background tasks in the main thread

    f_FirstCycle:= false;

    if Mother^.Module.MustUnload
      and Threadmanager.Done
      and not Mother^.Module.SessionNotSaved
    then Mother^.Module.ReadyToUnload:= true;
  end;

  function TModule.Cleanup: boolean;
  begin
    Result:= true;
    if Assigned(ThreadManager) then ThreadManager.Free;
    if Assigned(LoadBalancer) then LoadBalancer.Free;
    TAssetManager.Finalize;
    try
      CleanupLineInfo();
    except
      AddLog('WARNING: failed to clean up chelinfo! %0',[StopDying()]);
      Result:= false;
    end;
  end;

  procedure TModule.ProcessMessages;
  var
    e: PInputEvent;
  begin
    e:= Mother^.Input.Events;  //an array terminated with element with EventType of 0.
    if not Assigned(e) or not Assigned(Logic) then Exit;
    while e^.EventType <> 0 do begin
      Logic.PushInputEvent(e);
      inc(e);
    end;
  end;

  function TModule.GetConfigStr(name: Ansistring; defval: AnsiString = ''): AnsiString;
  begin
    Result:= string(Mother^.Config.GetString(Mother^.Module.PureName, PAnsiChar(name)));
    if Result = '' then Result:= defval;
  end;

  procedure TModule.SetConfigStr(name, val: Ansistring);
  begin
    Mother^.Config.SetString(Mother^.Module.PureName, PAnsiChar(name), PAnsiChar(val));
  end;

  procedure TModule.SetConfigInt (id: AnsiString; Value: integer);
  begin
    mo_hub.SetConfigInt(Mother^.Module.PureName, id, Value);
  end;

  function TModule.GetConfigInt (id: AnsiString; Defval: integer): integer;
  begin
    Result:= mo_hub.GetConfigInt(Mother^.Module.PureName, id);
    if Result = 0 then Result:= Defval;
  end;

  procedure TModule.Restart;
  begin
    Mother^.Module.State:= ms_Unloading;
  end;


  function PureName(a: string): string;
  //uniform file name, identic for both Windows and Linux
  begin
    Result:= ChangeFileExt(ExtractFileName(a), '');
    if (Length(Result) >=3) and (Copy(Result, 1, 3)='lib')
      then Result:=Copy(Result, 4, Length(Result) - 3);
  end;

  procedure TModule.NewSessionHash;
  begin
    Mother^.AssetKeeper.GenHash(@f_SessionHash);
    AfterEfCheck;
  end;

end.
