{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

    This unit contains classes for working with pak files
    (initially, only ZIP aka PK3 are supported)

 **********************************************************************}

{

Warning! All these classes are mostly invoked not from the main thread,
  but from the file management thread.

Locking is not necessary, though, as asset management guarantees
  that only one operation is performed at any given time

}

{$include mo_globaldefs.h}
{$typeinfo on}

unit mo_pakfiles;
interface
uses SysUtils, Classes {$ifdef windows}, Windows {$else} , baseunix, dl {$endif},
  mo_hub, chepersy, mo_classes, syncobjs, mo_threads, mo_assets, mo_mstream,
   {$if defined(windows) and defined(win_unicode_fnames)}
     un_filelist_wide
   {$else}
     un_filelist
   {$endif}
  , un_unzip
  ;


procedure RegisterPakFileClasses;

type

  TAbstractPackFile = class; //forward declaration

  PFileListEntry = ^TFileListEntry;
  TFileListEntry = packed record
    Hash: TAssetHash; //includes time and size
    name: TFileNameString;
    pak_file: TAbstractPackFile;
  end;

  TFileListArray = array of TFileListEntry;

  TArrayOfByte = array of byte;

  { TAbstractPackFile }

  TAbstractPackFile = class(THandleAsset)
    {
      The handle here is mother stream class instance
    }
  protected
    f_init_task: TAbstractThreadedTask;
    f_name, f_rootdir, f_readfname: TFileNameString;
    f_fsize: int64;
    f_fdate: TDateTime;
    f_list: TFileListArray;
    f_stream: TMotherStream;
    f_stream_handle: pointer;
  public
    function Initialize: boolean; virtual; //is ONLY called by threaded tasks from their thread!
    property List: TFileListArray read f_list write f_list;
    property Name: TFileNameString read f_name;
    property FileSize: int64 read f_fsize;
    property FileDate: TDateTime read f_fdate;
    property Stream: TMotherStream read f_stream;
    property StreamHandle: pointer read f_stream_handle;
    procedure RegisterFields; override;
    procedure GenResource; override;
    procedure FreeResource; override;
    function Actualize: boolean; override;
    function IsValidForServer: boolean; override;
//    procedure Scrape; override;

    //***TODO : make root dir work. Just a stub now.
    constructor Create(fname: TFileNameString; frootdir: TFileNameString = DS);

    //To be invoked from the file management tread only
    function ReadFile(fname: TFileNameString): TArrayOfByte; virtual; abstract;

    procedure UpdateHash; override; //Does nothing, hash creation
                                   //is delegated to TPakFileInitTask.Perform


    function Devour(c: TGenericAsset): boolean; override;
    procedure EraseContainer; override;
  end;

  TArrayOfPackFiles = array of TAbstractPackFile;

  { TZipFile }

  TZipFile = class(TAbstractPackFile)
  protected
    f_unzip: TUnzip;
  public
    procedure RegisterFields; override;
//    procedure Scrape; override;
    function ReadFile(fname: TFileNameString): TArrayOfByte; override;
    function Initialize: boolean; override; //is ONLY called by threaded tasks from their thread!

    function Devour(c: TGenericAsset): boolean; override;
    procedure EraseContainer; override;
  end;

implementation
  uses mo_module, mo_filemanager, crc;

  type

    { TZipFileOpenTask }

    TPakFileInitTask = class(TAbstractThreadedTask)
    private
      f_master: TAbstractPackFile;
      f_hash: TAssetHash;
      function Perform: boolean; override;
    public
      constructor Create(master: TAbstractPackFile; _hash: TAssetHash);
      function GetType: TThreadedTaskType; override;
      property Hash: TAssetHash read f_hash;
    end;

    constructor TPakFileInitTask.Create(master: TAbstractPackFile; _hash: TAssetHash);
    begin
      f_master:= master;
      f_hash:= _hash;
      Reschedule;
    end;

    function TPakFileInitTask.Perform: boolean;
    var
      fname: TFileNameString;
    begin
      Result:= true; //in case of failure, consider task completed as well
      if Mother^.Debug.Verbose then AddLog('TPakFileInitTask.Perform'#10#13'  %0 / %1', [f_master.Name, HashToHex(f_hash)]);

//bgSay('TPakFileInitTask.Perform: f_master=' + IntToHex(dword(pointer(f_master)), 8));
//bgSay('  f_master.Initialize=' + IntToHex(dword(@f_master.Initialize), 8));
//bgSay('  f_master is ' + f_master.ClassName);
      if not f_master.Initialize then begin
        if Mother^.Debug.Verbose then AddLog('Aborting task: failed to initialize the container.');
        Exit;
      end;
      f_master.Hash:= f_hash;
    end;


    function TPakFileInitTask.GetType: TThreadedTaskType;
    begin
      Result:= ttt_File;
    end;


procedure RegisterPakFileClasses;
begin
  RegClass(TAbstractPackFile);
  RegClass(TZipFile);
end;


{ TAbstractPackFile }

function TAbstractPackFile.Actualize: boolean;
begin
  if not Assigned(f_init_task)  then begin
    if not Assigned(Module.Logic.FileManager) then Exit(false);
    f_init_task:= TPakFileInitTask.Create(
      Self, Module.Logic.FileManager.GetHash(f_name));
  end;
  if not f_init_task.Completed then Exit(false);
  Result:= inherited;
  if Actualized then begin
    f_init_task.Free;
    pointer(f_init_task):= nil;
  end;
end;

procedure TAbstractPackFile.RegisterFields;
begin
  RegType(typeinfo(TFileListEntry), sizeof(TFileListEntry), [
    'hash', typeinfo(TAssetHash),
    'name', typeinfo(TFileNameString),
    'pak_file', typeinfo(TAbstractPackFile)

  (*
    'name', typeinfo(TFileNameString),
    'size', typeinfo(int64),
    'modify_time', typeinfo(TDateTime),
    'pak_file', typeinfo(TMemoryMappedFile),
    'pak_file_offset', typeinfo(int64)
   *)
  ]);
  RegType(typeinfo(TFileListArray));//, typeinfo(TFileListEntry));
  RegType(typeinfo(TArrayOfPackFiles));//, typeinfo(TAbstractPackFile));
  inherited RegisterFields;
  ListFields([
    '-ptr1', @f_init_task, CPS_POINTER,
    'name', @f_name, 'dir', @f_rootdir, '-readfname', @f_readfname, typeinfo(TFileNameString),
    'fsize', @f_fsize, typeinfo(int64),
    'fdate', @f_fdate, typeinfo(TDateTime),
    'list', @f_list, typeinfo(TFileListArray),
    '-stream', @f_stream, CPS_POINTER, //TMotherStream;
    'stream handle', @f_stream_handle, typeinfo(ptruint) //pointer
  ]);
end;

function TAbstractPackFile.Devour(c: TGenericAsset): boolean;
var i: integer;
begin
  f_stream_handle:= (c as TAbstractPackFile).StreamHandle;
  if not Assigned (f_stream_handle) then Exit(false);
  if (f_name <>  (c as TAbstractPackFile).Name)
    or (f_fsize <>  (c as TAbstractPackFile).FileSize)
    or (f_fdate <>  (c as TAbstractPackFile).FileDate)
      then Exit(false);
  f_stream:= TMotherStream.CreateFromHandle(f_stream_handle, false);
  f_list:= COPY((c as TAbstractPackFile).List);
  for i:=0 to High(f_list) do f_list[i].pak_file:= Self;
  Result:= inherited Devour(c);
end;

procedure TAbstractPackFile.EraseContainer;
begin
  if Assigned(f_stream) then f_stream.Free;
  inherited;
  f_stream_handle:= nil;
  SetLength(f_list, 0);
  f_name:= '';
  f_fsize:= 0;
  f_fdate:= 0;
end;

constructor TAbstractPackFile.Create(fname: TFileNameString; frootdir: TFileNameString = DS);
begin
  inherited Create([]);
  f_name:= fname;
  f_rootdir:= frootdir;
end;

procedure TAbstractPackFile.GenResource;
begin
  //Do nothing; The actual creating of stream and filling f_stream_handle is performed in the file thread
end;

procedure TAbstractPackFile.UpdateHash;
begin
  //do nothing
end;

function TAbstractPackFile.IsValidForServer: boolean;
begin
  Result:= true;
end;

procedure TAbstractPackFile.FreeResource;
begin
  if _handle <> 0 then begin
    if Assigned(f_stream) then begin
      f_stream.Free;
      pointer(f_stream):= nil;
    end;
    //This is emergency, mind you. This should not happen.
    //Because FreeResource is run in the main thread, as opposed to the
    //  file manager thread
    if Mother^.Debug.Verbose then AddLog('%0.FreeResource frees the mother stream. This is a contingency!', [string(Self.ClassName)]);
    Mother^.Stream.Free(pointer(f_stream_handle));
    AfterEfCheck;
    f_stream_handle:= nil;
  end;
end;

function TAbstractPackFile.Initialize: boolean; //is ONLY called by threaded tasks from their thread!
var
  fname: TFileNameString;
begin
//bgSay('  TAbstractPackFile.Initialize ' + f_name);
  Assert(Assigned(Module.ThreadManager) and (
      Mother^.Module.MultithreadingDisabled
      or Module.ThreadManager.IsCurrentThreadTheFileThread),
    string(Self.ClassName) + '.Initialize should only be called from one specific thread!');
//gSay('    A');
  if Assigned(f_stream) then Exit(true);
//bgSay('    B');
  try
    fname:= UnmangleFileName(f_name);
    if Mother^.Debug.Verbose then AddLog('Opening mother file stream %0 ...', [fname]);
//bgSay('    C');
    f_stream:= TMotherStream.CreateAsFileStream(fname, fmOpenRead + fmShareDenyWrite, false);
//bgSay('    D');
    f_stream_handle:= f_stream.Handle;
    _handle:= 1;
    Result:= true;
    if Mother^.Debug.Verbose then AddLog('  ok, handle=%0', [pointer(f_stream_handle)]);
  except
//bgSay('    BAMTS');
    AddLog(RuEn(
      'Не удалось открыть файл для чтения'#10#13'  %0'#10#13'%1',
      'Failed to open file for reading'#10#13'  %0'#10#13'%1')
      ,[fname, StopDying()]);
    Result:= false;
  end;
end;

function TZipFile.Initialize: boolean; //is ONLY called by threaded tasks from their thread!
var
  ufi: unz_file_info;
  i: integer;
  fn, comment: AnsiString;
begin
  Result:= inherited Initialize;
  if Result and not Assigned(f_unzip) then begin
//bgSay('  TZipFile.Initialize');
    if Assigned(f_unzip) then Exit(true);
    try
//bgSay('    A');
      if Mother^.Debug.Verbose then AddLog('  Attaching a zlib object to the stream ...');
      f_unzip:= TUnzip.Create(f_stream);
//bgSay('    B');
      repeat
        if not f_unzip.GetCurrentFileInfo (@ufi, fn, nil, 0, comment) then break;
        if ufi.uncompressed_size <= 0 then continue; //there are folders and shit
        SetLength(f_list, Length(f_list) + 1);
        with f_list[High(f_list)] do begin
//bgSay('      ' +  fn);
          name:= '*M/' + fn;
          pak_file:= Self;
          //Hash.Age:= ufi.dosDate;
          //Hash.NameCrc:= crc32(0, @name[1], Length(name) * SizeOf(name[1]));
          Hash.FileAge:= FileDateToDateTime(ufi.dosDate);
          Hash.NameMd5:= Md5String(name);
          Hash.FileSize:= ufi.uncompressed_size;
          if Mother^.Debug.Verbose then AddLog('    %0 , size=%1, ratio=%2%',
            [name, ufi.uncompressed_size, round (100.0 * ufi.compressed_size / max(1, ufi.uncompressed_size))]);
        end;
//      function GetCurrentFileInfo(pfile_info: unz_file_info_ptr; var FileName: AnsiString;
//extraField: pointer; extraFieldBufferSize: longint; var Comment: AnsiString): boolean; { ZEXPORT }
      until not f_unzip.GoToNextFile();
//bgSay('    C');
      if Mother^.Debug.Verbose then AddLog('  Parsed successfully, found %0 files.', [Length(f_list)]);
    except
//bgSay('    BUMTS');
      AddLog(RuEn(
        'Не удалось инициализировать распаковщик unzip'#10#13#10#13'%0',
        'Failed to initialize unzip unpacker'#10#13#10#13'%0')
        ,[StopDying()]);
      Result:= false;
    end;
  end;
//bgSay('    Ok');
end;


function TZipFile.Devour(c: TGenericAsset): boolean;
begin
  Result:= inherited Devour(c);
  if Assigned(f_stream) then f_unzip:= TUnzip.Create(f_stream);
end;

procedure TZipFile.EraseContainer;
begin
  if Assigned(f_unzip) then f_unzip.Free;
  inherited;
  pointer(f_unzip):= nil;
end;

procedure TZipFile.RegisterFields;
begin
  inherited;
  ListFields([
    '-unzip', @f_unzip, CPS_POINTER
  ]);
end;

function TZipFile.ReadFile(fname: TFileNameString): TArrayOfByte;
var
  ufi: unz_file_info;
  fn, comment: AnsiString;
  i, read_how_much: integer;
  err: boolean;
  q: Qword;
begin
//bgSay('TZipFile.ReadFile ' + fname);
  if Mother^.Debug.Verbose then AddLog('    %0.ReadFile(%1)...', [string(Self.ClassName), fname] );
  if not Initialize then begin
    AddLog(RuEn(
      'Загрузка прервана: не удалось инициализировать контейнер!'#10#13' файл %0'#10#13' контейнер %1',
      'Aborting the load: failed to initialize the container!'#10#13' file %0'#10#13' container %1'
      ), [fname, UnmangleFileName('*M/' + f_name)]);
    Exit(nil);
  end;
//bgSay('  b');
  if COPY(fname, 1, 3) <> '*M/' then begin
    AddLog(RuEn(
      'Загрузка прервана: невалидный базовый путь!'#10#13' файл %0'#10#13' контейнер %1',
      'Aborting the load: invalid base path!'#10#13' file %0'#10#13' container %1'
      ), [fname, UnmangleFileName('*M/' + f_name)]);
    Exit(nil);
  end;
//bgSay('  c');
  fname:= COPY(fname, 4, Length(fname) - 3);
//bgSay('  fname = ' + fname);
  if not f_unzip.LocateFile(fname)
  then begin
    AddLog(RuEn(
      'Загрузка прервана: файл не найден в контейнере!'#10#13' файл %0'#10#13' контейнер %1',
      'Aborting the load: file not found in the container!'#10#13' file %0'#10#13' container %1'
      ), ['*M/' + fname, UnmangleFileName('*M/' + f_name)]);
    Exit(nil);
  end;
//bgSay('  d');
  Mother^.Timer.UsecDelta(@q);
  err:= not f_unzip.GetCurrentFileInfo (@ufi, fn, nil, 0, comment);
  if not err then err:= not f_unzip.OpenCurrentFile();
  if not err then begin
    SetLength(Result, ufi.uncompressed_size);
    i:= 0;
    while i < Length(Result) do begin
      read_how_much:= f_unzip.ReadCurrentFile(@Result[i], Length(Result) - i);
      if read_how_much <=0 then begin
        err:= true;
        break;
      end;
      i+= read_how_much;
    end;
  end;
  if err then begin
    AddLog(RuEn(
      'Загрузка прервана: ошибка чтения из контейнера!'#10#13' файл %0'#10#13' контейнер %1',
      'Aborting the load: error reading from the container!'#10#13' file %0'#10#13' container %1'
      ), [fname, UnmangleFileName('*M/' + f_name)]);
    SetLength(Result, 0);
  end;
  if Mother^.Debug.Verbose then AddLog('      ..unpacked %0 bytes in %1 μs', [Length(Result), round(Mother^.Timer.UsecDelta(@q))]);
//bgSay('  Ok');
end;


end.
