{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

unit mo_texture;

{$include mo_globaldefs.h}

interface

uses
  sysutils, Classes, math, chepersy, mo_hub, typinfo,
  mo_classes, mo_threads, mo_assets, mo_image, mo_gmathbase, mo_filemanager;

  //incomplete, watch out for formats it doesn't know (returns 4).
  function TextureFormatToBytesPerTexel(format: GLuint): integer;

const
  GIANT_TEXTURE_THRESHOLD = 2048; {textures larger than this are uploaded
    part by part using glTexSubImage2d()
  }
type

  TTexturePropertiesEnum = (
    tp_cyclic, //wrap_s and wrap_t
    tp_nearest, //use _nearest for filtering
    tp_not_cycle_3rd_axis, //for storing an atlas of cycled 2d textures in a 3d texture
    tp_compressed, //use DXT. relies on the driver to do the compression.

    tp_sharp_alpha,  { for thin details that should not disappear (branches,
      leaves, chicken wire, etc.) Use a specialized mipmap building mechanism
      that propagates more opaque colors so that if even one of source pixels
      is opaque, the destination will be opaque too.
      also, if compressed, DXT1 will be used istead of DXT5 }

    tp_keep_image //do not delete the image decoding task right away, its fate
      // will be decided by the atlas owning this texture or some other manager.
  );
  TTexturePropertiesSet = set of TTexturePropertiesEnum;

  TTexture = class (TTextureLikeAsset)
  protected
    f_tprops: TTexturePropertiesSet;
    f_depth, //for 3d textures
    f_format,
    f_imgformat,
    f_max_mip: glUint;
    procedure AllocateBlankTextureImage; virtual;
    procedure GenResource; override;
    procedure FreeResource; override;
    function AssumedWeight: float; virtual; {
      for QualityFactor drop when not in texture memory.
      Returns assumed texture memory size in megabytes }
    function BlankImageFormatFromFormat: glEnum; virtual;
    function MustBeFreedFromMainThread: boolean; override;

  public
    property TextureProperties: TTexturePropertiesSet read f_tprops;
    property MaxMip: GLuint read f_max_mip;
    property Format: GLuint read f_format;
    property Depth: GLuint read f_depth;
    property ImageFormat: GLuint read f_imgformat;
    constructor Create(
      _width,
      _height: integer;
      _format: GLuint = 0; //0 means use DefaultFormat method
      _image_format: GLuint = 0; //0 means use DefaultImageFormat method
      _max_mip: GLuint = 0; //0 means no mipmaps
      tp: TTexturePropertiesSet = [];
      p: TAssetPropertiesSet = []);
    procedure RegisterFields; override;
    function SortOrder: ffloat; override;
    function IsResident: boolean; { Check if thexture is in the video memory,
        not in the system memory. Is used for curbing Quality Factor.
      Any modern driver won't even let us know, always returning true.
      But for video cards manufactured in mid-2000s, this check is vital
        as their video memory is abysmal and their drivers
        reply to this request properly.
      Gets called by the Bind() method, causes QF to drop sharply every time
        a texture is found to be not resident
    }
    function MinFilter: GLuint; virtual; //default depends on tp_nearest and mip number
    function MagFilter: GLuint; virtual; //default depends on tp_nearest
    function Target: GLuint; virtual; //default GL_TEXTURE_2D
    function DefaultFormat: GLuint; virtual; //GL_RGBA8 or GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
    function DefaultImageFormat: GLuint; virtual; //GL_BGRA
    function SmallestMipSize: integer; virtual; //default 16

    procedure ValidateTextureDimensions; virtual; { Checks maximum size,
      makes sure sizes are power of two on videocards not supporting NPOT,
      makes sure the highest mip level has dimensions divisible to 4 if
      the format is DXT
      Also culls the number of mip levels (see SmallestMipSize)
    }

    procedure Bind; {not even virtual.  Performs a sanity check, to avoid
      binding texture 0: quoting http://410chan.org/dev/res/4274.html#4274 ,
      "А все потому, что текстуры 0 нет, текстуры выделяются с номера 1,
      использование 0 ведет к неопределенному поведению,
      в том числе и мусор, в том числе может быть и чужая память, и следующая
      запись туда может вообще все распидорасить,"
      -- and that's horrible!
    }

    function Devour(c: TGenericAsset): boolean; override;

  end;


 // TImage = class; //forward declaration

  TTextureFromImage = class(TTexture)
    {NOTE: until the asset is actualized (i.e. the image is loaded, decoded
      and resized and all mip levels are loaded),
      the dimensions return 1,
      the format returns GL_RGB, and so on

     NOTE2: mip levels are always made by resampling the full image, not the
       previous mip level's scaled version. This could be *slow* but gives
       marginally better image quality
    }
  protected
    f_file_name: TFileNameString;
    f_file_read_task: TFileLoadTask;
    f_img_decode_task: TImageDecodeTask;
    f_img_resample_task: TImageResampleTask;
    f_uploading_mip_level, //with mipmapped textures, several resampling
      //steps are required
    f_uploaded_part, //truly gigantic textures would get uploaded across several
      //frames using texSubImage2D, to avoid stalling the main thread
    f_upload_stage //to make dancing through the steps more explicit;
      : integer;
  public
    constructor Create(
      _fname: TFileNameString;
      _max_mip: GLuint = 0;
      tp: TTexturePropertiesSet = [];
      ap: TAssetPropertiesSet = []);
    property FileName: TFileNameString read f_file_name;
    property DecodeTask: TImageDecodeTask read f_img_decode_task;
    property ResampleTask: TImageResampleTask read f_img_resample_task;
    procedure RegisterFields; override;
    destructor Destroy; override;
    function Actualize: boolean; override;
    function Devour(c: TGenericAsset): boolean; override;

    function TextureFormatFromImageFormat: GLuint; virtual; { Depending
      on f_imgformat and compression settings,
      may return things from GL_RGBA8 to various DXT
    }

    procedure UpdateHash; override; //replacing it with empty stub that does nothing

    function CedeImageDecodeTask: TImageDecodeTask; //resets the field to nil, too;
  end;

  procedure RegClasses;


implementation

uses mo_module;

  procedure TTexture.RegisterFields;
  begin
    RegType(typeinfo(TTexturePropertiesSet));//, typeinfo(TTexturePropertiesEnum));
    inherited;
    ListFields([
      'texture properties', @f_tprops, typeinfo(TTexturePropertiesSet),
      'depth', @f_depth,
      'format', @f_format,
      'image format', @f_imgformat,
      'max mip level', @f_max_mip
    ]);
  end;

  function TTexture.MustBeFreedFromMainThread: boolean; begin Result:= True end;

  procedure TTextureFromImage.RegisterFields;
  begin
    inherited;
    ListFields([
      'file name', @f_file_name, typeinfo(TFileNameString),

      //none of these get saved into session (zeroed out at loading)
      '-frtask', @f_file_read_task,
      '-idtask', @f_img_decode_task,
      '-irtask', @f_img_resample_task, '*pointer',
      '-uml', @f_uploading_mip_level,
      '-upa', @f_uploaded_part,
      '-ups', @f_upload_stage, typeinfo(integer)
    ]);
  end;



  procedure TTexture.GenResource;
  var rep: glUint;
  begin
    try
      glGenTextures(1, @_handle);
      CheckGlError;

      glEnable(Target());
      CheckGlError;
      glBindTexture(Target(), _handle);
      CheckGlError;



      glTexParameteri(Target(), GL_TEXTURE_MIN_FILTER, MinFilter());
      glTexParameteri(Target(), GL_TEXTURE_MAG_FILTER, MagFilter());
      if tp_cyclic in f_tprops
        then rep:= GL_REPEAT
        else rep:= GL_CLAMP_TO_EDGE;
      glTexParameteri(Target(), GL_TEXTURE_WRAP_S, rep);

      if (Target() = GL_TEXTURE_2D) or (Target() = GL_TEXTURE_3D)
        then glTexParameteri(Target(), GL_TEXTURE_WRAP_T, rep);

      if Target() = GL_TEXTURE_3D then begin
        if (tp_cyclic in f_tprops) and not (tp_not_cycle_3rd_axis in f_tprops)
          then rep:= GL_REPEAT
          else rep:= GL_CLAMP;
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, rep);
      end;

      //creates a blank, uninitialized texture image
      AllocateBlankTextureImage;

      UpdateHash;
    except
      OnGenResourceFailure(StopDying());
    end;
  end;




  constructor TTexture.Create(
    _width,
    _height: integer;
    _format: GLuint = 0; //0 means use DefaultFormat method
    _image_format: GLuint = 0; //0 means use DefaultImageFormat method
    _max_mip: GLuint = 0; //0 means no mipmaps
    tp: TTexturePropertiesSet = [];
    p: TAssetPropertiesSet = []);
  begin
    inherited Create(_width, _height, p);
    if _format = 0
      then f_format:= DefaultFormat
      else f_format:= _format;
    if _image_format = 0
      then f_imgformat:= DefaultImageFormat
      else f_imgformat:= _image_format;
    f_max_mip:= _max_mip;
    f_tprops:= tp;
  end;

  constructor TTextureFromImage.Create(
    _fname: TFileNameString;
    _max_mip: GLuint = 0;
    tp: TTexturePropertiesSet = [];
    ap: TAssetPropertiesSet = []
  );
  var
    i: integer;
    zm: TTextureFromImage;
  begin
    inherited Create(1, 1, GL_RGBA, GL_BGRA, _max_mip, tp, ap); //*temporary* properties
    f_file_name:= _fname;
  end;

  destructor TTextureFromImage.Destroy;
  begin
    if Assigned(f_file_read_task) then f_file_read_task:= nil;
    if Assigned(f_img_decode_task) then f_img_decode_task:= nil;
    if Assigned(f_img_resample_task) then f_img_resample_task:= nil;
    inherited;
  end;

  procedure TTextureFromImage.UpdateHash;
  begin
    //do NOTHING. Hash remains empty until the file is loaded.
  end;

  //f_upload_stage
  const
    i_reading_file = 0;
    i_decoding_image = 1;
    i_resampling_image = 2;
    i_uploading_image = 3;

  function TTextureFromImage.Actualize: boolean;
  var
    i, height_part, mw, mh: integer;
    texels: int64;
    img: TImageDecodeTask;
    q: qword;

  begin
    case f_upload_stage of
      i_reading_file: begin
        if not Assigned(f_file_read_task)
          then f_file_read_task:= TFileLoadTask.Create(f_file_name);
        if f_file_read_task.Completed
          then begin
            f_upload_stage:= i_decoding_image;
            f_hash:= f_file_read_task.Hash;
          end;
        Exit(false);
      end;
      i_decoding_image: begin
        if not Assigned(f_img_decode_task) then begin
          if not Assigned(f_file_read_task.PContent)
            or (f_file_read_task.ContentSize <= 0)
          then Die(RuEn(
            'Не удалась загрузка изображения из файла'#10#13'  %0',
            'Failed to load image file'#10#13'  %0')
            , [UnmangleFileName(f_file_read_task.FileName)]);
          f_img_decode_task:= TImageDecodeTask.Create(
            f_file_read_task.PContent, f_file_read_task.ContentSize);
        end;
        if f_img_decode_task.Completed then begin
          f_width:= f_img_decode_task.Width;
          f_height:= f_img_decode_task.Height;
          f_imgformat:= f_img_decode_task.Format;
          f_format:= TextureFormatFromImageFormat;
          ValidateTextureDimensions; //results may depend on the format
          if (f_width <> f_img_decode_task.Width) or (f_height <> f_img_decode_task.Height)
            then f_upload_stage:= i_resampling_image
            else f_upload_stage:= i_uploading_image;
        end;
        Exit(false);
      end;
      i_resampling_image: begin
        if not Assigned(f_img_resample_task) then begin
          mw:= f_width;
          mh:= f_height;
          for i:= 1 to f_uploading_mip_level do begin //calculate the size of the current mip level
            mw:= mw div 2;
            mh:= mh div 2;
          end;
          f_img_resample_task:= TImageResampleTask.Create(
            f_img_decode_task, mw, mh, f_img_decode_task.Format, tp_cyclic in f_tprops);
        end;
        if f_img_resample_task.Completed
          then f_upload_stage:= i_uploading_image;
        Exit(false);
      end;
      i_uploading_image: begin
        if Assigned(f_img_resample_task)
          then img:= f_img_resample_task
          else img:= f_img_decode_task; //if there was no need to resample

        if _handle = 0 then GenResource;

        height_part:= Math.min(img.Height, Math.max(16, round((GIANT_TEXTURE_THRESHOLD * GIANT_TEXTURE_THRESHOLD) / img.Width)));

        glBindTexture(Target(), _handle);
        CheckGlError;


//***TODO I messed up interfacing with Vampyre, so using glTexSubImage doesn't work
//  (due to foirmat incompatibilities, I presume)
//  Anyway, uploading a whole mip level of sane size (1024x1024) only takes a few microseconds
//  and Vampyre crashes anyway when trying to load images larger than roughly 2500x2500

        //if false //(f_uploaded_part > 0) or (height_part < img.Height)
        //then begin
        //  height_part:= Math.min(height_part, img.Height - f_uploaded_part);
        //  if Mother^.Debug.Verbose then AddLog('    Uploading part of image %0x%1 for mip %2 of texture %3', [img.Width, height_part, f_uploading_mip_level, _handle]);
        //  glTexSubImage2d(GL_TEXTURE_2D, f_uploading_mip_level,
        //    0, f_uploaded_part,
        //    img.Width, height_part, img.Format, GL_UNSIGNED_BYTE,
        //      img.PPixels + TextureFormatToBytesPerTexel(img.Format) * img.Width * f_uploaded_part);
        //end
        //else begin
        if Mother^.Debug.Verbose then AddLog(
          '    Uploading image %0x%1 for mip %2 of texture %3 (internal format %4, image format %5)',
          [img.Width, img.Height, f_uploading_mip_level, _handle, GlTextureFormatToString(f_format), GlTextureFormatToString(img.Format)]);
        Mother^.Timer.UsecDelta(@q);
//  glTexSubImage2D: procedure(target: GLenum; level, xoffset, yoffset: GLint; width, height: GLsizei; format, atype: GLenum; const pixels: Pointer); extdecl = nil;
          //glTexSubImage2d(GL_TEXTURE_2D, f_uploading_mip_level, 0, 0,
          //  img.Width, img.Height, img.Format, GL_UNSIGNED_BYTE, img.PPixels);
          glTexImage2d(GL_TEXTURE_2D, f_uploading_mip_level, f_format,
            img.Width, img.Height, 0, img.Format, GL_UNSIGNED_BYTE, img.PPixels);

          f_uploaded_part:= 0;
          height_part:= img.Height;
        //end;
        CheckGlError;

        if (Mother^.Debug.Verbose) then  AddLog(
          '    ..done in %0μs', [round(Mother^.Timer.UsecDelta(@q))]);

        f_uploaded_part+= height_part;
        if (f_uploaded_part >= img.Height) then begin
          //uploading is finished
          if Assigned(f_img_resample_task) then begin
            f_img_resample_task:= nil;
          end;
          if f_uploading_mip_level < f_max_mip then begin
            //there are more mip levels to upload. Make another pass.
            f_uploaded_part:=0;
            Inc(f_uploading_mip_level);
            if Assigned(f_img_resample_task) then begin
              f_img_resample_task:= nil;
            end;
            f_upload_stage:= i_resampling_image;
            Exit(false); //continue in the next frame.
          end;
          if not (tp_keep_image in f_tprops) then f_img_decode_task:= nil;
          f_uploading_mip_level:= 0;
          f_uploaded_part:= 0;
          f_upload_stage:= -1;
          MarkAsActualized;
          Exit(true);
        end;
        Exit(false);
      end;
    else
      Die('Unknown upload stage %0', [f_upload_stage]);
    end;
  end;

  function TTextureFromImage.CedeImageDecodeTask: TImageDecodeTask;
  begin
    Assert(tp_keep_image in f_tprops, 'TTextureFromImage.CedeImageDecodeTask() : texture properties must include tp_keep_image!');
    if not Actualized then Exit(nil);
    Assert(Assigned(f_img_decode_task), 'TTextureFromImage.CedeImageDecodeTask() : the task is NIL!');
    Assert(f_img_decode_task.Completed, 'TTextureFromImage.CedeImageDecodeTask() : the task is not completed yet!');
    Result:= f_img_decode_task;
    f_img_decode_task:= nil;
  end;

  procedure TTexture.FreeResource;
  begin
    glDeleteTextures(1, @_handle);
    CheckGlError;
    _handle:= 0;
  end;


  function TTexture.Devour(c: TGenericAsset): boolean;
  begin
    f_tprops:= (c as TTexture).TextureProperties;
    f_depth:= (c as TTexture).Depth;
    f_format:= (c as TTexture).Format;
    f_imgformat:= (c as TTexture).ImageFormat;
    f_max_mip:= (c as TTexture).MaxMip;
    Result:= inherited Devour(c);
  end;

  function TTextureFromImage.Devour(c: TGenericAsset): boolean;
  begin
    f_file_name:= (c as TTextureFromImage).FileName;
    Result:= inherited Devour(c);
  end;

  function TTexture.BlankImageFormatFromFormat: glEnum;
  begin
    case f_imgformat of
      GL_DEPTH_COMPONENT: Result:= GL_FLOAT;
    else
      Result:= GL_UNSIGNED_BYTE;
    end;

  end;

  procedure TTexture.AllocateBlankTextureImage;
  var
    i, w, h, d: integer;
    q: qword;
  begin
    if (Mother^.Debug.Verbose) then Mother^.Timer.UsecDelta(@q);
    w:= f_width;
    h:= f_height;
    d:= f_depth;
    for i:=0 to f_max_mip do begin
      case Target() of
        GL_TEXTURE_1D: begin
          glTexImage1D (
            GL_TEXTURE_1D, i, f_format, w,
                       0, f_imgformat,  BlankImageFormatFromFormat(), NIL);
          CheckGlError;
          if ((w = 1)) and (i < f_max_mip)
            then Die('%0 is too many a mip level for a 1D texture of %1', [f_max_mip, f_width]);
          w:= w div 2;
        end;
        GL_TEXTURE_2D: begin
          glTexImage2D (
            GL_TEXTURE_2D, i, f_format, w, h,
                       0, f_imgformat, BlankImageFormatFromFormat(), NIL);
          CheckGlError;
          if ((w = 1) or (h = 1)) and (i < f_max_mip)
            then Die('%0 is too many a mip level for a texture of %1x%2', [f_max_mip, f_width, f_height]);
          w:= w div 2;
          h:= h div 2;
        end;
        GL_TEXTURE_3D: begin
          glTexImage3D (
            GL_TEXTURE_1D, i, f_format, w, h, d,
                       0, f_imgformat, BlankImageFormatFromFormat(), NIL);
          CheckGlError;
          if ((w = 1) or (h = 1) or (d = 1)) and (i < f_max_mip)
            then Die('%0 is too many a mip level for a 3D texture of %1x%2x%3', [f_max_mip, f_width, f_height, f_depth]);
          w:= w div 2;
          h:= h div 2;
          d:= d div 2;
        end;
      else
        Die('Unsupported texture target %0h in TTexture.AllocateBlankTextureImage()', [IntToHex(Target(), 4)]);
      end;
    end;
    glTexParameteri(Target(), GL_TEXTURE_BASE_LEVEL, 0);
    glTexParameteri(Target(), GL_TEXTURE_MAX_LEVEL, f_max_mip);
    CheckGlError;
    if (Mother^.Debug.Verbose) then AddLog(
        '    Allocated a blank texture image of %1x%2, %3 with %4 mip levels in %0μs',
        [round(Mother^.Timer.UsecDelta(@q)), f_width, f_height, GlTextureFormatToString(f_format), f_max_mip + 1]);
  end;

  function TTexture.AssumedWeight: float;  //in megabytes
  var
    i: integer;
    texels, mw, omw: float;

  begin
    Result:= 0;
    if (_handle = 0) then Exit;
    mw:= 1.0;
    omw:= 1.0;
    for i:= 1 to MaxMip do begin  //adding the weight of mip levels
      case Target() of
        GL_TEXTURE_1D: omw*= 0.5;
        GL_TEXTURE_3D: omw*= 0.125;
      else
        omw*= 0.25;
      end;
      mw+= omw;
    end;
    case Target() of
      GL_TEXTURE_1D: texels:= f_width;
      GL_TEXTURE_3D: texels:= f_width * f_height * f_depth;
      GL_TEXTURE_CUBE_MAP: texels:= 6 * f_width * f_height;
    else
      texels:= f_width * f_height;
    end;

    Result:= texels * TextureFormatToBytesPerTexel(f_format) * mw * ( 1 / (1024 * 1024));
  end;

  function TTexture.SortOrder: ffloat;
  begin
    Result:= FF_ONE;
  end;

  function TTexture.IsResident: boolean;
  var
    isres: GLBoolean = GL_TRUE;
  begin
    if not Assigned(glAreTexturesResident) then Exit(true);
    Result:= glAreTexturesResident(1, @_handle, @isres) = GL_TRUE;
  end;

  function TTexture.MinFilter: GLuint; //default depends on tp_nearest and mip number
  begin
    if tp_nearest in f_tprops then begin
      if f_max_mip > 0
        then Result:= GL_NEAREST_MIPMAP_NEAREST
        else Result:= GL_NEAREST;
    end
    else begin
      if f_max_mip > 0
        then Result:= GL_LINEAR_MIPMAP_LINEAR
        else Result:= GL_LINEAR;
    end;
  end;

  function TTexture.MagFilter: GLuint; //default depends on tp_nearest
  begin
    if tp_nearest in f_tprops
      then Result:= GL_NEAREST
      else Result:= GL_LINEAR;
  end;

  function  TTexture.Target: GLuint; //default GL_TEXTURE_2D
  begin
    Result:= GL_TEXTURE_2D;
  end;

  function TTexture.SmallestMipSize: integer;
  begin
    Result:= 16;
  end;

  function TTexture.DefaultFormat: GLuint;  //GL_RGBA8
  begin
    if tp_compressed in f_tprops
      then begin
        if tp_sharp_alpha in f_tprops
          then Result:= GL_COMPRESSED_RGBA_S3TC_DXT1_EXT
          else Result:= GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
      end
      else Result:= GL_RGBA8;
  end;

  function TTexture.DefaultImageFormat: GLuint; //GL_BGRA
  begin
    Result:= GL_BGRA;
  end;

  procedure TTexture.Bind;
  begin
    if not Actualized then begin
//      if not Exists then Die(RuEn(
//        'Вызов Bind() у несуществующего текстурного ресурса (%0)',
//        'Call to Bind() of a non-existent texture asset (%0)'
//      ), [TellInvalidAsset(Self)]);

      glBindTexture(Target(), Mother^.GAPI.StubTexture);  //Mightily flashing each frame, too
      exit;
    end;

    if _handle = 0 then begin
      if Mother^.Debug.StopOnOpenGlErrors then Die(RuEn(
        'Попытка забиндить нулевую текстуру',
        'Attempt to bind texture number zero'));

      glBindTexture(Target(), Mother^.GAPI.StubTexture);
      exit;
    end;

    //Only works on older video cards.
    //But this is *vital* if you have, like, only 0.125 Gb video RAM
    if Module.LoadBalancer.Enabled and not IsResident
      then Module.LoadBalancer.TextureIsNotInTextureMemory(Self.AssumedWeight());


    glBindTexture(Target(), _handle);
    if not Mother^.Debug.StopOnOpenGlErrors then begin
      if glGetError() <> 0
        then glBindTexture(Target(), Mother^.GAPI.StubTexture);
    end
    else CheckGlError;
  end;

  procedure TTexture.ValidateTextureDimensions; { Checks maximum size,
      makes sure the sizes are power of two on videocards not supporting NPOT,
      makes sure the highest mip level has dimensions divisible to 4 if
      the format is DXT or that highest mip is at least 1 pixel otherwise
    }
    var i, cw, ch, cd, ccw, cch, ccd: integer;
  begin
    case Target() of
      GL_TEXTURE_1D: begin
        mo_gmathbase.ValidateTextureDimension(f_width);
        cw:= f_width;
        for i:= 1 to MaxMip do begin
          cw:= cw div 2;
          if cw < SmallestMipSize then begin
            f_max_mip:= i - 1;
            break;
          end;
        end;
      end;
      GL_TEXTURE_2D, GL_TEXTURE_CUBE_MAP: begin
        mo_gmathbase.ValidateTextureDimension(f_width);
        mo_gmathbase.ValidateTextureDimension(f_width);
        cw:= f_width;
        ch:= f_height;
        for i:= 1 to MaxMip do begin
          cw:= cw div 2;
          ch:= ch div 2;
          if (cw < SmallestMipSize) or (ch < SmallestMipSize) then begin
            f_max_mip:= i - 1;
            break;
          end;
        end;
        case f_format of
          GL_COMPRESSED_RGB_S3TC_DXT1_EXT,
          GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
          GL_COMPRESSED_RGBA_S3TC_DXT3_EXT,
          GL_COMPRESSED_RGBA_S3TC_DXT5_EXT: begin
            //Make sure the smallest mip level divides to 4x4
            cw:= f_width;
            ch:= f_height;
            for i:= 1 to MaxMip do begin  //first scale down to the smallest mip
              cw:= cw div 2;
              ch:= ch div 2;
            end;
            cw:= Math.max(4, cw and not integer($03)); //second, make sure it divides by 4
            ch:= Math.max(4, ch and not integer($03));
            if MaxMip > 0 then begin
              ccw:= cw shl MaxMip; //third, scale back up in multiples of 2
              cch:= ch shl MaxMip;
              mo_gmathbase.ValidateTextureDimension(ccw); //four, make sure we haven't exceeded max. texture size
              mo_gmathbase.ValidateTextureDimension(cch);
              if (ccw <> cw) or (cch <> ch) then begin
                Dec(f_max_mip);
                ValidateTextureDimensions; //if we did, cut one mip level off and call recursively
              end
              else begin
                //the texture size scales *up*
                f_width:= ccw;
                f_height:= cch;
              end;
            end
            else begin
              //no need to check if there's only one mip, as the formula above rounds down
              //and also leaves POT values untouched
              f_width:= cw;
              f_height:= ch;
            end;
          end;
        end;
      end;
      GL_TEXTURE_3D: begin
        mo_gmathbase.Validate3dTextureDimension(f_width);
        mo_gmathbase.Validate3dTextureDimension(f_height);
        mo_gmathbase.Validate3dTextureDimension(f_depth);
        cw:= f_width;
        ch:= f_height;
        cd:= f_depth;
        for i:= 1 to MaxMip do begin
          cw:= cw div 2;
          ch:= ch div 2;
          cd:= cd div 2;
          if (cw < SmallestMipSize) or (ch < SmallestMipSize) or (cd < SmallestMipSize) then begin
            f_max_mip:= i - 1;
            break;
          end;
        end;
        case f_format of
          GL_COMPRESSED_RGB_S3TC_DXT1_EXT,
          GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
          GL_COMPRESSED_RGBA_S3TC_DXT3_EXT,
          GL_COMPRESSED_RGBA_S3TC_DXT5_EXT: begin
            //Make sure the smallest mip level divides to 4x4x4
            //Make sure the smallest mip level divides to 4x4
            cw:= f_width;
            ch:= f_height;
            cd:= f_depth;
            for i:= 1 to MaxMip do begin  //first scale down to the smallest mip
              cw:= cw div 2;
              ch:= ch div 2;
              cd:= cd div 2;
            end;
            cw:= Math.max(4, cw and not integer($03)); //second, make sure it divides by 4
            ch:= Math.max(4, ch and not integer($03));
            cd:= Math.max(4, cd and not integer($03));
            if MaxMip > 0 then begin
              ccw:= cw shl MaxMip; //third, scale back up in multiples of 2
              cch:= ch shl MaxMip;
              ccd:= cd shl MaxMip;
              Validate3dTextureDimension(ccw); //four, make sure we haven't exceeded max. texture size
              Validate3dTextureDimension(cch);
              Validate3dTextureDimension(ccd);
              if (ccw <> cw) or (cch <> ch) or (ccd <> cd) then begin
                Dec(f_max_mip);
                ValidateTextureDimensions; //if we did, cut one mip level off and call recursively
              end
              else begin
                //the texture size scales *up*
                f_width:= ccw;
                f_height:= cch;
                f_depth:= ccd;
              end;
            end
            else begin
              //no need to check if there's only one mip, as the formula above rounds down
              //and also leaves POT values untouched
              f_width:= cw;
              f_height:= ch;
              f_depth:= cd;
            end;
          end;
          end;
        end;
    else
      Die('Unsupported target %0h', [IntToHex(Target(), 4)]);
    end;
  end;

  function TTextureFromImage.TextureFormatFromImageFormat: GLuint;
  begin
    //! Vampyre Imaging supports a lot of formats, but Chentrah doesn't,
    //  converting them all to these four:
    case f_imgformat of
      GL_LUMINANCE: Result:= GL_LUMINANCE;
      GL_LUMINANCE_ALPHA: Result:=  GL_LUMINANCE_ALPHA; //GL_LUMINANCE4_ALPHA4, anyone?
      GL_BGR, GL_RGB: begin
        if tp_compressed in f_tprops
          then Result:= GL_COMPRESSED_RGB_S3TC_DXT1_EXT
          else Result:= GL_RGB8;
      end;
      0, GL_BGRA, GL_RGBA: begin
        if tp_compressed in f_tprops
          then begin
            if tp_sharp_alpha in f_tprops
              then Result:= GL_COMPRESSED_RGBA_S3TC_DXT1_EXT
              else Result:= GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
          end
          else Result:= GL_RGBA8;
      end;
    else
      Die('Unsupported image format %0h', [IntToHex(f_imgformat, 4)]);
    end;
  end;

  function TextureFormatToBytesPerTexel(format: GLuint): integer;
  begin
    case format of
      GL_LUMINANCE, GL_ALPHA8, GL_LUMINANCE8, GL_LUMINANCE4_ALPHA4 : Result:=1;
      GL_LUMINANCE_ALPHA, GL_LUMINANCE8_ALPHA8, GL_LUMINANCE16,
        GL_ALPHA16 : Result:=2;
      GL_RGB, GL_BGR, GL_RGB8 : Result:=3;
      GL_RGBA, GL_BGRA, GL_RGBA8, GL_LUMINANCE16_ALPHA16 : Result:=4;
      GL_RGB16: Result:= 6;
      GL_RGBA16: Result:= 8;
      GL_COMPRESSED_RGB_S3TC_DXT1_EXT,
        GL_COMPRESSED_RGBA_S3TC_DXT1_EXT: Result:= 1; //Wrong, must be 0.5! Just never use DXT1.
      GL_COMPRESSED_RGBA_S3TC_DXT3_EXT,
        GL_COMPRESSED_RGBA_S3TC_DXT5_EXT: Result:= 1;
    else
      Result:=4;
    end;
  end;


  procedure RegClasses;
  begin
    RegClass(TTexture);
    RegClass(TTextureFromImage);
  end;

end.
