{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This unit contains threading support for chentrah module,
      employing a hack for the Windows Structured Exception Handling
      mechanism (see cl_seh_hack.pp)

    This unit has some definitions borrowed from the System unit (FreePascal RTL)
      since these were defined in the implementation section,
      unavailable to other units.

 **********************************************************************}

{
    There is an inherent delay, especially with chained threaded tasks,
      as checking for completeness and adding the next task is only performed
      in the main thread, once per frame in the owning asset's Actualize method

    So, a task consisting of three steps would take three frames to complete.

    This is considered acceptable
      as THE MAIN GOAL IS SMOOTHNESS RATHER THAN ANY PERFORMANCE BOOST.

    The multithreading (in the module) could be disabled altogether
     by seting disable_module_multithreading to 1 in the main ini file.
     The background tasks are then performed in the main thread, doing
     unspeakable things to smoothness and framerate stability.

}

{$codepage utf-8}

unit mo_threads;

interface
uses SysUtils, Classes {$ifdef windows}, Windows {$else} , baseunix, dl {$endif}
   , math, mo_hub, mo_classes, syncobjs, typinfo;

type

  TThreadedTaskType = (
    ttt_Logic,
    ttt_File,
    ttt_Imaging,
    ttt_General,
    ttt_Render //Execute in the main thread *only*, thus able to do OpenGL
    //  in parallel.
  );
  TThreadedTaskTypes = set of TThreadedTaskType;
  TThreadedTasks = array[TThreadedTaskType] of TSortedList;

  { TBackgroundThread }

  TBackgroundThread = class(TThread)
  protected
    f_done: boolean;
    f_MyId: TThreadID;
    f_TaskTypes: TThreadedTaskTypes;
    f_ready: boolean;
  public
    constructor Create(task_types: TThreadedTaskTypes; ssize: ptruint = 0);
    property Done: boolean read f_done;
//    function CheckForGenericDyingYells(call_stack_steps_to_skip: integer = 1): UnicodeString;
//    procedure AddCrashMessage(w: UnicodeString);
    property Id: TThreadID read f_MyId;
    procedure Execute; override;
    procedure Sleep; virtual;
//    procedure PassMotherCallbackError(w: UnicodeString);
//    function StopDying: UnicodeString;
    function PerformOneTask: longbool;
//    function ItsTimeToTerminate: longbool;
    procedure Wait;
    destructor Destroy; override;
    property TaskTypes: TThreadedTaskTypes read f_TaskTypes;
    property ReadyToGo: boolean read f_ready;
  end;

  TLogicThread = class(TBackgroundThread)
    //Will NOT terminate until Logic.Saved is set to true!
  public
    CriticalSection: TCriticalSection;
    procedure WaitFormainThread(duration: cardinal = INFINITE);
    function WaitFormainThreadButOnlyIfItIsDesperate : boolean;
    constructor Create();
    destructor Destroy; override;
    procedure Execute; override;
  end;



  { TThreadManager }

  TDummyThreadManager = class
    //a stub for when multithreading is disabled
  private
    TaskTypesToRunInMainThread: TThreadedTaskTypes;
    f_started: boolean;
    f_Done: boolean;
    f_tsc_since_last_render: qword;
    f_go: boolean;
  public
    StopRequested: longbool;
    RenderDoneLogicMayProceed, { Critical section not applicable as the render
      has an alternate mode that *can* overlap with logic.
      Logic, on the other hand, *canNOT* overlap with render. }
    LogicDoneRenderMayProceed, { this one is for the render heartbeat not to
      freeze if the logic is stuck doing something lengthy. The renderer will
      skip a couple frames causing a minor HOM effect, then switch to a black
      screen with a progress indicator.
      The wait timeout corresponds to 15 FPS (67 ms) }
    RenderDesperatelyNeedsNextFrame: { this is to patch the issue of events
      apparently working differently in Windows 7 from Windows XP: render
      gets locked out when logic becomes busy, FPS drops to 1..10. While in XP
      it always runs smoothly.
      This event signals the logic to wait on the next cycle, to let render
      a word in. The maximum wait time is based on 35 FPS. }
      TEventObject;
    UsSinceLastRender: double;
    property Started: boolean read f_started;
    property Done: boolean read f_Done;
    constructor Create;
    destructor Destroy; override;
    procedure Start; virtual;
    function Stopped: system.boolean; virtual;
    procedure RemoveMeImFreed(thr: TBackgroundThread); virtual; abstract;
    function GetThreadNumById(tid:  TThreadID): integer; virtual;
    function GetCurrentThread: TBackgroundThread; virtual;
    function IsCurrentThreadTheFileThread: boolean; virtual;
    procedure Pulse; virtual;
    function CallLogicForRender: boolean; virtual; //called by the main thread
    procedure RunBackgroundTasksInMainThread (
      RecommendedTimeLimitMicroseconds: double); {the limit can, and will be
        overdrafted: if below this, can start yet another task, it says.
        There is no telling how long that task would take.}
    procedure InitializeChepersy;
    property ThreadsGo: boolean read f_go;
  end;

  TThreadManager = class (TDummyThreadManager)
  private
    T: array of TBackgroundThread;
    Cs: TCriticalSection; //TRTLCriticalSection;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Start; override;
    function Stopped: system.boolean; override;
    procedure RemoveMeImFreed(thr: TBackgroundThread); override;
    function GetThreadNumById(tid:  TThreadID): integer; override;
    function GetCurrentThread: TBackgroundThread; override;
    function IsCurrentThreadTheFileThread: boolean; override;
    function GetLogicThread: TLogicThread;
    procedure Pulse; override;
    function CallLogicForRender: boolean; override;//called by the main thread
  end;

  { TAbstractThreadedTask }

  TAbstractThreadedTask = class(TCgeObject) //Is NEVER saved,
    // so it should NOT be registered.
    // There is a special validating check for that
  private
    f_completed,
    f_failed: boolean;

    {Be VERY careful with calling its methods.
      In theory, it should be stuck in the queue, not changing, but who knows?
      Better to limit to reading its fields and properties representing fields directly.
    }

  protected
    function Perform: boolean; virtual; abstract; //false means reschedule to try again later. Errors MUST raise an exception.
  public
    //The constructor MUST call Reschedule after it finishes all initialization!
    function GetType: TThreadedTaskType; virtual;
    procedure Execute;
    procedure Reschedule; //if it depends on something that is not ready yet, then move it to the end of queue.
    property Completed: boolean read f_completed;
    property Failed: boolean read f_failed;
  end;
  CAbstractThreadedTask = class of TAbstractThreadedTask;


implementation
  uses
     mo_module, mo_logic, chepersy;

  constructor TDummyThreadManager.Create;
  var
    TaskType: TThreadedTaskType;
  begin
    for TaskType in TThreadedTaskType do Self.TaskTypesToRunInMainThread+= [TaskType];
    LogicDoneRenderMayProceed:= TEventObject.Create(nil, true, false, 'LogicDoneRenderMayProceed');
    RenderDoneLogicMayProceed:= TEventObject.Create(nil, true, false, 'RenderDoneLogicMayProceed');
    RenderDesperatelyNeedsNextFrame:= TEventObject.Create(nil, true, false, 'RenderDesperatelyNeedsNextFrame');
  end;

  constructor TThreadManager.Create;
  begin
    LogicDoneRenderMayProceed:= TEventObject.Create(nil, true, false, 'LogicDoneRenderMayProceed');
    RenderDoneLogicMayProceed:= TEventObject.Create(nil, true, false, 'RenderDoneLogicMayProceed');
    RenderDesperatelyNeedsNextFrame:= TEventObject.Create(nil, true, false, 'RenderDesperatelyNeedsNextFrame');
    Self.cs:= TCriticalSection.Create; //InitCriticalSection(Self.cs);
  end;

  destructor TDummyThreadManager.Destroy;
  begin
    LogicDoneRenderMayProceed.Free;
    RenderDoneLogicMayProceed.Free;
    RenderDesperatelyNeedsNextFrame.Free;
    inherited;
  end;

  destructor TThreadManager.Destroy;
  var i: integer;
  begin
    Cs.Enter;
    for i:= 0 to High(T) do
      if Assigned(T[i]) then T[i].Free;
    Cs.Leave;
    Cs.Free; //DoneCriticalSection(Self.cs);
    inherited;
  end;

  function TThreadManager.GetLogicThread: TLogicThread;
  begin
    if High(T) < 0 then Exit(nil);
    if not Assigned(T[0]) then Exit(nil);
    Result:= T[0] as TLogicThread;
  end;

  function TDummyThreadManager.GetThreadNumById(tid:  TThreadID): integer;
  begin
    Result:= -1;
  end;

  function TThreadManager.GetThreadNumById(tid:  TThreadID): integer;
  var i: integer;
  begin
    Result:= -1;
    EnterWv(Cs);
    for i:=0 to High(T) do begin
      if Assigned(T[i]) and (T[i].Id = tid) then begin
        Result:= i;
        break;
      end;
    end;
    Cs.Leave;
  end;

  function TDummyThreadManager.IsCurrentThreadTheFileThread: boolean;
  begin
    Result:= true;
  end;

  function TThreadManager.IsCurrentThreadTheFileThread: boolean;
  var bt: TBackgroundThread;
  begin
    EnterWv(Cs);
    bt:= Self.GetCurrentThread();
    if not Assigned(bt)
      then Result:= false
      else Result:= ttt_File in bt.TaskTypes;
    Cs.Leave;
  end;

  function TDummyThreadManager.GetCurrentThread: TBackgroundThread;
  begin
    Result:= nil;
  end;

  function TThreadManager.GetCurrentThread: TBackgroundThread;
  var i: integer;
  begin
    i:= GetThreadNumById(GetCurrentThreadId());
    if i < 0 then Exit(nil);
    Result:= T[i];
  end;

  procedure TThreadManager.RemoveMeImFreed(thr: TBackgroundThread);
  var i, j: integer;
  begin
    j:= -1;
    for i:= 0 to High(T) do
      if T[i] = thr then begin
        j:= i;
        break;
      end;
    if j >=0 then begin
      Cs.Enter;
      pointer(T[j]):= nil;
      Cs.Leave;
//Dis may happen when the manager is processing da list... aaaand, Boom.
     // if j < High(T) then T[j]:= T[High(T)];
     // SetLength(T, Length(T) - 1);
    end;
//    if Length(T) < 1 then Mother^.Module.HasThreadsRunning:= false;
  end;


  procedure TDummyThreadManager.Start;
  begin
    if Mother^.Debug.Verbose then AddLog('Starting dummy thread manager (multithreading is disabled)');
    f_started:= true;
    InitializeChepersy;
  end;


  procedure TThreadManager.Start;
  var
    TotalStacks: integer; // in megabytes
    i: integer;
    q: qword;
  begin
    if Mother^.Debug.Verbose then AddLog('The module is starting its thread manager...');
    f_started:= true;

    if Mother^.Module.MultithreadingDisabled then begin
      AddLog(RuEn(
        'Многопоточность запрещена, собираюсь выполнять все фоновые задачи в основном потоке!',
        'Multithreading is disabled, going to perform all background tasks in the main thread!'));
//      Mother^.Module.HasThreadsRunning:= false;
      Exit;
    end;

    Mother^.Timer.UsecDelta(@q);
    // ***TODO: add checking for the largest continuos region of unallocated address space with VirtualQuery

    //Mother threads have small stacks
    TotalStacks:= 16 * (1 + Mother^.State.NumMotherThreads); //assumed value!
    SetLength(T, 1);

    //the file reading thread is always there
       if Mother^.Debug.Verbose then AddLog('  creating the control logic thread #0...');
    T[0]:= TLogicThread.Create();
    TotalStacks+= Module.ControlLogicThreadStackSize div (1024 * 1024);

    if Mother^.CPU.Count <= 3 then begin
      //there is no sense in spawning them like roaches. one core for render, one for logic.
      SetLength(T, 2);
        if Mother^.Debug.Verbose then AddLog('  creating the universal thread #1...');
      T[1]:= TBackgroundThread.Create([ttt_File, ttt_Imaging, ttt_General]);
      TotalStacks+= Module.BackgroundThreadStackSize div (1024 * 1024);

    end else begin
      SetLength(T, 3);

         if Mother^.Debug.Verbose then AddLog('  creating the file thread #1...');
      T[1]:= TBackgroundThread.Create([ttt_File]);
      TotalStacks+= Module.BackgroundThreadStackSize div (1024 * 1024);

      //the firtst default thread (also image processing) is always there.
        if Mother^.Debug.Verbose then AddLog('  creating the default thread #2...');
      T[2]:= TBackgroundThread.Create([ttt_Imaging, ttt_General]);
      TotalStacks+= Module.BackgroundThreadStackSize div (1024 * 1024);

      //IF there are more CPU cores AND there's enough RAM and address space, spawn additional threads:
      for i:= 3 to
        //There are two actively working threads: the main thread and the logic thread.
        {$ifdef cpu32}
          min(3, Mother^.CPU.Count - 2) //thread stacks *exhaust* 32-bit address space like no tomorrow!
        {$else}
          min(6, Mother^.CPU.Count - 2)
        {$endif}
      do begin
        // The stacks are never going to occupy their full size.
        // That's just a precaution against heavily recursed chepersy operations.
        // So, their total is being compared to RAM total.
        if (TotalStacks + (Module.BackgroundThreadStackSize div (1024 * 1024)) > Mother^.State.PhysMemory)
          {$ifdef cpu32}
            //running out of address space is a very real danger for 32-bit aplications.
            //you won'ty believe how fragmented this shit usually is!
            or (TotalStacks > 250)
          {$endif}
         then break;
        SetLength(T, Length(T) + 1);
          if Mother^.Debug.Verbose then AddLog('  creating thread #%0...',[High(T)]);
        T[High(T)]:= TBackgroundThread.Create([ttt_General]);
        TotalStacks+= Module.BackgroundThreadStackSize div (1024 * 1024);
      end;

    end;
      if Mother^.Debug.Verbose then AddLog('  total thread stacks size (unconfirmed) is %0 Mb', [TotalStacks]);

    if Mother^.Debug.Verbose then AddLog('  Creating suspended threads took %0 microseconds', [round(Mother^.Timer.UsecDelta(@q))]);

    for i:= 0 to high(T) do begin
      if Mother^.Debug.Verbose then AddLog('  launching thread #%0...', [i]);
      T[i].Start;
    end;

    if Mother^.Debug.Verbose then AddLog('  Starting threads took %0 microseconds', [round(Mother^.Timer.UsecDelta(@q))]);
  end;

  procedure TDummyThreadManager.RunBackgroundTasksInMainThread(RecommendedTimeLimitMicroseconds: double);
  var
    q: Qword;
    MicrosecondsPassed: double;
    task: TAbstractThreadedTask;
  begin
    Mother^.Timer.UsecDelta(@q);
    MicrosecondsPassed:= 0.0;
    repeat //always execute at least one task
      task:= nil;
      if Assigned(Module.Logic)
        then task:= Module.Logic.GetThreadedTask(Self.TaskTypesToRunInMainThread);
      if not Assigned(task) then break;
      try
        task.Execute;
      except
        Die(RuEn(
          'Крах выполнения %0 в основном потоке',
          'Failed to perform %0 in the main thread'
          ), [string(task.ClassName)]);
      end;
      MicrosecondsPassed+= Mother^.Timer.UsecDelta(@q);
    until MicrosecondsPassed > RecommendedTimeLimitMicroseconds;
  end;

  procedure TDummyThreadManager.Pulse;
  var
    q: Qword;
    MicrosecondsPassed: double;
    oldwhen: TLogicNextPulseWhenEnum;
    i, cycles: integer;
  begin
    if not Assigned(Module.Logic) then Exit;

    Mother^.Timer.UsecDelta(@q);
    MicrosecondsPassed:= 0.0;

    try
      if not Assigned(Module.Logic) then begin
        try
            NoteThreadOccupation(tlk_Logic);
          MyLogClass.CreateOrLoad; { Assigns itself to the Module.Logic field.
             May load session if one exists, instead of just creating self. }
          Module.Logic.OnCreate;
        except
          Die(RuEn(
            'Сбой создания управляющей логики',
            'failed to create the control logic'));
        end;
          NoteThreadOccupation(tlk_Working);
      end;


      Module.Logic.FromMainThreadStallIfGoingTooFast;

      oldwhen:= Module.Logic.NextPulseWhen;
      case oldwhen of
        lnf_WaitForRender: Module.Logic.Pulse;
        lnf_Steady100Hz: repeat
            Module.Logic.Pulse;
          until Module.Logic.MicrosecondsLaggingBehind() <= 0;
        lnf_ASAPRenderAt30: begin
          for i:=1 to Math.max(2,  round( Mother^.Timer.FPS / 30))
                                            { a negative feedback of sorts }
            do Module.Logic.Pulse;
          Mother^.Timer.CullFrameRate:= 999; { Bullshit }
        end;
      else
        Die(MI_ERROR_PROGRAMMER_NO_BAKA, ['Unsupported Logic.NextPulseWhen='
          + GetEnumname( typeinfo(TLogicNextPulseWhenEnum),
            Ord(Module.Logic.NextPulseWhen)
           )]);
      end;
      LogicDoneRenderMayProceed.SetEvent();
    except
      Die(RuEn(
        'Крах выполнения логики в основном потоке (многопоточность запрещена)',
        'Failed to process logic in the main thread, as per disabled multithreading'));
    end;

    Self.RunBackgroundTasksInMainThread(
      Module.LoadBalancer.RecommendedTimeLimitForBackgroundTasksInMainThread
      - Mother^.Timer.UsecDelta(@q));

    f_done:= not Mother^.Module.SessionNotSaved and Mother^.Module.MustUnload;
  end;


  procedure TThreadManager.Pulse;
  var
    i: integer;
    w, tn: UnicodeString;
    AllDone: boolean = true;
    a: array of UnicodeString;
    ppw: PPUnicodeChar;
    pes: PMotherSEHState;
  begin
    w:= '';

    if not f_go then begin
      for i:= 0 to High(T) do begin
        if not Assigned(T[i]) then continue;
        if not T[i].ReadyToGo then Exit;
      end;
      f_go:= true;
      if Mother^.Debug.Verbose then AddLog('All threads are ready to go, the thread manager begins functioning.');
    end;

    //check thread statuses so that a screened crash message is always raised in the main thread
    for i:= 0 to High(T) do begin
      if not Assigned(T[i]) then continue;
      if not T[i].Done then continue; //thread always finishes execution if it crashes
      pes:= GetExceptionState(T[i].Id);
      if pes^.NowDying then begin
        if Mother^.Debug.Verbose then AddLog('The thread manager found that thread #%0 has crashed. Processing its error messages...',[i]);
        if w <> '' then w+= #10#13#10#13;
        w+= PUnicodeCharToUnicodeString(pes^.ThreadTitle) + ':'#10#13;
        w+= PUnicodeCharToUnicodeString(Mother^.Module.StopDying(pes));
      end;
    end;

    if w <> '' then begin
      if Mother^.Debug.Verbose then AddLog('The thread manager found that one or more threads have crashed. Will die in their stead now.');
      Mother^.ExceptionState.NowDying:= true;
      Mother^.ExceptionState.CallStackLogged:= true;//prevent from building call stack
      try //screen it! Mother reraises when appropriate.
        Die(w);
      except
      end;
      Mother^.ExceptionState.CallStackLogged:= false;
    end;

    EnterWv(Cs);
    f_Done:= true;
    for i:= 0 to High(T) do begin
      if Assigned(T[i]) then f_Done:= f_Done and T[i].Done;
    end;
    Cs.Leave;

{    if AllDone
      then //Mother^.Module.HasThreadsRunning:= false
      else
}
    if
      not Mother^.ExceptionState.NowDying
      and not Mother^.Module.ApplicationExiting
      and (High(T) > 0)
      and Assigned(T[0])
      and Assigned(Module)
      and Assigned(Module.Logic)
      and Assigned(Module.LoadBalancer)
    then Self.RunBackgroundTasksInMainThread(
       Module.LoadBalancer.RecommendedTimeLimitForBackgroundTasksInMainThread);
   end;


  //what I made this method for? I forgot >:(
  function TDummyThreadManager.Stopped: system.boolean;
  begin


  end;

  function TThreadManager.Stopped: system.boolean;
  begin


  end;

  function TDummyThreadmanager.CallLogicForRender: boolean;
  begin
    if not Assigned(Module.Logic) then Exit(false);
    Module.CachedLogicNextPulseWhen:= Module.Logic.NextPulseWhen;
    Result:= Module.Logic.Render;
  end;


  function TThreadmanager.CallLogicForRender: boolean;
  var
    LT: TLogicThread;
    wr, wr2: TWaitResult;
    MaximumWaitUntilAskLogicToWait,
    ExtraWaitUntilDespaired: longint;
//q: qword;
    function LogicDoRender: boolean;
    begin
      UsSinceLastRender:= 0.0;
      Mother^.Timer.UsecDelta(@f_tsc_since_last_render);

      LogicDoneRenderMayProceed.ResetEvent;
      try
//Mother^.Timer.UsecDelta(@q);
        EnterWv(LT.CriticalSection);
//addlog('E %0', [round(Mother^.Timer.UsecDelta(@q) / 1000)]);
        Module.CachedLogicNextPulseWhen:= Module.Logic.NextPulseWhen;

        if Result then begin
        end;

        Result:= Module.Logic.Render;

      finally
        LT.CriticalSection.Leave;
      end;
    end;
  begin
    LT:= GetLogicThread;
    if not Assigned(LT) then Exit(false);
    if not Assigned(Module.Logic) then Exit(false);
    //render continues drawing a blank screen  on at least 15 FPS if logic freezes

    UsSinceLastRender+= Mother^.Timer.UsecDelta(@f_tsc_since_last_render);

    MaximumWaitUntilAskLogicToWait:= Math.Max(0, Math.Min(1000 div 15,
      (1000 div 35) - round (UsSinceLastRender / 1000)));
    ExtraWaitUntilDespaired:= Math.Max(0, 1000 div 15 - MaximumWaitUntilAskLogicToWait);

//Mother^.Timer.UsecDelta(@q);
      NoteThreadOccupation (tlk_WaitingSync);
    wr:= LogicDoneRenderMayProceed.WaitFor(MaximumWaitUntilAskLogicToWait);
      NoteThreadOccupation (tlk_Working);
//addlog('MaximumWaitUntilAskLogicToWait = %0, actual time = %1', [MaximumWaitUntilAskLogicToWait, round(Mother^.Timer.UsecDelta(@q) / 1000.0)]);
    case wr of
      wrSignaled: begin
        Result:= LogicDoRender;
        Exit;
      end;
      wrTimeout: begin
        RenderDesperatelyNeedsNextFrame.SetEvent;
          NoteThreadOccupation (tlk_WaitingSync);
        wr2:= LogicDoneRenderMayProceed.WaitFor(ExtraWaitUntilDespaired);
          NoteThreadOccupation (tlk_Working);
        try
          case wr2 of
            wrSignaled: begin
              Result:= LogicDoRender;
            end;
         else
           Result:= false;
         end;
        finally
          RenderDesperatelyNeedsNextFrame.ResetEvent;
        end;
      end;
    else
     Result:= false;
    end;
  end;


{  function TThreadManager.Crashed: system.boolean;
  var i: integer;
  begin
    for i:= 0 to high(T) do if Assigned(T[i]) and T[i].Crashed then Exit(true);
    Result:= false;
  end;
}
  {$ifdef windows}
    {$ifdef cpu32}
      //No need to. {$include mo_threads_seh_hack.inc}
    {$endif}
  {$endif}

 // =============================================================================

  procedure TBackgroundThread.Execute;
  var
   tn, tt: UnicodeString;
   i: integer;
   pes: PMotherSehState;
  begin
    f_Done:= false;
    Self.f_MyId:= GetCurrentThreadId();

    if ttt_Imaging in f_TaskTypes then Mother^.Image.CorrectThreadId:= Self.f_MyId;
      //will exit with an error if try to call imaging API from any other thread

    if Mother^.Debug.Verbose then AddLog('Execute begins, Id=%0',[pointer(GetCurrentThreadId())]);
      {$ifdef cpu32}
      {MSDN: Issuing SetUnhandledExceptionFilter replaces the existing top-level
      exception filter for all existing and all future threads in the calling process.
      }

//Why the FUCK was I thinking it's a per-thread setting?

      //SetUnhandledExceptionFilter(@MyExceptionFilter);
      {$endif}

    i:= Module.Threadmanager.GetThreadNumById(GetCurrentThreadId());

    tt:= RuEn('общего назначения', 'general-purpose');
    if ttt_Logic in f_TaskTypes then tt:= RuEn('управляющей логики', 'control logic');
    if ttt_File in f_TaskTypes then tt:= RuEn('для файловых операций', 'file processing');
    if ttt_Imaging in f_TaskTypes then tt:= RuEn('для обработки изображений', 'image-processing');
    tn:= PervertedFormat(RuEn(
      'Фоновый поток модуля №%0 %1',
      'Module background %1 thread #%0'), [i, tt]);

    Mother^.Core.AllocateThreadExceptionState(PUnicodeChar(tn));
    { The allocated state
      is used by mother callback functions the thread may call.
      For the module, it's a black box }


   (*  Накуй не нужно
      FillChar(CgeSignalHandler, sizeof(CgeSignalHandler), 0);
      CgeSignalHandler.sa_handler := SigActionHandler(@MySignalDispatcher);
      CgeSignalHandler.sa_flags:=SA_SIGINFO;
      FpSigAction(SIGFPE, @CgeSignalHandler, @OldSigFpeHandler);
      FpSigAction(SIGSEGV,@CgeSignalHandler, @OldSigSegvHandler);
      FpSigAction(SIGBUS, @CgeSignalHandler, @OldSigBusHandler);
      FpSigAction(SIGILL, @CgeSignalHandler, @OldSigIllHandler);
    *)
    pes:= GetExceptionState();
    pes^.AbbrTitleForIndicator:= 'bg' + IntToStr(i);

    f_ready:= true;
    repeat sysutils.Sleep(1) until Module.Threadmanager.ThreadsGo;
    if Mother^.Debug.Verbose then AddLog('Work started (Id=%0)',[pointer(GetCurrentThreadId())]);

    try
      repeat
        if not Self.PerformOneTask then Self.Sleep;
      until
        pes^.NowDying //There could be errors screened by accident
        //or Self.ItsTimeToTerminate;
        or Mother^.Module.MustUnload
        or (Mother^.Module.State = ms_Crashed);
      if pes^.NowDying then AddLog('silent CRASH');
    except
      AddLog('CRASH!');
      // do NOT call StopDying, thus leaving thread exception state for the thread manager to process
    end;
      NoteThreadOccupation(tlk_Sleeping);

    f_done:= true;
    pes^.ThreadFinished:= true;
    if Mother^.Debug.Verbose then AddLog('Execute ended');
  end;

  procedure TDummyThreadmanager.InitializeChepersy;
  begin
//sleep(2000);
//AddLog('Бляааа..');
//byte(nil^):= 0;
//AddLog('..ть.');

     if Mother^.Debug.Verbose then AddLog('Initializing Chepersy...');
    InitCgeClasses;
    ClassesRegistrationStart;
    RegClass(MyLogClass);
    ClassesRegistrationEnd;
     if Mother^.Debug.Verbose then AddLog('Chepersy initialized.');
  end;


  procedure TLogicThread.Execute;
  var
   tn: UnicodeString;
   i, sll: integer;
   pes: PMotherSehState;
   pot: boolean;
   when: TLogicNextPulseWhenEnum;
   lag: double;
  begin
    f_Done:= false;
    f_MyId:= GetCurrentThreadId();
    if Mother^.Debug.Verbose then AddLog('Execute begins, Id=%0',[pointer(GetCurrentThreadId())]);
    i:= Module.Threadmanager.GetThreadNumById(GetCurrentThreadId());
    tn:= PervertedFormat(RuEn(
      'Логический поток модуля №%0',
      'Module logic thread #%0'), [i]);
    Mother^.Core.AllocateThreadExceptionState(PUnicodeChar(tn));
    pes:= GetExceptionState();
    pes^.AbbrTitleForIndicator:= 'LOG';
    f_ready:= true;
    repeat sysutils.Sleep(1) until Module.Threadmanager.ThreadsGo;
    if Mother^.Debug.Verbose then AddLog('Work started (Id=%0)',[pointer(GetCurrentThreadId())]);

    try
      //Mother^.Module.LogicThreadRunning:= true;
      NoteThreadOccupation(tlk_Logic);
      if not chepersy.CpsFinalized then Module.ThreadManager.InitializeChepersy;

      repeat
        try
          EnterWv(CriticalSection);
          if not Assigned(Module.Logic) then begin
            try
                if Mother^.Debug.Verbose then AddLog('Creating control logic, %0...',[string(MyLogClass.ClassName)]);
              MyLogClass.CreateOrLoad; { Assigns itself to the Module.Logic field.
                 May load session if one exists, instead of just creating self. }
              Module.Logic.OnCreate;
                if Mother^.Debug.Verbose then AddLog('Finished creating of %0',[string(MyLogClass.ClassName)]);
            except
              CriticalSection.Leave;
              Die(RuEn(
                'Сбой создания управляющей логики',
                'failed to create the control logic'));
            end;
          end;
          //if Self.ItsTimeToTerminate then Module.Logic.MustSave;
          try
            when:= Module.Logic.NextPulseWhen;
            case when of
              lnf_WaitForRender : begin
                CriticalSection.Leave;
                WaitFormainThread(INFINITE);
                EnterWv(CriticalSection);
                Module.Logic.Pulse;
              end;
              lnf_Steady100Hz: begin
                lag:= Module.Logic.MicrosecondsLaggingBehind; //includes AccumulatedLag
                Module.Logic.AccumulatedLag:= Math.Min(500000.0, Math.Max(0, lag));
                sll:= Math.Max(0, Math.Min( round( 1000/15), round(- lag / 1000)));
                CriticalSection.Leave;
                if not WaitFormainThreadButOnlyIfItIsDesperate { may wait much
                                           longer if FPS has dropped below 30 }
                then Mother^.Timer.HiResSleep(sll); //let the critical section do all work

    //addlog('(%0)', [sll]);
                EnterWv(CriticalSection);
                Module.Logic.Pulse;
              end;
              lnf_ASAPRenderAt30: begin
                CriticalSection.Leave;
                WaitFormainThreadButOnlyIfItIsDesperate;
                EnterWv(CriticalSection);
                Module.Logic.Pulse; //again, the critical section provides
              end;
            else
              Die(MI_ERROR_PROGRAMMER_NO_BAKA, ['Unsupported Logic.NextPulseWhen='
                + GetEnumname(typeinfo(TLogicNextPulseWhenEnum), Ord(when))
                + ' (ord=' + IntToStr(Ord(when)) + ')']);
            end;
          except
            { The exception is processed by the thread manager
                in the main thread. See TThreadManager.Pulse. }
            if Mother^.Debug.Verbose then AddLog('Module.Logic.Pulse crash.');
          end;
        finally
          CriticalSection.Leave;
        end;
      until
        pes^.NowDying //There could be errors screened by accident
        or Mother^.Module.MustUnload
        or (Mother^.Module.State = ms_Crashed) { Note: ms_Crashed is used
             only bor debugging (BSOD mode). The module is not unloaded but
             freezes to preserve the console content. Nevertheless, if there is
             a session to save it will be saved and all module threads will
             stop.
             Normally, module crashing leads straight to ms_Unloading.
           }
        or not Assigned(Module.Logic)
           //or not Mother^.Module.SessionNotSaved
           //or Module.Logic.Saved
        or Module.Logic.Crashed
      ;

      if Assigned(Module.Logic)
        and not Module.Logic.Crashed
        and Mother^.Module.SessionNotSaved
      then begin
        AddLog(RuEn(
          'Нештатное завершение, пытаюсь сохранить сессию.',
          'Incorerect shutdown, trying to save the session.'));
        Module.Logic.EmergencySaveSession;
      end;

      if pes^.NowDying then AddLog('silent CRASH');
      pes^.ThreadFinished:= true;
    except
      if Mother^.Debug.Verbose then AddLog('CONTROL LOGIC THREAD CRASH!');
      // do NOT call StopDying, thus leaving thread exception state for the thread manager to process
    end;
      NoteThreadOccupation(tlk_Sleeping);
    f_done:= true;
//    Mother^.Module.LogicThreadRunning:= false;
    if Mother^.Debug.Verbose then AddLog('Execute ended');
  end;

  constructor TBackgroundThread.Create(task_types: TThreadedTaskTypes; ssize:ptruint = 0);
  begin
    ssize:= max(ssize, Module.BackgroundThreadStackSize);
    try
      inherited Create(true, ssize);
      FreeOnTerminate:= false;
    except
      Die(RuEn('Сбой при создании потока (стек %0 Мб)', 'Failed to create thread (stack %0 Mb)'), [(ssize * 1.0) / (1024 * 1024)] );
    end;
    f_TaskTypes:= task_types;
    Inc(Mother^.State.NumModuleThreads);
//    Mother^.Module.HasThreadsRunning:= true;
  end;


  constructor TLogicThread.Create();
  begin
    inherited Create([ttt_Logic], Module.ControlLogicThreadStackSize);
    CriticalSection:= TCriticalSection.Create;
  end;

  destructor TLogicThread.Destroy();
  begin
    CriticalSection.Free;
    inherited;
  end;

  destructor TBackgroundThread.Destroy;
  var
    myname: UnicodeString;
  begin
    myname:= Self.ClassName;
    if Mother^.Debug.Verbose then AddLog('%0.Destroy begins',[myname]);
    Dec(Mother^.State.NumModuleThreads);
    Module.ThreadManager.RemoveMeImFreed(Self);
    inherited;
    if Mother^.Debug.Verbose then AddLog('%0.Destroy ended',[myname]);
  end;


{  procedure TBackgroundThread.AddCrashMessage(w: UnicodeString);
  begin
    //if Self.f_CrashMessage <> '' then w+= #10#13#10#13;
    Self.f_CrashMessage:= w + Self.f_CrashMessage;
    Self.f_crashed:= true;
  end;
}
(*  function TBackgroundThread.CheckForGenericDyingYells(call_stack_steps_to_skip: integer): UnicodeString;
  var
    w: UnicodeString;
    a, ca:pointer;
    i: integer;
  begin
    result:= '';
    if ExceptObject is Exception then begin
      w:= TellException(ExceptObject as Exception);
      if (w <> '' ) and (result <> '') then result+= #10#13#10#13 + w;
    end;
    if not CallStackLogged then begin
      CallStackLogged:= Yes;
      i:= 0;
      a:= get_frame;
      while Assigned(a) do begin
        ca:= get_caller_addr(a);
        {ignore the first three callers. These always are:
          0). call to this method from TThreadManager.PassDying()
          1). call to TThreadManager.PassDying() from Die()
          2). Call to Die() itself
        }
        if Assigned(ca) then begin
          if i = (call_stack_steps_to_skip) then begin
            if result <> '' then result+= #10#13#10#13;
            result+= 'Call stack:';
          end;
          if i >= (call_stack_steps_to_skip) then result+= #10#13'  ' + ShortExpAddr(ca);
          inc(i);
        end
        else Break;
        if i > MAX_STACK_FRAMES_DUMP then break;
        a:= get_caller_frame(a);
      end;
    end;
  end;

*)

(*
Идея состоит в том, что поток забирает задачу из списка, после чего она пропадает из дерева.
Попытка сохранить работающую задачу должна приводить к краху.
Это не на неё должны ссылаться, а она должна ссылаться.
Выполненные задачи тоже в дерево не входят.
*)


  procedure TBackgroundThread.Sleep;
  begin
    {
     I don't think this is of much use, as threads usually can't sleep
      for short enough time intervals.
     TimeBeginPeriod(1) may have worked in Windows 98, or even XP,
       but I doubt it works in Vista and above

     ***TODO: perform a series of tests to measure the sleeping granularity
        of Windows 7 as I did back in 1998 for Windows 95!
    }
    Mother^.Timer.HiResSleep(round(1000 / Math.min(200, 1.7 * Math.max(FPS_CULL_DEFAULT, Mother^.Timer.CullFrameRate))));
  end;

  procedure TLogicThread.WaitFormainThread(duration: cardinal = INFINITE);
  var
    wr: TWaitResult;
    remaining, chunk: cardinal;
  begin
     NoteThreadOccupation(tlk_WaitingSync);
    remaining:= duration;
    repeat
      chunk:= Math.min(1000 div 15, remaining);
      wr:= Module.ThreadManager.RenderDoneLogicMayProceed.WaitFor(chunk);
      case wr of
         wrSignaled: begin
           Module.ThreadManager.RenderDoneLogicMayProceed.ResetEvent;
           remaining:= 0;
         end
      else
        if remaining <> INFINITE then remaining-= chunk;
        if Mother^.Module.State = ms_Crashed then remaining:= 0; {
          In this rare case, the main thread gets suspended until user responds
          to the blue screen of death
        }
      end;
    until remaining = 0;
      NoteThreadOccupation(tlk_Working);
  end;

  function TLogicThread.WaitFormainThreadButOnlyIfItIsDesperate: boolean;
  begin
    Result:= wrSignaled =
                Module.ThreadManager.RenderDesperatelyNeedsNextFrame.WaitFor(0);
    if Result
      then WaitForMainThread(1000 div 15);
  end;

{
  function TBackgroundThread.StopDying: UnicodeString;
  begin
    Result:= f_crashMessage;
    f_crashed:= false;
  end;
}

  function TBackgroundThread.PerformOneTask: longbool;
  var
    t: TAbstractThreadedTask = nil;
  begin
    if Mother^.State.BatteryRemaining <= 100 then begin
      //all threads except the two default ones go into hibernation
      //like bears in winter when laptop is running on battery
      if not ((ttt_File in f_TaskTypes) or (ttt_Imaging in f_TaskTypes))
        then sysutils.Sleep(1000);
    end;
    if Assigned(Module.Logic)
      then t:= Module.Logic.GetThreadedTask(f_TaskTypes); //is responsible for critical section
    if Assigned(t) then begin
      if Mother^.Debug.Verbose then AddLog('got a task: %0',[string(t.ClassName)]);
        NoteThreadOccupation(tlk_Working);
      t.Execute;
      if Mother^.Debug.Verbose then begin
        if t.Completed
          then AddLog('  completed %0',[string(t.ClassName)])
          else AddLog('  rescheduled %0',[string(t.ClassName)]);
//Now this is automatic because threaded tasks are refcounted too.
//If it doesn't have someone else holding a reference to it, it dies.
//        if t.FreeOnCompletion
//          then AddLog('  deleting unmanaged %0',[string(t.ClassName)]);;
      end;
//      if t.FreeOnCompletion then t.Free;
      Result:= true;
    end
    else Result:= false;
  end;

{
  function TBackgroundThread.ItsTimeToTerminate: longbool;
  begin
    Result:= Mother^.Module.State <> ms_Loaded;
  end;
}

{
  procedure TBackgroundThread.PassMotherCallbackError(w: UnicodeString);
  begin
    f_module_callback_error:= w;
  end;
 }

  procedure TBackgroundThread.Wait;
  begin


  end;

  // ==========================================================================

  function TAbstractThreadedTask.GetType: TThreadedTaskType;
  begin
    Result:= ttt_General;
  end;


  procedure TAbstractThreadedTask.Execute;
  begin
    if Perform
      then f_completed:= true //and then it just sits there until the owner asset
                      //checks its threaded task status in its Actualize method
      else Reschedule;
  end;


  procedure TAbstractThreadedTask.Reschedule;
  begin
    //worst case, a thread would spend its alotted time interval rescheduling tasks in an endless cycle.
    Assert(Assigned(Module.Logic), 'Control logic is not valid!');
    Module.Logic.AddThreadedTask(Self);
  end;


{  procedure TAbstractThreadedTask.Finish;
  begin
    //do nothing;
  end;
}
end.
